//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup Tool "FUNSSEL" developed for the 
// Simulator League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public 
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#include "FilterSocket.h"
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <errno.h>
#include <iomanip>
#include <signal.h>

// cwz-build-command: make
//#define DEBUG

/* if defined, all messages are logged. Messages smaller than MSG_LENGTH
   with full content, for longer messages the length is printed
*/
//#define MSG_LENGTH 512

class FilterSocket_Error: public Exception {
public:
  FilterSocket_Error(const char *msg, int clt);
  FilterSocket_Error(const char *stamp, int what_err, int clt);
};

FilterSocket_Error::FilterSocket_Error(const char *stamp, int err, int clt)
{
  char buf[256];
  if (clt > 0)
    sprintf(buf, "filtersocket: errno=%d in %s for client %d",
	    err, stamp, clt);
  else
    sprintf(buf, "filtersocket: errno=%d in %s", err, stamp);
  descr = buf;
}

FilterSocket_Error::FilterSocket_Error(const char *error, int clt)
{
  char buf[256];
  if (clt > 0)
    sprintf(buf, "filtersocket: error %s for client %d", error, clt);
  else
    sprintf(buf, "filtersocket: error %s", error);
  descr = buf;
}

FilterSocket::FilterSocket(int fakePort, const char *remote_host,
			   int sniffedPort,
			   int maxSize)
  throw (Exception): maxMsgSize(maxSize)
{
  // set up fake port
  primary.sock = socket(AF_INET, SOCK_DGRAM, 0);
  fcntl(primary.sock, F_SETFL, O_NONBLOCK|O_ASYNC);
  fcntl(primary.sock, F_SETOWN, getpid());

  struct sockaddr_in saddr;
  bzero((char *)&saddr, sizeof(saddr));
  saddr.sin_family      = AF_INET;
  saddr.sin_addr.s_addr = htonl(INADDR_ANY);
  saddr.sin_port        = htons(fakePort);
  if (bind(primary.sock, (struct sockaddr *)&saddr, sizeof(saddr)) < 0)
    throw FilterSocket_Error("bind server port", errno, -1);

  // prepare "login" address of server
  struct hostent *hp;
  hp = gethostbyname(remote_host);
  if (!hp)
    throw FilterSocket_Error("server host unknown", -1);

  bzero((char *)&primary.addr, sizeof(primary.addr));
  primary.addr.sin_port   = htons(sniffedPort);
  primary.addr.sin_family = AF_INET;
  primary.addr.sin_addr   = *((in_addr *)hp->h_addr);
}

FilterSocket::~FilterSocket() throw (Exception)
{
  if (close(primary.sock) < 0)
    throw FilterSocket_Error("close server port", errno, -1);
  list<FilterClient>::const_iterator it;
  for (it = clients.begin(); it != clients.end(); it++)
    if (close(it->sock) < 0)
      throw FilterSocket_Error("close client port", errno, it->sock);
}

void FilterSocket::drop(int id) throw (Exception)
{
  // search for a client with that id
  list<FilterClient>::iterator it;
  for (it = clients.begin(); it != clients.end(); it++) {
    if (it->sock == id)
      break;
  }
  if (it == clients.end())
    return;
  if (close(it->sock) < 0)
    throw FilterSocket_Error("close client port", errno, it->sock);
  clients.erase(it);
}

FilterMessage FilterSocket::getNextMessage() throw (Exception)
{
  // step 1: search for a message for the server
  sockaddr_in caddr;
  size_t size = sizeof(caddr);
  char buf[maxMsgSize];
  int  n = recvfrom(primary.sock, (void *)buf, maxMsgSize,
		    0, (sockaddr *)&caddr, &size);

  if ((n == -1) && (errno != EAGAIN))
    throw FilterSocket_Error("polling server port", errno, -1);

  if (n > 0) {
    // a message was sent to the server
    string msg(buf, n);
    const FilterClient &client = findClient(caddr);
#ifdef DEBUG
    cout << "recvd msg for server from " << client.sock
	 << " " << ntohs(client.addr.sin_port)
	 << " " << hex << client.addr.sin_addr.s_addr << dec << endl;
#ifdef MSG_LENGTH
    if (n < MSG_LENGTH) {
      char test[n + 1];
      memcpy(test, buf, n);
      test[n] = 0;
      cout << "<<" << test << ">> " << n << endl;
    }
    else
      cout << "len " << n << endl;
#endif
#endif
    return FilterMessage(msg, client.sock, FilterMessage::ToServer);
  }

  // step 2: search for a message from the server on client ports
  list<FilterClient>::iterator it;
  for (it = clients.begin(); it != clients.end(); it++) {
    sockaddr_in tmp_addr;
    size_t size = sizeof(tmp_addr);
    int  n = recvfrom(it->sock, (void *)buf, maxMsgSize,
		      0, (sockaddr *)&tmp_addr, &size);

    if ((n == -1) && (errno != EAGAIN))
      throw FilterSocket_Error("polling client port", errno, it->sock);

    if (n > 0) {
      // a message was sent to the client
      // use senders address as server for subsequent sends
      // this is noda's protocol...
#ifdef DEBUG
      cout << "recvd from server for " << it->sock
	   << " " << ntohs(it->addr.sin_port)
	   << " " << hex << it->addr.sin_addr.s_addr << dec << endl;
      if (it->mysrv.sin_port != tmp_addr.sin_port)
	cout << "got new port "
	     << ntohs(tmp_addr.sin_port) << " "
	     << hex << tmp_addr.sin_addr.s_addr << dec << endl;
#ifdef MSG_LENGTH
      if (n < MSG_LENGTH) {
	char test[n + 1];
	memcpy(test, buf, n);
	test[n] = 0;
	cout << "<<" << test << ">> " << n << endl;
      }
      else
	cout << " len " << n << endl;
#endif
#endif
      it->mysrv = tmp_addr;
      string   msg(buf, n);
      return FilterMessage(msg, it->sock, FilterMessage::FromServer);
    }
  }
  // no message at all
  return FilterMessage(string(), -1, 0);
}

void FilterSocket::sendToClient(const FilterClient &client, const char *buf, int n)
  throw (Exception)
{
#ifdef DEBUG
  cout << "sending to " << client.sock
       << " " << ntohs(client.addr.sin_port)
       << " " << hex << client.addr.sin_addr.s_addr << dec 
       << " as server" << endl;
#ifdef MSG_LENGTH
  if (n < MSG_LENGTH) {
    char test[n + 1];
    memcpy(test, buf, n);
    test[n] = 0;
    cout << "<<" << test << ">> " << n << endl;
  }
  else
    cout << " len " << n << endl;
#endif
#endif
  // messages for proxy clients are not forwarded
  if (client.addr.sin_addr.s_addr == htonl(INADDR_ANY))
    return;

  int sent;
  for (;;) {
    sent = sendto(primary.sock, (void *)buf, n, 0, (sockaddr *)&client.addr,
		  sizeof(client.addr));
    if (sent == n)
      break;
    else {
      if (errno == ECONNREFUSED) {
	// client died. ignore from now on
	cerr << "client " << client.sock << " not reachable" << endl;
	drop(client.sock);
	return;
      }

      if (errno != EAGAIN)
	throw FilterSocket_Error("transmission to client", errno, client.sock);
    }
  }
}

void FilterSocket::sendForClient(const FilterClient &client, const char *buf, int n)
  throw (Exception)
{
#ifdef DEBUG
  cout << "sending to server at " << ntohs(client.mysrv.sin_port)
       << " " << hex << client.mysrv.sin_addr.s_addr << dec
       << " from socket " << client.sock << endl;
#ifdef MSG_LENGTH
  if (n < MSG_LENGTH) {
    char test[n + 1];
    memcpy(test, buf, n);
    test[n] = 0;
    cout << "<<" << test << ">> " << n << endl;
  }
  else
    cout << " len " << n << endl;
#endif
#endif
  int sent;
  for (;;) {
    sent = sendto(client.sock, (void *)buf, n, 0,
		  (sockaddr *)&client.mysrv, sizeof(client.mysrv));
    if (sent == n)
      break;
    else {
      if (errno == ECONNREFUSED) {
	// client died. ignore from now on
	cerr << "client " << client.sock << " not reachable" << endl;
	drop(client.sock);
	return;
      }
      
      if (errno != EAGAIN)
	throw FilterSocket_Error("transmission to client", errno, client.sock);
    }
  }
}

void FilterSocket::send(const FilterMessage &msg) throw (Exception)
{
  int id = msg.client;
  // search for a client with that id
  list<FilterClient>::const_iterator it;
  for (it = clients.begin(); it != clients.end(); it++) {
    if (it->sock == id)
      break;
  }
  if (it == clients.end())
    // ignore packages for not yet started clients
    return;
  if (it->addr.sin_addr.s_addr == htonl(INADDR_ANY))
    throw FilterSocket_Error("send to non physical client", id);

  if (msg.fromClient())
    sendForClient(*it, msg.message.data(), msg.message.length());
  else
    sendToClient(*it, msg.message.data(), msg.message.length());
}

bool FilterSocket::clientExists(int client)
{
  list<FilterClient>::const_iterator it;
  for (it = clients.begin(); it != clients.end(); it++)
    if (it->sock == client) 
      return true;  
  return false;
}

const FilterSocket::FilterClient &FilterSocket::findClient(sockaddr_in addr)
{
  list<FilterClient>::const_iterator it;
  for (it = clients.begin(); it != clients.end(); it++) {
    if ((it->addr.sin_port == addr.sin_port) &&
	(it->addr.sin_addr.s_addr == it->addr.sin_addr.s_addr))
      return *it;
  }
  // ok, a new client connected
  FilterClient clnt;
  // store addresses
  clnt.addr = addr;
  // first send goes out to the soccerserver main port
  clnt.mysrv = primary.addr;

  // open socket for sending in clients name
  clnt.sock = socket(AF_INET, SOCK_DGRAM, 0);
  fcntl(clnt.sock, F_SETFL, O_NONBLOCK|O_ASYNC);
  fcntl(clnt.sock, F_SETOWN, getpid());

  struct sockaddr_in saddr;
  bzero((char *)&saddr, sizeof(saddr));
  saddr.sin_family      = AF_INET;
  saddr.sin_addr.s_addr = htonl(INADDR_ANY);
  saddr.sin_port        = htons(0);
  if (bind(clnt.sock, (struct sockaddr *)&saddr, sizeof(saddr)) < 0)
    throw FilterSocket_Error("bind client port", errno, -1);

  clients.push_back(clnt);
  return clients.back(); 
}

int FilterSocket::connect() throw (Exception)
{
  FilterClient clnt;
  // illegal address since there is no physical client behind it
  clnt.addr.sin_addr.s_addr = htonl(INADDR_ANY);
  clnt.mysrv = primary.addr;

  // open socket for sending in clients name
  clnt.sock = socket(AF_INET, SOCK_DGRAM, 0);
  fcntl(clnt.sock, F_SETFL, O_NONBLOCK|O_ASYNC);
  fcntl(clnt.sock, F_SETOWN, getpid());

  struct sockaddr_in saddr;
  bzero((char *)&saddr, sizeof(saddr));
  saddr.sin_family      = AF_INET;
  saddr.sin_addr.s_addr = htonl(INADDR_ANY);
  saddr.sin_port        = htons(0);
  if (bind(clnt.sock, (struct sockaddr *)&saddr, sizeof(saddr)) < 0)
    throw FilterSocket_Error("bind proxy client port", errno, -1);

  clients.push_back(clnt);
  return clnt.sock;
}

void FilterSocket::myhandler(int)
{
}

void FilterSocket::enableInts()
{
  sigset_t sset;
  sigemptyset(&sset);
  struct sigaction saction;
  saction.sa_handler = myhandler;
  saction.sa_flags   = SA_RESTART;
  saction.sa_mask    = sset;
  sigaddset(&sset, SIGIO);
  sigprocmask(SIG_UNBLOCK, &sset, NULL);
  sigaction(SIGIO, &saction, NULL);
}
