//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup Tool "FUNSSEL" developed for the 
// Simulator League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public 
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#include "ICFile.h"
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <zlib.h>
#include <sys/stat.h>

//#define DEBUG

// cwz-build-command: make FUNSSEL

class IOError: public Exception {
public:
  IOError(const string &error, const string &f) :
    Exception("file " + f + ": " + error) {}
};

class IndexError: public Exception {
public:
  IndexError() : Exception("cannot open index") {}
};

ICFile::ICFile(const string &_name) throw (Exception)
  : file(NULL)
{
  open(_name);
}

ICFile::~ICFile()
{
  if (file)
    fclose((FILE *)file);
}

bool ICFile::open(const string &_name, bool write) throw (Exception)
{
  if (file)
    fclose((FILE *)file);
  record_starts.erase(record_starts.begin(), record_starts.end());

  if (_name != "")
    name  =_name;
  if (write) {
    file = fopen(name.c_str(), "w+");
    if (!file)
      throw IOError("cannot open", name);
    wopen = true;
    return true;
  }

  // open for read only
  wopen = false;
  // if file exists, create index base
  struct stat dummy;
  if (stat(name.c_str(), &dummy) == 0) {
    file = fopen(name.c_str(), "r+");
    try {
      setupIndexList();
      return true;
    }
    catch (Exception e) {
      // oops, this file is not for us...
      fclose((FILE *)file);
      file = NULL;
      record_starts.erase(record_starts.begin(), record_starts.end());
    }
  }
  return false;
}

void ICFile::setupIndexList() throw (Exception)
{
  record_starts.erase(record_starts.begin(), record_starts.end());

  fprintf(stderr, "indexing %s (may take a while)\n", name.c_str());
  fprintf(stderr, "         0 bytes done");
  fflush(stderr);
  int start = 0;
  int count = 0;
  while (!feof((FILE *)file)) {
    if (count++ % 10 == 0)
      fprintf(stderr, "\r%10d", start);

    record_starts.push_back(start);
    string dummy;

    fseek((FILE *)file, start, SEEK_SET);
    // may read more than needed
    int len = readNoSeek(dummy);
    if (len == -1) {
      // this entry was not readable
      record_starts.erase(record_starts.end() - 1);
      return;
    }
    start  += len;
  }
  fputc('\n', stderr);
}

void ICFile::push_back(const string &data) throw (Exception)
{
#ifdef DEBUG
  cout << "packing..." << endl;
#endif  
  if (!wopen)
    throw IOError("not open for writing", name);

  unsigned char buf[256];
  z_stream z;
  z.next_in   = (Bytef *)data.data();
  z.avail_in  = data.length();
  z.next_out  = buf;
  z.avail_out = 256;
  z.zalloc    = Z_NULL;
  z.zfree     = Z_NULL;
  z.data_type = Z_BINARY;

  fseek((FILE *)file, 0, SEEK_END);

  int pos = ftell((FILE *)file);

  int state = deflateInit(&z, Z_DEFAULT_COMPRESSION);
  while (state == Z_OK) {
    state = deflate(&z, Z_FINISH);
    if ((state != Z_OK) && (state != Z_STREAM_END))
      break;
    fwrite(buf, sizeof(char), z.next_out - buf, (FILE *)file);
    z.next_out  = buf;
    z.avail_out = 256;
  }
  deflateEnd(&z);
  if (state != Z_STREAM_END)
    throw IOError("could not write", name);

  record_starts.push_back(pos);
#ifdef DEBUG
  cout << "finished..." << endl;
#endif  
}

string ICFile::read(int record) throw (Exception)
{
  if (record < 0)
    throw IndexError();
  if (record >= (int) record_starts.size())
    throw IndexError();
  fseek((FILE *)file, record_starts[record], SEEK_SET);
  string res;
  if (readNoSeek(res) == -1)
    throw IOError("cannot read", name);
  return res;
}

int ICFile::readNoSeek(string &result)
{
  int length = 0;
  unsigned char buf_out[256];
  unsigned char buf_in[256];
  z_stream z;

  fread(buf_in, sizeof(char), 256, (FILE *)file);
  z.next_in   = buf_in;
  z.avail_in  = 256;
  z.next_out  = buf_out;
  z.avail_out = 256;
  z.zalloc    = Z_NULL;
  z.zfree     = Z_NULL;

  int state = inflateInit(&z);
  while (state == Z_OK) {
    if (z.avail_in == 0) {
      length    += 256;
      if (feof((FILE *)file))
	return -1;
      fread(buf_in, sizeof(char), 256, (FILE *)file);
      z.next_in  = buf_in;
      z.avail_in = 256;
    }
    state = inflate(&z, Z_SYNC_FLUSH);
    if ((state != Z_OK) && (state != Z_STREAM_END))
      break;
    result.append((char *)buf_out, string::size_type(z.next_out - buf_out));
    z.next_out  = buf_out;
    z.avail_out = 256;
  }
  length += z.next_in - buf_in;
  inflateEnd(&z);
  if (state != Z_STREAM_END)
    return -1;
  return length;
}
