//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#ifndef GRID_H
#define GRID_H
#include "Ensure.h"

/*
  This file implements a grid template. There is no .C file!
  To avoid compiler warnings the interface also has methods
  with int, but int is always typecasted to unsigned int internally.
 */

// To create an object of this template use e.g.:
// Grid<Grade> gradeGrid(30, 200); to create a grid of Grades
// Grid<Vector2d> vecGrid(120, 200); to create a gird of Vector2ds
// Be carefull! 
// Grid<Whatever> g(-1,-1) is a huge Grid with about
// sizeof(Whatever*)*2^((sizeof(int)+1)*8) bytes +
// sizeof(Whatever) bytes for each set entry

template<class type>
class Grid {
 public:

  // create a gird of size (height, width)
  // height and width are typecasted to unsigend int
  // the gird is initialized with NULL pointers! 
  // so don't get(i, j)-> without a check
  Grid( const int height, 
	const int width   );

  Grid( const unsigned int height, 
	const unsigned int width   );

  Grid( const Grid<type> &grid );

  ~Grid( void );

  // get a pointer to the element at (i, j)
  // can be NULL !!!
  // 0 <= i < height, 0 <= j < width
  const type* get( const unsigned int i, 
		   const unsigned int j  ) const;
  const type* get( const int i, 
		   const int j  ) const
  { return get((unsigned int)i, (unsigned int)j); };

  enum GridErr {
    GRID_OK,   // everything is OK 
    GRID_OOB,  // index is out of bounds
  };

  // set the element at (i, j) to a copy of elt
  // if elt == NULL the entry at (i, j) is deleted and set to NULL
  // 0 <= i < height, 0 <= j < width
  GridErr set( const unsigned int i,
	       const unsigned int j,
	       const type *elt       );
  GridErr set( const int i,
	       const int j,
	       const type *elt )
  { return set((unsigned int)i, (unsigned int)j, elt); };

  Grid<type>& operator=( Grid<type> &grid );

 private:
  type ***grid;
  unsigned int height;
  unsigned int width;

  // create a grid of size (height, width) in memory
  // does not remove an old one!
  void makeGrid( unsigned int height,
		 unsigned int width   );
  // del the grid
  void delGrid( unsigned int height,
		unsigned int width   );
};

/////////////////////////////////////////////////////////////////////
//
//                      Implementation
//
/////////////////////////////////////////////////////////////////////

template<class type>
Grid<type>::Grid( const int h, 
		  const int w  )
  : grid(NULL),
    height((unsigned int)h),
    width((unsigned int)w)
{
  makeGrid(height, width);
}

template<class type>
Grid<type>::Grid( const unsigned int h, 
		  const unsigned int w  )
  : grid(NULL),
    height(h),
    width(w)
{
  makeGrid(height, width);
}

template<class type>
Grid<type>::Grid( const Grid<type> &g )
  : grid(NULL),
    height(g.height),
    width(g.width)
{
  makeGrid(height, width);
  for (unsigned int i = 0; i < height; i++)
    for (unsigned int j = 0; j < width; j++)
      if (g.grid[i][j])
	grid[i][j] = new type(*g.grid[i][j]);
}

template<class type>
Grid<type>::~Grid( void )
{
  delGrid(height, width);
}

template<class type>
void
Grid<type>::makeGrid( unsigned int h,
		      unsigned int w  )
{
  grid =  new (type**)[h]; // create the first column
  for (unsigned int i = 0; i < h; i++)
    grid[i] = new (type*)[w]; // create the rows
  
  // initalize the grid
  for (unsigned int i = 0; i < h; i++)
    for (unsigned int j = 0; j < w; j++)
      grid[i][j] = NULL;
}

template<class type>
void
Grid<type>::delGrid( unsigned int h,
		     unsigned int w  )
{
  for (unsigned int i = 0; i < h; i++) {
    for (unsigned int j = 0; j < w; j++) 
      delete grid[i][j];  // delete the elements
    delete [] grid[i];  // delete one row
  }
  delete [] grid; // delete the rest
  grid = NULL;
}

template<class type>
inline
const type*
Grid<type>::get(const unsigned int i, 
		const unsigned int j  ) const
{
  ensure(i < height && j < width);
  if (i < height && j < width)
    return grid[i][j];

  return NULL;
}

template<class type>
Grid<type>::GridErr
Grid<type>::set(const unsigned int i, 
		const unsigned int j,
		const type *elt       )
{
  ensure(i < height && j < width);
  if (i < height && j < width) {
    if (grid[i][j]) delete grid[i][j];
    if (elt) 
      grid[i][j] = new type(*elt);
    else
      grid[i][j] = NULL;
    return GRID_OK;
  }
  return GRID_OOB;
}

template<class type>
Grid<type>&
Grid<type>::operator=( Grid<type> &g )
{
  delGrid(height, width); // delete the old grid

  // create the new one
  width = g.width;
  height = g.height;

  makeGrid(height, width);

  // copy the elements
  for (unsigned int i = 0; i < height; i++)
    for (unsigned int j = 0; j < width; j++)
      if (g.grid[i][j])
	grid[i][j] = new type(*g.grid[i][j]);

  return g;
}
#endif /* GRID_H */














