//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////// 
//   $Id: snapshot.C,v 1.86 2001/08/06 00:41:22 mrb Exp $
//   $Manager: cm$
//
//   snapshot.C
//
////////////////////////////////////////////////////////////////////////


#include "unistd.h"
#include "math.h"
#include "mytrace.h"
#include "snapshot.h"
#include "Ensure.h"
#include "audiomsg.h"
#include "Polygon.h"


// output when calculating ball position from visual data
//#define BALLPOSTRACE(output) cout << output

// output in collision detection algorithm
//#define COLLDETECTTRACE(output) cout << output

// output generated from this command can be plotted by gplot <file> <time>.
// It shows the update of our position as determined by the flags seen.
//#define FLAGTRACE(output) cout << output


#ifndef BALLPOSTRACE
#define BALLPOSTRACE(output)
#else
#warning BALLPOSTRACE debug trace BALLPOSTRACE is activated!
#endif

#ifndef COLLDETECTTRACE
#define COLLDETECTTRACE(output)
#else
#warning COLLDETECTTRACE debug trace COLLDETECTTRACE is activated!
#endif

#ifndef FLAGTRACE
#define FLAGTRACE(output)
#else
#warning FLAGTRACE debug trace FLAGTRACE is activated!
#endif


//////////////////////////////////////////////////////////////////////
//
// utility functions
//
//////////////////////////////////////////////////////////////////////

bool inField( const Vector2d &pos, double security )
{
  if( fabs( pos.getX() ) > FIELD_LENGTH/2 - security )
    return false;
  if( fabs( pos.getY() ) > FIELD_WIDTH/2 - security )
    return false;
  return true;
}


bool inOurPenaltyArea( const Vector2d &pos, double security )
{
  return ( ( pos.getX() > - 0.5 * FIELD_LENGTH + security ) &&
	   ( pos.getX() < - 0.5 * FIELD_LENGTH + PENALTY_AREA_LENGTH - security ) &&
	   ( fabs( pos.getY() ) < 0.5 * PENALTY_AREA_WIDTH - security ) );
}


bool inOtherPenaltyArea( const Vector2d &pos, double security )
{
  return ( ( pos.getX() < 0.5 * FIELD_LENGTH - security ) &&
	   ( pos.getX() > 0.5 * FIELD_LENGTH - PENALTY_AREA_LENGTH + security ) &&
	   ( fabs( pos.getY() ) < 0.5 * PENALTY_AREA_WIDTH - security ) );
}


//////////////////////////////////////////////////////////////////////
//
// utility functions for quantization and de-quantization
//
//////////////////////////////////////////////////////////////////////

inline double
quantize( double val, double qstep )
  // simulates quantization of an exact variable
{
  return rint( val / qstep ) * qstep;
}


inline double
nodaDist( double dist, double qstep )
  // simulates quantization of an exact distance according to Noda
  // return value is equivalent to return value from server
{
  return quantize( exp( quantize(log( dist + ( 1.0e-10 ) ), qstep ) ), 0.1 );
}


inline int
nodaDir( Angle dir )
  // simulates quantization of an exact direction according to Noda
  // return value is equivalent to return value from server
{
  return DEG_TO_RAD( rint( RAD_TO_DEG( dir ) ) );
}


inline Angle
leftDirLimit( Angle nodaDir )
  // returns minimum Angle that would still be quantized to nodaDir
{
  // this "rint" has nothing to do with quantization but with the fact
  // that soccerserver only sends integers and so "nodaDir" must be
  // integer when transformed back to DEG
  return DEG_TO_RAD( rint( RAD_TO_DEG( nodaDir ) ) - 0.5 );
}


inline Angle
rightDirLimit( Angle nodaDir )
  // returns maximum Angle that would still be quantized to nodaDir
{
  // this "rint" has nothing to do with quantization but with the fact
  // that soccerserver only sends integers and so "nodaDir" must be
  // integer when transformed back to DEG
  return DEG_TO_RAD( rint( RAD_TO_DEG( nodaDir ) ) + 0.5 );
}


double
upperDistLimit( double nodaDist, double qstep )
  // returns maximum distance that would still be quantized to nodaDist
{  
  // From the outer quantization, adding 0.05 to nodaDist gives the absolute 
  // maximum for the possible values of the exp(...) expression. 
  
  double d = nodaDist + 0.05;
  
  // Invert exp()

  if( d > 0 )
    d = log( d );
  else
    {
      cerr << "NODA'S ERROR: illigal quantization result from server !" << endl;
      return 1.0;   // senseless value (but d < 0 is senseless as well)
    }

  // As exp is monotonous, log(nodaDist+0.05) is the absolute upper limit
  // for the possible values of the inner quantization. But as these values
  // must also have precision qstep, we can restrict d further to the lower of 
  // the two quantization limits of the inner quantization.
  
  d = floor( d / qstep ) * qstep;

  // Now invert the inner quantization to upper limit.

  d = d + 0.5 * qstep;

  // Finally invert log.

  return exp( d );
}


double
lowerDistLimit( double nodaDist, double qstep )
  // returns minimum distance that would still be quantized to nodaDist
{
  // Like above, only choose lower limits when inverting quantizations,  
  // and pick the upper limit when restricting the inner quantizations value. 

  double d = nodaDist - 0.05;

  // distances are positive
  if( d <= EPS )
    return 0.0;

  // Invert exp()

  d = log(d);

  d = ceil( d / qstep ) * qstep;

  d = d - 0.5 * qstep;

  return exp( d );
}


//////////////////////////////////////////////////////////////////////
//
// Class WorldObject
//
//////////////////////////////////////////////////////////////////////


bool
WorldObject::inCorridor( Vector2d source, Vector2d target, double d, Angle alpha ) const
{
  double sx = source.getX();
  double sy = source.getY();

  double tx = target.getX();
  double ty = target.getY();

  double ox = pos.getX();
  double oy = pos.getY();

  // check if object is in circle of diameter d around target
  if( distTo( target ) <= 0.5 * d )
    return true;

  // check if object is in very small circle around source
  // so that inCorridor is consistent if source is (e.g.) the calling player's position
  if( distTo( source ) <= 1e-4 )
    return true;

  // project object onto the line source->target
  double proj = ( ( (sx-tx) * (sx-ox) + (sy-ty) * (sy-oy) ) /
		  ( (sx-tx) * (sx-tx) + (sy-ty) * (sy-ty) ) );
  
  if ( proj < 0 || proj > 1 )
    // projection point is outside the line
    return false;

  // calculate distance between object and the line source->target
  double dist = sqr( ox - sx - proj * (tx-sx) ) + sqr( oy - sy - proj * (ty-sy) );

  if ( dist > 0.25 * d * d )
    // object is too far away from line
    return false;

  if ( fabs( angleBetween( target - source, pos - source ) ) > alpha )
    // object is not in cone with angle 2 * alpha
    return false;

  return true;
}


bool
WorldObject::inField( double security ) const
{
  if( fabs( pos.getX() ) > FIELD_LENGTH/2 - security )
    return false;
  if( fabs( pos.getY() ) > FIELD_WIDTH/2 - security )
    return false;
  return true;
}


bool
WorldObject::inOurPenaltyArea( double security ) const
{
  return ( ( pos.getX() > - 0.5 * FIELD_LENGTH + security ) &&
	   ( pos.getX() < - 0.5 * FIELD_LENGTH + PENALTY_AREA_LENGTH - security ) &&
	   ( fabs( pos.getY() ) < 0.5 * PENALTY_AREA_WIDTH - security ) );
}


bool
WorldObject::inOtherPenaltyArea( double security ) const
{
  return ( ( pos.getX() < 0.5 * FIELD_LENGTH - security ) &&
	   ( pos.getX() > 0.5 * FIELD_LENGTH - PENALTY_AREA_LENGTH + security ) &&
	   ( fabs( pos.getY() ) < 0.5 * PENALTY_AREA_WIDTH - security ) );
}


//////////////////////////////////////////////////////////////////////
//
// Class WorldMovedObject
//
//////////////////////////////////////////////////////////////////////


void
WorldMovedObject::calcVelocity( const WorldMovedObject& relObject, double distChange, double dirChange )
{
  Vector2d relpos = pos - relObject.pos;

  v = ( distChange / norm( relpos ) ) * relpos;
  relpos.perp();
  acc( dirChange * relpos + relObject.v );
}


void
WorldMovedObject::setPos( const SensorObject& object, const WorldPlayer& my )
{
  pos = my.pos + radAng2Vector2d( object.dist, object.dir + my.faceDir );
}


//////////////////////////////////////////////////////////////////////
//
// Class WorldPlayer
//
//////////////////////////////////////////////////////////////////////


bool
WorldPlayer::ballCatchable( const WorldBall& ball, const Angle &catchAngle ) const
{
  // check if player is in penalty area
  if ( !inOurPenaltyArea() )
    return false;
  
  // relative position of the ball to the player
  Vector2d relPos = radAng2Vector2d( distTo( ball ), bodyAngleTo( ball ) - catchAngle );

  // check if ball is in catchable area
  if( fabs( relPos.getY() ) <= 0.5 * CATCHABLE_AREA_W )
    if( relPos.getX() >= 0.0 && relPos.getX() <= CATCHABLE_AREA_L )
      return true;

  return false;
}


double
WorldPlayer::kickPowerRate( const WorldBall& ball ) const
{
  // Here I use M_PI instead of PI because it is no Vector2d calculation
  return KICK_POWER_RATE * ( 1 - 0.25 * fabs( bodyAngleTo( ball ) ) / M_PI -
				 0.25 * ( distTo( ball ) - BALL_SIZE - size ) / 
				 kickableMargin);
}


void
WorldPlayer::setBySensorData( SensorPlayer player, const WorldPlayer& my )
{
  // approximate player's distance if not valid
  if( !IsDistValid( player.dist ) )
    player.dist = dist( my.pos, pos );
  setPos( player, my );
  age = 0;
  coneTime = 0;
  if( player.goalie != GOALIE_UNKNOWN )
    isGoalie = player.goalie;
  if( IsDirValid( player.body_dir ) )
    {
      bodyDir = player.body_dir + my.faceDir;
      if( IsDirValid( player.face_dir ) )
	faceDir = player.face_dir;
    }
  if( IsDistValid( player.dist_change ) && IsDirValid( player.dir_change ) )
    calcVelocity( my, player.dist_change, player.dir_change );
}


//////////////////////////////////////////////////////////////////////
//
// Class WorldBall
//
//////////////////////////////////////////////////////////////////////

void
WorldBall::calcPos( const SensorObject& object, const WorldPlayer& my )
  // keep simulated position if it is compatible with received values
  // otherwise: calculate position of ball using received values
  
{
  // step 0: setting up computation variables

  // dir we received from the server
  // (value was quantized by server before sending it to us)
  Angle received_dir_quantized  = object.dir;

  // dir we simulated, quantized the way the server would quantize it
  Angle sim_dir                 = my.faceAngleTo( *this );
  Angle sim_dir_quantized       = nodaDir( sim_dir );

  // dist we received from the server, was quantized before
  double received_dist_quantized = object.dist;

  // dist we simulated, quantized the way the server would quantize it
  double sim_dist               = my.distTo( *this );
  double sim_dist_quantized     = nodaDist( sim_dist, QUANTIZE_STEP );
  
  // dir to use when received dir is not compatible with sim dir,
  // depends on the direction (left/right) of the detected dir error
  Angle new_dir;

  // dist to use when received dist is not compatible with sim dist,
  // depends on the direction (closer/further) of the detected dir error
  double new_dist;
  
  // flag to show if a new value differs from its simulated value
  bool changed = false;
  

  // step 1: check direction

  Angle left  = leftDirLimit( received_dir_quantized );
  Angle right = rightDirLimit( received_dir_quantized );

  BALLPOSTRACE( "sim ball dir: " << RAD_TO_DEG( sim_dir )
		<< " left limit: " << RAD_TO_DEG( left )
		<< " right limit: " << RAD_TO_DEG( right )
		<< " ===> " );
  
  if( fabs( received_dir_quantized - sim_dir_quantized ) < EPS )
    {
      new_dir = sim_dir;
      BALLPOSTRACE( "sim dir is ok." << endl );
    }
  else
    {
      if( received_dir_quantized < sim_dir_quantized )
	{
	  // simulation was too far right
	  new_dir = right - Angle(EPS); // use received dir tile right edge
	}
      else
	{
	  // simulation was too far left
	  new_dir = left + Angle(EPS); // use received dir tile left edge
	}
      
      changed = true;
      BALLPOSTRACE( "sim dir was wrong, changed it to " << RAD_TO_DEG(new_dir) << "." << endl );
    }
  
  
  // step 2: check distance
 
  double lower = lowerDistLimit( received_dist_quantized, QUANTIZE_STEP );
  double upper = upperDistLimit( received_dist_quantized, QUANTIZE_STEP );

  BALLPOSTRACE( "sim ball dist: " << sim_dist << " lower limit: " << lower
		<< " upper limit: " << upper << " ===> " );
  
  if( fabs( received_dist_quantized - sim_dist_quantized ) < EPS )
    {
      new_dist = sim_dist;
      BALLPOSTRACE( "sim dist is ok." << endl );
    }
  else
    {
      if( received_dist_quantized < sim_dist_quantized )
	{
	  // simulation was too far away
	  new_dist = upper - EPS; // use received dir tile outer edge
	}
      else
	{
	  // simulation was too close
	  new_dist = lower + EPS; // use received dir tile closer edge
	}
      
      changed = true;
      BALLPOSTRACE( "sim dist was wrong, changed it to " << new_dist << endl );
    }
  
  // step 3: set ball pos
  
  if( changed )
    {
      pos = my.pos + radAng2Vector2d( new_dist, new_dir + my.faceDir );
    }  
}


//////////////////////////////////////////////////////////////////////
//
// Class WorldSense
//
//////////////////////////////////////////////////////////////////////

Angle 
WorldSense::getViewAngle( const view_width viewWidth )
{
  switch( viewWidth )
    {
    case VIEW_WIDTH_NARROW : return DEG_TO_RAD( 22.5 );
    case VIEW_WIDTH_NORMAL : return DEG_TO_RAD( 45.0 );
    case VIEW_WIDTH_WIDE   : return DEG_TO_RAD( 90.0 );
    }

  return Angle( 0.0 );
}  


Angle
WorldSense::getViewAngle() const
{
  return getViewAngle( viewWidth );
}

//////////////////////////////////////////////////////////////////////
//
// class WorldPlayerType
//
//////////////////////////////////////////////////////////////////////


class ostream & operator<<( class ostream &os, const WorldPlayerType &pt)
{
  return os << "(PlayerType " << pt.state
	    << " ((speed_max " << pt.playerSpeedMax 
	    << ") (player_decay " << pt.playerDecay
	    << ") (inertia_moment " << pt.inertiaFactor
	    << ") (dash_power_rate " << pt.dashPowerRate
	    << ")) (player_size " << pt.playerSize
	    << ") ((kickable_margin " << pt.kickableMargin
	    << ") (kick_rand " << pt.kickRand
	    << ")) ((stamina_inc_max " << pt.staminaIncMax
	    << ") (extra_stamina " << pt.extraStamina
	    << ") (effort_max " << pt.effortMax
	    << ") (effort_min " << pt.effortMin
	    << ")))";    
}

//////////////////////////////////////////////////////////////////////
//
// Class Snapshot
//
//////////////////////////////////////////////////////////////////////

Snapshot::Snapshot()
{
  time = 0;
  visualTime = 0;
  lastCatchTime = 0;
  
  collState = NONE;

  ourGoalieNo = UNKNOWN;     // goalie numbers are unknown
  otherGoalieNo = UNKNOWN;

  ourSide  = SIDE_LEFT;          // there is no SIDE_UNKNOWN
  playMode = PLAY_MODE_UNKNOWN;
  ourScore = otherScore = 0;

  // set goal positions (remember: coordinate system is always the same!)
  otherGoal.pos = GOALPOS( GOAL_RIGHT );
  ourGoal.pos = GOALPOS( GOAL_LEFT );
}

void
Snapshot::setSeeInvalid()
{
  int i;

  my.age = ball.age = ourGoal.age = otherGoal.age = ETERNAL;
  for( i = 0; i < 11; i++ )
    ourTeamArray[i].age = otherTeamArray[i].age = ETERNAL;
}


void
Snapshot::setSenseInvalid()
{
  senseBody.age = ETERNAL;  
}


void
Snapshot::setInvalid()
{
  setSenseInvalid();
  setSeeInvalid();
}


bool 
Snapshot::calcMyPos( SensorData &data )
{
  int   numLinesFound = 0;   // number of lines player has seen
  int   numFlagsFound = 0;   // number of flags player has seen

  // step 1 : count valid flags and save the two closest ones

  int closest = -1;
  double closestDist = 10000;
  int second=-1;
  double secondDist = 10000;

  for( int i = 0; i< NUM_FLAGS; i++ )
    {
      if( data.flag[i].valid && data.flag[i].which != FLAG_UNKNOWN 
	  && IsDistValid( data.flag[i].dist ) )
	{
	  numFlagsFound++;
	  if( data.flag[i].dist < secondDist )
	    {
	      second = i;
	      secondDist = data.flag[i].dist;
	    }
	  if( data.flag[i].dist < closestDist )
	    {
	      second = closest;
	      secondDist = closestDist;
	      closest = i;
	      closestDist = data.flag[i].dist;
	    }
	}
    }

  if( numFlagsFound == 0 )
    {
      FLAGTRACE( "Cannot determine any flag!" << endl );
      return false;
    }

  // step 2 : try to calculate face direction with one line

  for( int i = 0; i < 3; i++ )
    {
      if( data.line[i].valid && data.line[i].which != LINE_UNKNOWN )
	{
	  numLinesFound++;
	  
	  if( numLinesFound > 1 )   // this means player is outside the field
	    continue;
	  
	  Angle alpha = data.line[i].dir;

	  switch( data.line[i].which )
	    {
	    case LINE_RIGHT:
	      alpha = alpha > 0 ? PI_2 - alpha : -PI_2 - alpha;
	      break;
	      
	    case LINE_BOTTOM:
	      alpha = alpha > 0 ? PI - alpha : -alpha;
	      break;
	      
	    case LINE_LEFT:
	      alpha = alpha > 0 ? -PI_2 - alpha : PI_2 - alpha;
	      break;
	      
	    case LINE_TOP:
	      alpha = alpha > 0 ? -alpha : -PI - alpha;
	      break;
	      
	    default:
	      ensure( false );
	    }

	  Angle l( leftDirLimit( alpha ) );
	  Angle r( rightDirLimit( alpha ) );
	  alpha = bisect( l, r );
	  my.faceDir = alpha;
	}
    }

  // step 3 : calculate face direction, if it could not be calculated with one line

  if( numLinesFound != 1 )
    {
      if( numFlagsFound == 1 )
	{
	  FLAGTRACE( "Only determined one flag!" << endl );
	  return false;
	}

      Vector2d diffAbsPos = FLAGPOS( data.flag[closest].which ) - FLAGPOS( data.flag[second].which );
      
      Vector2d diffRelPos = radAng2Vector2d( data.flag[closest].dist, data.flag[closest].dir )
	- radAng2Vector2d( data.flag[second].dist, data.flag[second].dir );
      
      my.faceDir = diffAbsPos.getAngle() - diffRelPos.getAngle();
    }

  // step 4 : calculate absolute position with polygon intersection
  //          if intersection fails use old method (weighted use of flag polygons)

  Polygon posPoly;                      // intersection polygon of all flag polygons
  bool firstFlag = true;                // is the current flag the first valid flag?

  Vector2d myWeightedPos( 0.0, 0.0 );   // my position calculated with old method (if intersecting fails)
  double weightsum = 0.0;               // weightsum for old calculation method

  for( int i = 0; i < NUM_FLAGS; i++ )
    {
      if( data.flag[i].valid && data.flag[i].which != FLAG_UNKNOWN
	  && IsDistValid( data.flag[i].dist ) )
	{
	  Polygon flagPoly;   // position polygon for current flag

	  Vector2d flagpos = FLAGPOS( data.flag[i].which );   // get flag position
	  
	  // de-quantized absolute area of flag position
	  double top = upperDistLimit( data.flag[i].dist, QUANTIZE_STEP_L );
	  double bot = lowerDistLimit( data.flag[i].dist, QUANTIZE_STEP_L );

	  // assume maximal error of 1.5 deg
	  Angle l( leftDirLimit( data.flag[i].dir ) + my.faceDir - DEG_TO_RAD( 1.5 ) );
	  Angle r( rightDirLimit( data.flag[i].dir ) + my.faceDir + DEG_TO_RAD( 1.5 ) );
	  
	  // limitations of our own position, as determined by flag
	  Vector2d lowleft = flagpos - radAng2Vector2d( top, r );
	  Vector2d lowmiddle = flagpos - radAng2Vector2d( top, bisect( l, r ) );
	  Vector2d lowright = flagpos - radAng2Vector2d( top, l );
	  Vector2d hileft = flagpos - radAng2Vector2d( bot, r );
	  Vector2d hiright = flagpos - radAng2Vector2d( bot, l );
	  
	  // flag polygon coordinates trace
 	  FLAGTRACE( data.time << " "  << lowright.getX() << " " << lowright.getY() << endl );
 	  FLAGTRACE( data.time << " " << hiright.getX() << " " << hiright.getY() << endl );
 	  FLAGTRACE( data.time << " " << hileft.getX() << " " << hileft.getY() << endl );
 	  FLAGTRACE( data.time << " " << lowleft.getX() << " " << lowleft.getY() << endl ); 
 	  FLAGTRACE( data.time << " " << lowmiddle.getX() << " " << lowmiddle.getY() << endl );
 	  FLAGTRACE( data.time << " " << lowright.getX() << " " << lowright.getY() << endl << endl );

	  if( firstFlag )
	    {
	      // just fill posPoly (nothing to intersect yet)
	      posPoly.addPoint( lowleft );
	      posPoly.addPoint( hileft );
	      posPoly.addPoint( hiright );
	      posPoly.addPoint( lowright );
	      posPoly.addPoint( lowmiddle );

	      firstFlag = false;
	    }
	  else
	    {
	      // intersect posPoly with new flag polygon
	      flagPoly.addPoint( lowleft );
	      flagPoly.addPoint( hileft );
	      flagPoly.addPoint( hiright );
	      flagPoly.addPoint( lowright );
	      flagPoly.addPoint( lowmiddle );

	      posPoly ^= flagPoly;
	    }

	  // old position calculation method
	  double weight = 1.0 / ( 1.0 + data.flag[i].dist );
	  weightsum += weight;
	  myWeightedPos += ( flagpos - radAng2Vector2d( 0.5 * ( top + bot ), bisect( l, r ) ) ) * weight;
	}
    }
  
  // intersection polygon coordinates trace
  for( int k = 0; k < posPoly.nop(); k++ )
    FLAGTRACE( data.time << " " << posPoly.points()[k].getX()
	       << " " << posPoly.points()[k].getY() << endl );
  FLAGTRACE( endl );
  
  if( posPoly.nop() > 0 )
    {
      // new method (polygon intersection) worked
      my.pos = posPoly.getInside();

      // new method position trace
      FLAGTRACE( data.time << " " << my.pos.getX()-0.2 << " " << my.pos.getY()-0.2 << endl );
      FLAGTRACE( data.time << " " << my.pos.getX()+0.2 << " " << my.pos.getY()+0.2 << endl << endl );
      FLAGTRACE( data.time << " " << my.pos.getX()-0.2 << " " << my.pos.getY()+0.2 << endl );
      FLAGTRACE( data.time << " " << my.pos.getX()+0.2 << " " << my.pos.getY()-0.2 << endl << endl );
    }
  else
    {
      // new method (polygon intersection) failed, use weighted position
      my.pos = myWeightedPos * ( 1.0 / weightsum );

      // old method position trace
      FLAGTRACE( data.time << " " << my.pos.getX()-0.2 << " " << my.pos.getY() << endl );
      FLAGTRACE( data.time << " " << my.pos.getX()+0.2 << " " << my.pos.getY() << endl << endl );
      FLAGTRACE( data.time << " " << my.pos.getX() << " " << my.pos.getY()-0.2 << endl );
      FLAGTRACE( data.time << " " << my.pos.getX() << " " << my.pos.getY()+0.2 << endl << endl );
    }

  my.age = 0;
  my.state = VISUAL;
  
  // if we are playing from right to left turn coordinate system by 180 deg
  if( ourSide == SIDE_RIGHT )
    {
      my.pos.perp();
      my.pos.perp();
      my.faceDir += PI;
    }

  return true;
}


bool 
Snapshot::adjustPlayers( SensorData &data ) 
{
  int          i = 0, j = 0;

  int          maxAge = 0;         // for searching for oldest players in player arrays
  int          maxAgeIndex = 0;

  int          minAge = ETERNAL;   // for searching for minimal distance in player arrays
  int          minAgeIndex = 0;

  WorldPlayer  dummy;              // for calculating distances in search algorithm

  bool         marked[21];         // for marking players that have been processed
  for( i = 0; i < 21; i++ )
    marked[i] = false;

  // step 1 : update all players with visible number (simply overwrite)

  for( i = 0; i < 21; i++ )
    {
      SensorPlayer &player = data.player[i];
      if( player.valid && player.number > 0 && IsDistValid( player.dist ) )
	{
	  if( player.team == TEAM_US )
	    {
	      // set goalie flags if goalie is seen for the first time
	      if( player.goalie == GOALIE_TRUE && ourTeamArray[player.number-1].isGoalie != GOALIE_TRUE )
		{
		  for( j = 0; j < 11; j++ )
		    if( j != player.number - 1 )
		      ourTeamArray[j].isGoalie = GOALIE_FALSE;
		  ourGoalieNo = player.number;
		}
	      ourTeamArray[player.number-1].setBySensorData( player, my );
	      ourTeamArray[player.number-1].state = VISUAL;
	      marked[i] = true;
	    }
	  else if( player.team == TEAM_THEM )
	    {
	      // set goalie flags if goalie is seen for the first time
	      if( player.goalie == GOALIE_TRUE && otherTeamArray[player.number-1].isGoalie != GOALIE_TRUE )
		{
		  for( j = 0; j < 11; j++ )
		    if( j != player.number - 1 )
		      otherTeamArray[j].isGoalie = GOALIE_FALSE;
		  otherGoalieNo = player.number;
		}
	      otherTeamArray[player.number-1].setBySensorData( player, my );
	      otherTeamArray[player.number-1].state = VISUAL;
	      marked[i] = true;
	    }
	}
    }
  
  // step 2 : try to update "felt" players (inside VISIBLE_DISTANCE area):
  //          - search in both teams for players with small distance to new
  //            sensor information and replace the youngest among them
  //          - if no player is close enough, overwrite oldest information
  //            in the OPPONENT team

  for( i = 0; i < 21; i++ )
    {
      SensorPlayer &player = data.player[i];
      if( !marked[i] && player.valid && player.number <= 0 && player.team == TEAM_UNKNOWN )
	{
	  if( IsDistValid( player.dist ) && player.dist <= VISIBLE_DISTANCE )
	    {
	      int minAgeMyTeam = ETERNAL;
	      int minAgeIndexMyTeam = 0;

	      int minAgeOtherTeam = ETERNAL;
	      int minAgeIndexOtherTeam = 0;

	      bool putInOtherTeam = true;
	      maxAge = 0;
	      dummy.setPos( player, my );

	      // search in both teams
	      for( j = 0; j < 11; j++ )
		{
		  // search in our team
		  // do not overwrite player himself and players that have already been updated
		  // do not overwrite goalie with players that are too far away from our goal
		  if( ( j != my.number - 1 ) && ( ourTeamArray[j].age > 0 ) &&
		      ( j != ourGoalieNo - 1 || dummy.pos.getX() < -FIELD_LENGTH / 3.0 ) )
		    {
		      // look for youngest player among close players
		      if( ourTeamArray[j].distTo( dummy ) < 2.0 + ourTeamArray[j].vmax * min( ourTeamArray[j].age, 3 ) ||
			  my.distTo( ourTeamArray[j] ) <= VISIBLE_DISTANCE + my.vmax )
			{
			  if( ourTeamArray[j].age < minAgeMyTeam )
			    {
			      minAgeMyTeam = ourTeamArray[j].age;
			      minAgeIndexMyTeam = j;
			    }
			}
		    } 

		  // search in opponent team
		  // do not overwrite players that have already been updated
		  // do not overwrite goalie with players that are too far away from opponent goal
		  if( ( otherTeamArray[j].age > 0 ) &&
		      ( j != otherGoalieNo - 1 || dummy.pos.getX() > FIELD_LENGTH / 3.0 ) )
		    {
		      // look for oldest player
		      if( otherTeamArray[j].age > maxAge )
			{
			  maxAgeIndex = j;
			  maxAge = otherTeamArray[j].age;
			}
		      // look for youngest player among close players
		      if( otherTeamArray[j].distTo( dummy ) < 2.0 + otherTeamArray[j].vmax * min( otherTeamArray[j].age, 3 ) ||
			  my.distTo( otherTeamArray[j] ) <= VISIBLE_DISTANCE + my.vmax )
			{
			  if( otherTeamArray[j].age < minAgeOtherTeam )
			    {
			      minAgeOtherTeam = otherTeamArray[j].age;
			      minAgeIndexOtherTeam = j;
			    }
			}
		    } 
		}

	      // is there any close player in the opponent team?
	      if( minAgeOtherTeam < ETERNAL )
		{
		  maxAgeIndex = minAgeIndexOtherTeam;  // use this index for update
		  maxAge = minAgeOtherTeam;
		}
	      // if not - maybe in our team?
	      else if( minAgeMyTeam < ETERNAL )
		{
		  maxAgeIndex = minAgeIndexMyTeam;  // use this index for update
		  maxAge = minAgeMyTeam;
		  putInOtherTeam = false;
		}
	      // replace player (either oldest or "youngest but closest")
	      if( maxAge > 0 )
		{
		  // sort player into one of the teams
		  if( putInOtherTeam )
		    {
		      otherTeamArray[maxAgeIndex].setBySensorData( player, my );
		      if( minAge < ETERNAL && otherTeamArray[maxAgeIndex].state != UNSAFE )
			otherTeamArray[maxAgeIndex].state = FUZZY;
		      else
			otherTeamArray[maxAgeIndex].state = UNSAFE;
		      marked[i] = true;
		    }
		  else
		    {
		      ourTeamArray[maxAgeIndex].setBySensorData( player, my );
		      if( minAge < ETERNAL && ourTeamArray[maxAgeIndex].state != UNSAFE )
			ourTeamArray[maxAgeIndex].state = FUZZY;
		      else
			ourTeamArray[maxAgeIndex].state = UNSAFE;
		      marked[i] = true;
		    }
		}
	    }
	}
    }

  // step 3 : try to update players with visible color but no visible number
  //          by comparing them with the players in the snapshot :
  //          search for players with small distance to new sensor information
  //          and replace the youngest among them

  for( i = 0; i < 21; i++ )
    {
      SensorPlayer &player = data.player[i];
      if( !marked[i] && player.valid && player.number <= 0 && IsDistValid( player.dist ) )
	{
	  if( player.team == TEAM_US )
	    {
	      minAge = ETERNAL;
	      double minDist = 10000;
	      dummy.setPos( player, my );
	      
	      for( j = 0; j < 11; j++ )
		{
		  // do not overwrite player himself and players that have already been updated
		  // do not overwrite goalie with players that are too far away from our goal
		  if( ( j != my.number - 1 ) && ( ourTeamArray[j].age > 0 ) &&
		      ( j != ourGoalieNo - 1 || dummy.pos.getX() < -FIELD_LENGTH / 3.0 ) )
		    {
		      // maximal error of distance caused by Noda's quantization
		      double distError = ( upperDistLimit( player.dist, QUANTIZE_STEP ) -
					   lowerDistLimit( player.dist, QUANTIZE_STEP ) );
		      // look for youngest player among close players
		      if( ourTeamArray[j].distTo( dummy ) <
			  2.0 + ourTeamArray[j].vmax * min( ourTeamArray[j].age, 3 ) + distError )
			{
			  if( ourTeamArray[j].age < minAge - 1 ||
			      ( ourTeamArray[j].age <= minAge + 1 &&
				ourTeamArray[j].distTo( dummy ) < minDist ) )
			    {
			      minAge = ourTeamArray[j].age;
			      minAgeIndex = j;
			      minDist = ourTeamArray[j].distTo( dummy );
			    }
			}
		    } 
		}
	      // we have found one close player
	      if( minAge < ETERNAL )
		{
		  ourTeamArray[minAgeIndex].setBySensorData( player, my );
		  if( ourTeamArray[minAgeIndex].state != UNSAFE )
		    ourTeamArray[minAgeIndex].state = FUZZY;
		  else
		    ourTeamArray[minAgeIndex].state = UNSAFE;
		  marked[i] = true;
		}
	    }
	  else if( player.team == TEAM_THEM )
	    {
	      minAge = ETERNAL;
	      double minDist = 10000;
	      dummy.setPos( player, my );

	      for( j = 0; j < 11; j++ )
		{
		  // do not overwrite players that have already been updated
		  // do not overwrite goalie with players that are too far away from opponent goal
		  if( ( otherTeamArray[j].age > 0 ) &&
		      ( j != otherGoalieNo - 1 || dummy.pos.getX() > FIELD_LENGTH / 3.0 ) )
		    {
		      // maximal error of distance caused by Noda's quantization
		      double distError = ( upperDistLimit( player.dist, QUANTIZE_STEP ) -
					   lowerDistLimit( player.dist, QUANTIZE_STEP ) );
		      // look for youngest player among close players
		      if( otherTeamArray[j].distTo( dummy ) <
			  2.0 + otherTeamArray[j].vmax * min( otherTeamArray[j].age, 3 ) + distError )
			{
			  if( otherTeamArray[j].age < minAge - 1 ||
			      ( otherTeamArray[j].age <= minAge + 1 &&
				otherTeamArray[j].distTo( dummy ) < minDist ) )
			    {
			      minAge = otherTeamArray[j].age;
			      minAgeIndex = j;
			      minDist = otherTeamArray[j].distTo( dummy );
			    }
			}
		    } 
		}
	      // we have found one close player
	      if( minAge < ETERNAL )
		{
		  otherTeamArray[minAgeIndex].setBySensorData( player, my );
		  if( minAge < ETERNAL && otherTeamArray[minAgeIndex].state != UNSAFE )
		    otherTeamArray[minAgeIndex].state = FUZZY;
		  else
		  otherTeamArray[minAgeIndex].state = UNSAFE;
		  marked[i] = true;
		}
	    }
	}
    }

  // step 4 : try to update the remaining players with visible color but no visible number
  //          by comparing them with the players in the snapshot :
  //          search for player with smallest distance to new sensor information and replace
  //          him if he is old enough

  for( i = 0; i < 21; i++ )
    {
      SensorPlayer &player = data.player[i];
      if( !marked[i] && player.valid && player.number <= 0 && IsDistValid( player.dist ) )
	{
	  if( player.team == TEAM_US )
	    {
	      int minDistIndex = -1;
	      double minDist = 10000;
	      dummy.setPos( player, my );

	      for( j = 0; j < 11; j++ )
		{
		  // do not overwrite player himself and players that have already been updated
		  // do not overwrite goalie with players that are too far away from our goal
		  if( ( j != my.number - 1 ) && ( ourTeamArray[j].age > 0 ) &&
		      ( j != ourGoalieNo - 1 || dummy.pos.getX() < -FIELD_LENGTH / 3.0 ) )
		    {
		      // look for closest player
		      if( minDistIndex == -1 || ourTeamArray[j].distTo( dummy ) < minDist )
			if( ourTeamArray[j].age > 3 )
			  {
			    minDist = ourTeamArray[j].distTo( dummy );
			    minDistIndex = j;
			  }
		    }
		}
	      // we have found one close player
	      if( minDistIndex != -1 )
		{
		  ourTeamArray[minDistIndex].setBySensorData( player, my );
		  ourTeamArray[minDistIndex].state = UNSAFE;
		  marked[i] = true;
		}
	    }
	  else if( player.team == TEAM_THEM )
	    {
	      int minDistIndex = -1;
	      double minDist = 10000;
	      dummy.setPos( player, my );

	      for( j = 0; j < 11; j++ )
		{
		  // do not overwrite players that have already been updated
		  // do not overwrite goalie with players that are too far away from opponent goal
		  if( ( otherTeamArray[j].age > 0 ) &&
		      ( j != otherGoalieNo - 1 || dummy.pos.getX() > FIELD_LENGTH / 3.0 ) )
		    {
		      // look for closest player
		      if( minDistIndex == -1 || otherTeamArray[j].distTo( dummy ) < minDist )
			if( otherTeamArray[j].age > 3 )
			  {
			    minDist = otherTeamArray[j].distTo( dummy );
			    minDistIndex = j;
			  }
		    }
		}
	      // we have found one close player
	      if( minDistIndex != -1 )
		{
		  otherTeamArray[minDistIndex].setBySensorData( player, my );
		  otherTeamArray[minDistIndex].state = UNSAFE;
		  marked[i] = true;
		}
	    }
	}
    }

  // step 5 : - cleanup : set for players which are in view cone and not too
  //            far away but too old age to ETERNAL and state to UNSAFE
  //          - same for players that I should have felt
  
  for( i = 0; i < 11; i++ )
    {
      if( i + 1 != my.number && ourTeamArray[i].coneTime > 4 )
	{
	  ourTeamArray[i].age = ETERNAL;
	  ourTeamArray[i].state = UNSAFE;
	}
      // do not set opponent goalie's age to ETERNAL  
      if( i + 1 != otherGoalieNo && otherTeamArray[i].coneTime > 4 )
	{
	  otherTeamArray[i].age = ETERNAL;
	  otherTeamArray[i].state = UNSAFE;
	}
    }
  
  return true;  
}


bool 
Snapshot::adjustBall( SensorData &data, const Vector2d& oldBallRelpos )
{
  bool positionAdjusted = false;
  bool velocityAdjusted = false;

  // set velocity to zero in standard situations
  switch( playMode )
    {
    case PLAY_MODE_BEFORE_KICK_OFF :
    case PLAY_MODE_TIME_OVER :
    case PLAY_MODE_KICK_OFF_US :
    case PLAY_MODE_KICK_OFF_THEM :
    case PLAY_MODE_KICK_IN_US :
    case PLAY_MODE_KICK_IN_THEM :
    case PLAY_MODE_FREE_KICK_US :
    case PLAY_MODE_FREE_KICK_THEM :
    case PLAY_MODE_CORNER_KICK_US :
    case PLAY_MODE_CORNER_KICK_THEM :
    case PLAY_MODE_GOAL_US :
    case PLAY_MODE_GOAL_THEM :
    case PLAY_MODE_DROP_BALL :
    case PLAY_MODE_OFFSIDE_US :
    case PLAY_MODE_OFFSIDE_THEM :
    case PLAY_MODE_HALF_TIME :
    case PLAY_MODE_TIME_EXTENDED :
      ball.v.setXY( 0.0, 0.0 );
      velocityAdjusted = true;
      break;

    default :
      break;
    }

  // set ball position in standard situations
  switch( playMode )
    {
    case PLAY_MODE_BEFORE_KICK_OFF :
    case PLAY_MODE_KICK_OFF_US :
    case PLAY_MODE_KICK_OFF_THEM :
      ball.pos.setXY( 0.0, 0.0 );
      positionAdjusted = true;
      ball.state = VISUAL;   // this makes sense even if we have not seen the ball!
      break;

    case PLAY_MODE_CORNER_KICK_US :
      // of course, this only works when we have seen the ball just before
      ball.pos.setXY( 0.5 * FIELD_LENGTH - CKICK_MARGIN,
		      ball.pos.getY() < 0.0 ? -0.5 * FIELD_WIDTH + CKICK_MARGIN
		      : 0.5 * FIELD_WIDTH - CKICK_MARGIN );
      // it is better not to use the following line!
      // positionAdjusted = true;
      break;

    case PLAY_MODE_CORNER_KICK_THEM :
      // of course, this only works when we have seen the ball just before
      ball.pos.setXY( -0.5 * FIELD_LENGTH + CKICK_MARGIN,
		      ball.pos.getY() < 0.0 ? -0.5 * FIELD_WIDTH + CKICK_MARGIN
		      : 0.5 * FIELD_WIDTH - CKICK_MARGIN );
      // it is better not to use the following line!
      // positionAdjusted = true;
      break;

    case PLAY_MODE_KICK_IN_US :
    case PLAY_MODE_KICK_IN_THEM :
      // of course, this only works when we have seen the ball just before
      if( fabs( ball.pos.getY() ) > 0.5 * FIELD_WIDTH - 3.0 )
	ball.pos.setY( ball.pos.getY() < 0.0 ? -0.5 * FIELD_WIDTH : 0.5 * FIELD_WIDTH );
      // it is better not to use the following line!
      // positionAdjusted = true;
      break;

    default :
      break;
    }

  // if we have got visual information AND seen the ball, calculate its position
  // and velocity if this has not been done above
  if( data.type == SENSOR_TYPE_SEE && data.ball.valid )
    {
      // keep simulated ball (need that for collision check and position difference comparison)
      WorldBall simBall = ball;

      // we have seen the ball, so set its age to zero
      ball.age = 0;
      ball.coneTime = 0;

      // approximate dist if not valid; in any case ball is a little away, so it needs a closer look!
      // this is only needed for bad quality sensor data - but we never use that feature!
      if( !IsDistValid( data.ball.dist ) )
	data.ball.dist = my.distTo( ball ) < 20.0 ? 20.0 : my.distTo( ball );

      // update ball position
      if( !positionAdjusted )
	{
	  ball.calcPos( data.ball, my );   // update of my position should be done before!
	  positionAdjusted = true;
	}

      ball.state = VISUAL;

      // update ball velocity
      if( !velocityAdjusted )
	{
	  // check if dist_change and dir_change are valid, then update ball velocity
	  if( IsDistValid( data.ball.dist_change ) && IsDirValid( data.ball.dir_change ) )
	    {
	      ball.calcVelocity( my, data.ball.dist_change, data.ball.dir_change );
	      velocityAdjusted = true;
	    }
	}

      // check if there might have been a collision and set velocity to zero
      // this does not work in any case! (because of distance quantization)
      // (e.g. the possible values 0.385 ... 0.451 (for a player with standard
      // size 0.3) are all quantized to 0.4)
      if( !velocityAdjusted )
	{
	  if( data.ball.dist <= nodaDist( BALL_SIZE + my.size, QUANTIZE_STEP ) + 1e-4 )
	    {
	      ball.v.setXY( 0.0, 0.0 );
	      velocityAdjusted = true;
	      ball.state = FUZZY;   // because information might be incorrect
	      COLLDETECTTRACE( "COLL: There might have been a player<->ball collision!!!" << endl );
	    }
	}

      // or calculate velocity from position difference, but only if ball is young
      // enough and the position difference does not fit our simulated velocity
      if( !velocityAdjusted )
	{
	  switch( simBall.age )
	    {
	    case 1 :   // visual/audial info in the last cycle
	      {
		// problem: there might have been a collision with another player
		// in this case we can not use my.pos and my.v for the calculation
		bool maybePlayerCollision = false;

		// check maximum player size
		double maxPlayerSize = PLAYER_SIZE;
		for( int i = 0; i < playerParams.player_types; i++ )
		  if( playerType[i].state != UNKNOWN )
		    if( playerType[i].playerSize > maxPlayerSize )
		      maxPlayerSize = playerType[i].playerSize;

		for( int i = 0; i < 21; i++ )
		  {
		    SensorPlayer &player = data.player[i];
		    if( player.valid && IsDistValid( player.dist ) )
		      if( player.dist < my.size + maxPlayerSize + 0.1 )   // safety distance
			{
			  maybePlayerCollision = true;
			  break;
			}
		  }

		COLLDETECTTRACE( "COLL: Posdiff = " << simBall.distTo( ball ) << ", ball.age was 1" << endl );

		// in case of collision with other player (guessed) use old algorithm
		if( maybePlayerCollision )
		  {
		    COLLDETECTTRACE( "COLL: maybe collision with player, using old algorithm" << endl );
		    
		    simBall.pos -= simBall.v * ( 1.0 / BALL_DECAY );
		    ball.v = ( ball.pos - simBall.pos ) * BALL_DECAY;
		    velocityAdjusted = true;
		    ball.state = FUZZY;   // because information might be incorrect
		  }
		else
		  {
		    // calculate ball velocity from position difference
		    ball.v = ( ball.pos - ( my.pos - my.v * ( 1.0 / my.decay ) + oldBallRelpos ) ) * BALL_DECAY;
		    velocityAdjusted = true;
		  }
	      }
	      break;

	    case 2 :   // latest visual/audial info two cycles ago

	      if( simBall.distTo( ball ) > 0.6 )   // [C] (but ok)
		{
		  // The simulation in this case is not very reliable.
		  // The main assumption is that the view is WIDE (because we receive
		  // visual data every 2 cycles). In this case the absolute position
		  // of the player is not too bad.
		  COLLDETECTTRACE( "COLL: Posdiff does not match velocity, error = " );
		  COLLDETECTTRACE( simBall.distTo( ball ) << ", ball.age was 2" << endl );
			  
		  // ball position two timesteps ago
		  simBall.pos -= simBall.v * ( ( 1.0 + BALL_DECAY ) / sqr( BALL_DECAY ) );
			  
		  // calculate ball velocity from position difference
		  ball.v = ( ball.pos - simBall.pos ) * ( sqr( BALL_DECAY ) / ( 1.0 + BALL_DECAY ) );
		  velocityAdjusted = true;
		  ball.state = FUZZY;   // because information might be incorrect
		}
	      break;

	    default :  // simulated ball too old
	      ball.state = FUZZY;
	      break;
	    }
	}
    }

  // values for ball.state now:
  // -----------------------------------------------------------------------------------
  // case                                                                        state
  // -----------------------------------------------------------------------------------
  // ball vel. seen                                                              VISUAL
  // ball vel. not seen, maybe player<->ball collision                           FUZZY
  // ball vel. not seen, simBall's age == 1, sim. fits posdiff                   VISUAL
  // ball vel. not seen, simBall's age == 1, sim. does not fit posdiff           VISUAL
  // ball vel. not seen, simBall's age == 1, sim. does not fit posdiff,
  //      collision with other player(s) possible                                FUZZY
  // ball vel. not seen, simBall's age == 2, sim. fits posdiff                   VISUAL
  // ball vel. not seen, simBall's age == 2, sim. does not fit posdiff           FUZZY
  // ball vel. not seen, simBall's age > 2                                       FUZZY

  return ( positionAdjusted || velocityAdjusted );
}


bool
Snapshot::adjustGoals( SensorData &data )
{
  // set goal positions (remember: coordinate system is always the same!)
  otherGoal.pos = GOALPOS( GOAL_RIGHT );
  ourGoal.pos = GOALPOS( GOAL_LEFT );

  // use visual information only to set "age" (so that player knows when he
  // has seen the goals for the last time
  for( int i = 0; i < 2; i++ )
    if( data.goal[i].valid )
      if( data.goal[i].team == TEAM_THEM )
	otherGoal.age = 0;
      else if( data.goal[i].team == TEAM_US )
	ourGoal.age = 0;
  
  return true;
}


bool
Snapshot::adjustSense( SensorData &data ) 
{
  senseBody.stamina = data.stamina;
  senseBody.effort = data.effort;
  senseBody.v = radAng2Vector2d( data.speed, data.speed_angle );

  senseBody.headAngle       = data.head_angle;
  senseBody.kickCount       = data.kick_count;
  senseBody.dashCount       = data.dash_count;
  senseBody.turnCount       = data.turn_count;
  senseBody.sayCount        = data.say_count;
  senseBody.turnNeckCount   = data.turn_neck_count;
  senseBody.catchCount      = data.catch_count;      // new in 7.0 protocol
  senseBody.moveCount       = data.move_count;
  senseBody.changeViewCount = data.change_view_count;

  senseBody.viewWidth = data.my_view_width;
  senseBody.viewQuality = data.my_view_quality;

  senseBody.age = 0;

  return true;
}


bool   
Snapshot::adjustSee( SensorData &data, const Vector2d& oldBallRelpos )
{
  time = data.time;

  calcMyPos( data );

  adjustBall( data, oldBallRelpos );

  adjustPlayers( data );

  adjustGoals( data );
  
  return true;
}


bool   
Snapshot::adjustExact( SensorData &data )
{
  time = data.time;

  // adjust my position
  my.pos.setXY( 0.0, 0.0 );
  my.v.setXY( 0.0, 0.0 );
  my.age     = 0;
  my.decay   = 0.0;
  my.vmax    = 0.0;
  my.bodyDir = 0.0;
  my.faceDir = 0.0;
  my.number  = 0;
  my.isGoalie = GOALIE_FALSE;   // is coach !

  // adjust ball
  if( data.exactball.valid )
    {
      ball.age = 0;
      if (ourSide == SIDE_LEFT) {
	ball.pos.setXY( data.exactball.x, data.exactball.y );
	ball.v.setXY( data.exactball.x_vel, data.exactball.y_vel );
      } else {
	ball.pos.setXY( - data.exactball.x, - data.exactball.y );
	ball.v.setXY( - data.exactball.x_vel, - data.exactball.y_vel );
      }
    }

  // adjust players
  for( int i = 0; i < 22; i++ )
    {
      ExactPlayer &player = data.exactplayer[i];
      if( player.valid && player.number > 0 )   // then all information is complete, esp. the numbers
	{
	  if( player.team == TEAM_US )
	    {
	      // set goalie flags if goalie is seen for the first time
	      if( player.goalie == GOALIE_TRUE && ourTeamArray[player.number-1].isGoalie != GOALIE_TRUE )
		{
		  for( int j = 0; j < 11; j++ )
		    if( j != player.number - 1 )
		      ourTeamArray[j].isGoalie = GOALIE_FALSE;
		  ourGoalieNo = player.number;
		}

	      ourTeamArray[player.number-1].age = 0;
	      ourTeamArray[player.number-1].isGoalie = player.goalie;
	      ourTeamArray[player.number-1].bodyDir = player.body_dir;
	      ourTeamArray[player.number-1].faceDir = player.face_dir;
	      if (ourSide == SIDE_LEFT) {
		ourTeamArray[player.number-1].pos.setXY( player.x, player.y );
		ourTeamArray[player.number-1].v.setXY( player.x_vel, player.y_vel );
	      } else {
		ourTeamArray[player.number-1].pos.setXY( - player.x, - player.y );
		ourTeamArray[player.number-1].v.setXY( - player.x_vel, - player.y_vel );
	      }
	    }
	  else if( player.team == TEAM_THEM )
	    {
	      // set goalie flags if goalie is seen for the first time
	      if( player.goalie == GOALIE_TRUE && otherTeamArray[player.number-1].isGoalie != GOALIE_TRUE )
		{
		  for( int j = 0; j < 11; j++ )
		    if( j != player.number - 1 )
		      otherTeamArray[j].isGoalie = GOALIE_FALSE;
		  otherGoalieNo = player.number;
		}
	      
	      otherTeamArray[player.number-1].age = 0;
	      otherTeamArray[player.number-1].isGoalie = player.goalie;
	      otherTeamArray[player.number-1].bodyDir = player.body_dir;
	      otherTeamArray[player.number-1].faceDir = player.face_dir;
	      if (ourSide == SIDE_LEFT) {
		otherTeamArray[player.number-1].pos.setXY( player.x, player.y );
		otherTeamArray[player.number-1].v.setXY( player.x_vel, player.y_vel );
	      } else {
		otherTeamArray[player.number-1].pos.setXY( - player.x, - player.y );
		otherTeamArray[player.number-1].v.setXY( - player.x_vel, - player.y_vel );
	      }
	    }
	}
    }
  
  
  // adjust goals
  for( int i = 0; i < 2; i++ )
    {
      if( data.exactgoal[i].valid )
	{
	  if( data.exactgoal[i].team == TEAM_US )
	    {
	      if (ourSide == SIDE_LEFT) {
		ourGoal.pos.setXY( data.exactgoal[i].x, data.exactgoal[i].y );
	      } else {
		ourGoal.pos.setXY( - data.exactgoal[i].x, - data.exactgoal[i].y );
	      }
	      ourGoal.age = 0;
	    }
	  if( data.exactgoal[i].team == TEAM_THEM )
	    {
	      if (ourSide == SIDE_LEFT) {
		otherGoal.pos.setXY( data.exactgoal[i].x, data.exactgoal[i].y );
	      } else {
		otherGoal.pos.setXY( - data.exactgoal[i].x, - data.exactgoal[i].y );
	      }
	      otherGoal.age = 0;
	    }
	}
    }
  
  return true;
}


int
Snapshot::ourTeamClosestTo( const Vector2d& target, int maxAge, int index) const
{
  bool          sorted[11];                // for marking players that have already been sorted
  bool          validEntryFound = false;   // Is there a player left who can be sorted?
  int           validEntryIndex = 0;       // index of next player to be sorted
  
  if( index < 1 ) return UNKNOWN;
  if( index > 11 ) return UNKNOWN;

  for( int i = 0; i < 11; i++ )
    sorted[i] = false;
  
  // search until "index" players are marked in index array
  
  for( int i = 0; i < index; i++ )
    {
      validEntryFound = false;
      
      // search for next player for index array

      for( int j = 0; j < 11; j++ )
	{
	  // Is player young enough and farer away than those we already sorted?

	  if ( !sorted[j] && ourTeamArray[j].age <= maxAge )
	    {
	      if( validEntryFound )
		{
		  // do not use dist() for distance check (avoid sqare root)
		  if( sqrnorm( ourTeamArray[j].pos - target ) < sqrnorm( ourTeamArray[validEntryIndex].pos - target ) )
		    validEntryIndex = j;
		}
	      else
		{
		  validEntryFound = true;
		  validEntryIndex = j;
		}
	    }
	}

      if( validEntryFound )
	sorted[validEntryIndex] = true;
      else continue;
    }

  if( validEntryFound )
    return validEntryIndex + 1;   // return player number

  return UNKNOWN;   // not enough valid players found
}


int
Snapshot::otherTeamClosestTo( const Vector2d& target, int maxAge, int index ) const
{
  bool          sorted[11];                // for marking players that have already been sorted
  bool          validEntryFound = false;   // Is there a player left who can be sorted?
  int           validEntryIndex = 0;       // index of next player to be sorted
  
  if( index < 1 ) return UNKNOWN;
  if( index > 11 ) return UNKNOWN;

  for( int i = 0; i < 11; i++ )
    sorted[i] = false;
  
  // search until "index" players are marked in index array
  
  for( int i = 0; i < index; i++ )
    {
      validEntryFound = false;
      
      // search for next player for index array

      for( int j = 0; j < 11; j++ )
	{
	  // Is player young enough and farer away than those we already sorted?

	  if ( !sorted[j] && otherTeamArray[j].age <= maxAge )
	    {
	      if( validEntryFound )
		{
		  // do not use dist() for distance check (avoid sqare root)
		  if( sqrnorm( otherTeamArray[j].pos - target ) < sqrnorm( otherTeamArray[validEntryIndex].pos - target ) )
		    validEntryIndex = j;
		}
	      else
		{
		  validEntryFound = true;
		  validEntryIndex = j;
		}
	    }
	}

      if( validEntryFound )
	sorted[validEntryIndex] = true;
      else continue;
    }

  if( validEntryFound )
    return validEntryIndex + 1;   // return player number

  return UNKNOWN;   // not enough valid players found
}


int
Snapshot::ourTeamDistIndex( int number, const Vector2d& target, int maxAge ) const
{
  int closerPlayers = 0;   // number of players that are closer to "target" than player no. "number"
  
  if( number < 1 ) return 0;
  if( number > 11 ) return 0;

  if( ourTeamArray[number - 1].age > maxAge ) return 0;

  double sqrPlayerDist = sqrnorm( ourTeamArray[number - 1].pos - target );

  for( int i = 0; i < 11; i++ )
    if( i != number - 1 )   // do not check player himself
      if( ourTeamArray[i].age <= maxAge )   // only check young players
	if( sqrnorm( ourTeamArray[i].pos - target ) < sqrPlayerDist )
	  closerPlayers++;

  return closerPlayers + 1;
}


int
Snapshot::otherTeamDistIndex( int number, const Vector2d& target, int maxAge) const
{
  int closerPlayers = 0;   // number of players that are closer to "target" than player no. "number"
  
  if( number < 1 ) return 0;
  if( number > 11 ) return 0;

  if( otherTeamArray[number - 1].age > maxAge ) return 0;

  double sqrPlayerDist = sqrnorm( otherTeamArray[number - 1].pos - target );

  for( int i = 0; i < 11; i++ )
    if( i != number - 1 )   // do not check player himself
      if( otherTeamArray[i].age <= maxAge )   // only check young players
	if( sqrnorm( otherTeamArray[i].pos - target ) < sqrPlayerDist )
	  closerPlayers++;

  return closerPlayers + 1;
}


int
Snapshot::ourTeamInCorridor( const Vector2d& source, const Vector2d& target, double d,
			     const Angle& alpha, int maxAge ) const
{
  int numPlayers = 0;

  for( int i = 0; i < 11; i++ )
    {
      if( ourTeamArray[i].age > maxAge )
	continue;

      if( ourTeamArray[i].inCorridor( source, target, d, alpha ) )
	numPlayers++;
    }

  return numPlayers;
}


int
Snapshot::otherTeamInCorridor( const Vector2d& source, const Vector2d& target, double d,
			       const Angle& alpha, int maxAge ) const
{
  int numPlayers = 0;

  for( int i = 0; i < 11; i++ )
    {
      if( otherTeamArray[i].age > maxAge )
	continue;

      if( otherTeamArray[i].inCorridor( source, target, d, alpha ) )
	numPlayers++;
    }

  return numPlayers;
}


double
Snapshot::ourOffsideLine( int maxAge ) const
{
  double   offX = 0.0;          // x-value for offside line
  double   maxX = 0.0;          // maximum x-value of opponent team (closest to opponent goal)
  double   tmpMaxX = 0.0;       // just for swapping the two x-values above // T.Hummrich

  // this is soccerserver code: it does not know any goalie...

  for( int i = 0; i < 11; i++ )
    {
      if( otherTeamArray[i].age <= maxAge )
	{
	  double playerX = otherTeamArray[i].pos.getX();
	  // we are careful: players could have moved away from their goal
	  playerX -= ( min( otherTeamArray[i].age, maxAge / 2 ) * 
		       otherTeamArray[i].dashPowerRate * MAXPOWER );

	  if( playerX > offX )      // If a player is beyond the offside-line
	    {
	      offX = playerX;       // ...take his x-coordinate for the offside value
	      if( offX > maxX )     // But if he is even beyond the max-line
		{
		  tmpMaxX = maxX;   // ...the maxX becomes the new offX and vice versa
		  maxX = offX;                  
		  offX = tmpMaxX;
		}
	    }

	}
    }

  if( ball.pos.getX() > offX )
    offX = ball.pos.getX();

  return offX;
}


double
Snapshot::otherOffsideLine( int maxAge ) const
{
  double   offX = 0.0;          // x-value for offside line
  double   minX = 0.0;          // minimum x-value of our team (closest to our goal)
  double   tmpMinX = 0.0;       // just for swapping the two x-values above // T.Hummrich

  // this is soccerserver code: it does not know any goalie...

  for( int i = 0; i < 11; i++ )
    {
      if( ourTeamArray[i].age <= maxAge )
	{
	  double playerX = ourTeamArray[i].pos.getX();

	  if( playerX < offX )      // If a player is beyond the offside-line
	    {
	      offX = playerX;       // ...take his x-coordinate for the offside value
	      if( offX < minX )     // But if he is even beyond the min-line
		{
		  tmpMinX = minX;   // ...the maxX becomes the new offX and vice versa
		  minX = offX;                  
		  offX = tmpMinX;
		}
	    }

	}
    }
  
  if( ball.pos.getX() < offX )
    offX = ball.pos.getX();

  return offX;
}

  
double Snapshot::dist(const Snapshot &s, const Snapshot &t, const Vector2d &pos)
{
  // calculates difference between two snaps 
  // (sum of object pos differences biased towards given position)
  double diff_value = 0,dist_from_pos=0;  
  const int ourpl_mult=1,othpl_mult=1,ballpos_mult=3,
    ballv_mult=4,max_dist=80;
  
  // The Ball
  // Calculate distance from chosen player 
  dist_from_pos = norm(pos-s.ball.pos);
  // Max Distance 
  if (dist_from_pos>max_dist)
    {dist_from_pos = max_dist;}
  // Adjust to wanted range (quadratic in distance)
  dist_from_pos = 2-(2*sqr((dist_from_pos)/max_dist));
  // Add ball position to error sum
  diff_value += norm(s.ball.pos-t.ball.pos) * dist_from_pos * 
    ballpos_mult;
  // Add ball speed to error sum
  diff_value += norm(s.ball.v-t.ball.v) * dist_from_pos * 
    ballv_mult;
  
  for (int i=0;i<11;i++)
    {
      // Our players
      // Calculate distance from chosen position
      dist_from_pos = norm(pos-s.ourTeamArray[i].pos);
      // Max Distance 
      if (dist_from_pos>max_dist)
	{dist_from_pos = max_dist;}
      // Adjust to wanted range (quadratic in distance)
      dist_from_pos = 2-(2*sqr((dist_from_pos)/max_dist));
      // Add to error sum
      diff_value += norm(s.ourTeamArray[i].pos-t.ourTeamArray[i].pos) * dist_from_pos * ourpl_mult;

      // Opponent's players
      // Calculate distance from chosen position
      dist_from_pos = norm(pos-s.otherTeamArray[i].pos);
      // Max Distance 
      if (dist_from_pos>max_dist)
	{dist_from_pos = max_dist;}
      // Adjust to wanted range
      dist_from_pos = 2-(2*sqr((dist_from_pos)/max_dist));
      // Add to error sum
      diff_value += norm(s.otherTeamArray[i].pos-t.otherTeamArray[i].pos) * dist_from_pos * othpl_mult;
    }
  return(diff_value);
}

double Snapshot::dist(const Snapshot &s, const Snapshot &t)
{
  // calculates difference between two snaps (sum of object pos differences)
  double diff_value = 0;
  
  // The Ball
  diff_value += norm(s.ball.pos-t.ball.pos);
  
  for (int i=0;i<11;i++)
    {
      // Our players
      diff_value += norm(s.ourTeamArray[i].pos-t.ourTeamArray[i].pos);
      // Opponent's players
      diff_value += norm(s.otherTeamArray[i].pos-t.otherTeamArray[i].pos);
    }
  return(diff_value);
}

double Snapshot::dist(const Snapshot &s, const Snapshot &t, int playernr)
{
  // calculates difference between two snaps 
  // (sum of object pos differences biased towards selected player position)
  double diff_value=0;
  diff_value = Snapshot::dist(s,t,s.ourTeamArray[playernr-1].pos);
  return(diff_value);
}

double Snapshot::balldist(const Snapshot &s, const Snapshot &t)
{
  // calculates difference between two snaps 
  // (sum of object pos differences biased towards ball position)
  double diff_value=0;
  diff_value = Snapshot::dist(s,t,s.ball.pos);
  return(diff_value);
}


bool
Snapshot::binaryWriteToStream( ostream &out ) const
{
  // writes current snap to stream in binary format. Header is 
  // size of Snapshot object in bytes.
  
  if( !out )
    {
      cerr << "Snapshot: output stream not open when trying to write" 
	   << endl;
      return false;          // error: stream not open
    }

  int size = sizeof( *this );

  out << size << endl;       // write header, for version checking of old logfiles
  
  char* tmp = ( char* )this;
  for( int i = 0; i < size; i++ )
    {
      out.put( *tmp );
      tmp++;
    }
  
  if( !out ) return false;   // some error
  else return true;          // writing was ok
}


bool
Snapshot::binaryReadFromStream( istream &in )
{
  // reads binary coded Snapshot from stream. If header is 
  // wrong, or read operation was unsuccessful (missing bytes in stream)
  // returns false, otherwise true;

  if( !in )
    {
      cerr << "Snapshot: input stream not open or faulty when trying to read" 
	   << endl;
      return false; 
    }
  
  int size;
  in >> size;               // try to read header
  if( !in )
    {
      return false;         // eof of stream was reached
    }

  if( size != sizeof( *this ) )
    {
      cerr << "Snapshot: incompatible old recording cannot be used anymore"
	   << endl;
      return false;
    }
  
  int i = 0;
  char* tmp = ( char* )this;
  in.get( *tmp );           // read blank separator at end of header

  while( in && i < size )
    {
      in.get( *tmp );
      tmp++;
      i++;
    }

  if( !in ) 
    return false;   // some weird error appeared while reading from stream
  else 
    return true;    // reading was ok

} 


class ostream& operator << ( class ostream &os, Snapshot& s )
{
  int i;

  os << "time       : " << s.time << endl
     << "ourSide    : " << ( int )s.ourSide << endl
     << endl
     << "my.pos     : " << s.my.pos << endl 
     << "my.bodyDir : " << s.my.bodyDir << endl
     << "my.faceDir : " << s.my.faceDir << endl
     << "my.age         : " << s.my.age << endl << endl;
  
  for( i = 0; i < 11; i++ )
    {
      os << "ourTeam(" << i + 1 << ") : pos : " << s.ourTeam( i + 1 ).pos
	 << ", age : " << s.ourTeam( i + 1 ).age 
	 << endl << endl;
    }
  for( i = 0; i < 11; i++ )
    {
      os << "otherTeam(" << i << ") : pos : " << s.otherTeam( i + 1 ).pos
	 << ", age : " << s.otherTeam( i + 1 ).age
	 << endl << endl;
    }

  os << "ball : pos : " << s.ball.pos
     << ", age : " << s.ball.age << endl
     << "otherGoal.age : " << s.otherGoal.age << endl
     << "ourGoal.age   : " << s.ourGoal.age << endl << endl;
  
  os << "-----------< SenseBody > --------------" << endl;
  os << "age       : " << s.senseBody.age << endl
     << "stamina   : " << s.senseBody.stamina << endl
     << "effort    : " << s.senseBody.effort << endl
     << "velocity  : Length " << s.senseBody.v.getLength()
     << ", Angle " << s.senseBody.v.getAngle() << endl
     << "headAngle : " << s.senseBody.headAngle << endl
     << "kickCount       : " << s.senseBody.kickCount << endl
     << "dashCount       : " << s.senseBody.dashCount << endl
     << "turnCount       : " << s.senseBody.turnCount << endl
     << "sayCount        : " << s.senseBody.sayCount << endl
     << "turnNeckCount   : " << s.senseBody.turnNeckCount << endl
     << "catchCount      : " << s.senseBody.catchCount << endl
     << "moveCount       : " << s.senseBody.moveCount << endl
     << "changeViewCount : " << s.senseBody.changeViewCount << endl;

  os << "***************************************" << endl;
  
  return os;
}


//////////////////////////////////////////////////////////////////////
//
// Operators for giving snapshot to other players using an audio msg
//
//////////////////////////////////////////////////////////////////////


class AudioMsg& operator << ( class AudioMsg& aum, const Snapshot& s )
{
  if( aum.isCoachMsg() )
    {
      aum << s.ball.pos.getX() << s.ball.pos.getY();
      for( int i = 1; i <= 11; i++ )
	aum << s.ourTeam( i ).pos.getX() << s.ourTeam( i ).pos.getY()
	    << s.otherTeam( i ).pos.getX() << s.otherTeam( i ).pos.getY();
    }
  else
    {
      aum << Val5B( s.ball.pos, s.ball.age, s.ball.state )
	  << Val5B( s.ball.v, s.ball.age );
      for( int i = 1; i <= 11; i++ )
	aum << Val5B( s.ourTeam( i ).pos, s.ourTeam( i ).age, s.ourTeam( i ).state ) 
	    << Val5B( s.otherTeam( i ).pos, s.otherTeam( i ).age, s.otherTeam( i ).state );
      
      // add +1 to numbers because NodaChar can not handle negative numbers
      aum << NodaChar( char( s.ourGoalieNo + 1 ) );
      aum << NodaChar( char( s.otherGoalieNo + 1 ) );
    }
  return aum;
}

class AudioMsg& operator >> ( class AudioMsg& aum, Snapshot& s )
{
  Val5B vv;
  NodaChar nc;
  WorldPlayer player;
  double d;

  s.time = aum.getServerTime();
  
  if( aum.isCoachMsg() )
    {
      aum >> d;
      s.ball.pos.setX( d );
      aum >> d;
      s.ball.pos.setY( d );
      s.ball.age = 0;
      s.ball.state = VISUAL;
      s.ball.v.setXY( 0.0, 0.0 );

      player.age = 0;
      player.state = VISUAL;
      player.v.setXY( 0.0, 0.0 );

      for( int i = 1; i <= 11; i++ )
	{
	  aum >> d;
	  player.pos.setX( d );
	  aum >> d;
	  player.pos.setY( d );
	  s.setOurTeam( i, player );

	  aum >> d;
	  player.pos.setX( d );
	  aum >> d;
	  player.pos.setY( d );
	  s.setOtherTeam( i, player );
	}
      
      // from tactical module: opponent player types
      for( int i = 1; i <= 11; i++ )
	{
	  player = s.otherTeam( i );
	  aum >> nc;
	  player.type = int( nc ) - 1;
	  s.setOtherTeam( i, player );
	}
    }
  else
    {
      aum >> vv;
      s.ball.pos = vv;
      s.ball.age = vv.getAge();
      s.ball.state = vv.getState();
      if( !vv.isValid() )
	s.ball.age = ETERNAL;
      
      aum >> vv;
      s.ball.v = vv;
      if( !vv.isValid() )
	s.ball.age = ETERNAL;
      
      for( int i = 1; i <= 11; i++ )
	{
	  aum >> vv;
	  player.pos = vv;
	  player.state = vv.getState();
	  if( vv.isValid() )
	    player.age = vv.getAge();
	  else
	    player.age = ETERNAL;
	  
	  s.setOurTeam( i, player );
    
	  aum >> vv;
	  player.pos = vv;
	  player.state = vv.getState();
	  if( vv.isValid() )
	    player.age = vv.getAge();
	  else
	    player.age = ETERNAL;
	  
	  s.setOtherTeam( i, player );
	}

      // subtract -1 that were added because NodaChar can not handle negative numbers
      aum >> nc;
      s.ourGoalieNo = int( nc ) - 1;
      
      aum >> nc;
      s.otherGoalieNo = int( nc ) - 1;
    }

  return aum;
}





bool
Snapshot::binaryWriteToStreamOSI( ostream &out ) const
{
  // writes current snap to stream in non-binary format. 
  // OSI: independ from OS
  
  if( !out )
    {
      cerr << "Snapshot: output stream not open when trying to write" 
	   << endl;
      return false;          // error: stream not open
    }

  const char *BLANK = " ";

  // write all data to the file

  // writing time and some body-data 
  out << time               << BLANK
      << (int)ourSide       << BLANK
      << my.pos.getX()      << BLANK
      << my.pos.getY()      << BLANK
      << my.number          << BLANK
      << (double)my.bodyDir << BLANK
      << (double)my.faceDir << BLANK
      << my.age             << BLANK
      << (int)(senseBody.viewWidth) << BLANK;
   
  // writing data of our team
  for( int i = 0; i < 11; i++ )
    out << ourTeam( i + 1 ).pos.getX() << BLANK
	<< ourTeam( i + 1 ).pos.getY() << BLANK
	<< ourTeam( i + 1 ).age << BLANK
	<< (double)(ourTeam( i + 1 ).bodyDir) << BLANK;
  
  // writing data of other team
  for( int i = 0; i < 11; i++ )
    out << otherTeam( i + 1 ).pos.getX() << BLANK
	<< otherTeam( i + 1 ).pos.getY() << BLANK
	<< otherTeam( i + 1 ).age << BLANK
	<< (double)(otherTeam( i + 1 ).bodyDir) << BLANK;

  // writing data of the ball and the goals
  out << ball.pos.getX() << BLANK
      << ball.pos.getY() << BLANK
      << ball.age        << BLANK
      << otherGoal.age   << BLANK
      << ourGoal.age     << BLANK
      << ball.v.getX()   << BLANK
      << ball.v.getY()   << BLANK;
  
  // writing sensebody data
  out << senseBody.age             << BLANK
      << senseBody.stamina         << BLANK
      << senseBody.effort          << BLANK
      << senseBody.v.getX()        << BLANK
      << senseBody.v.getY()        << BLANK
      << senseBody.headAngle       << BLANK
      << senseBody.kickCount       << BLANK
      << senseBody.dashCount       << BLANK
      << senseBody.turnCount       << BLANK
      << senseBody.sayCount        << BLANK
      << senseBody.turnNeckCount   << BLANK
      << senseBody.catchCount      << BLANK
      << senseBody.moveCount       << BLANK
      << senseBody.changeViewCount << BLANK;


  if( !out ) 
    return false;   // some error
  else 
    return true;    // writing was ok
}


bool
Snapshot::binaryReadFromStreamOSI( istream &in )
{
  // reads non-binary coded Snapshot from stream. 
  // OSI: independ from OS

  if( !in )
    {
      cerr << "Snapshot: input stream not open or faulty when trying to read"
	   << endl;
      return false; 
    }
  
  // variables used for reading complex data (Vector2d, ...)
  double x, y, angle;
  int theSide, viewWidth;
  
  // reading time and some body-data 
  in >> time           
     >> theSide
     >> x >> y
     >> my.number;
  ourSide = (playing_side)theSide;
  my.pos.setXY(x, y); 
       
  in >> angle;
  my.bodyDir = Angle(angle);    
  in >> angle 
     >> my.age
     >> viewWidth;
  senseBody.viewWidth= (view_width)viewWidth;
  my.faceDir = Angle(angle);
  
  // reading data of our team
  for( int i = 0; i < 11; i++ )
    {
      in >> x >> y
	 >> ourTeamArray[i].age
	 >> angle;
      ourTeamArray[i].pos.setXY(x, y);
      ourTeamArray[i].bodyDir= Angle(angle);
      ourTeamArray[i].number = i+1;
    } 
  // reading data of other team
  for( int i = 0; i < 11; i++ )
    {
      in >> x >> y  
	 >> otherTeamArray[i].age
	 >> angle;
      otherTeamArray[i].pos.setXY(x, y);
      otherTeamArray[i].bodyDir= Angle(angle);
      otherTeamArray[i].number = i+1;
    }

  // reading data of the ball and the goals
  in >> x >> y
     >> ball.age      
     >> otherGoal.age 
     >> ourGoal.age;
  ball.pos.setXY(x, y);      

  in >> x >> y;
  ball.v.setXY(x, y);
  
  // reading sensebody data
  in >> senseBody.age           
     >> senseBody.stamina       
     >> senseBody.effort
     >> x >> y
     >> angle
     >> senseBody.kickCount     
     >> senseBody.dashCount     
     >> senseBody.turnCount     
     >> senseBody.sayCount      
     >> senseBody.turnNeckCount
     >> senseBody.catchCount
     >> senseBody.moveCount
     >> senseBody.changeViewCount;

  senseBody.v.setXY(x, y);
  senseBody.headAngle = (Angle)angle;     

  if( !in ) 
    return false;   // some weird error appeared while reading from stream
  else 
    return true;    // reading was ok

}

int Snapshot::nextTimeOfChannel( const AumulChannel ch ) const
{
  for (int i=0; i<7; i++)
    if (getChannelTime(i+time) == ch)
      return i+time;
  return -2;
}

AumulChannel Snapshot::getChannelTime( const int time ) const
{
  switch (time%3) {
  case 0:
    return AUMUL_NONE;
    break;
  case 1:
    return AUMUL_EMRG;
    break;
  case 2:
  default:
    if ((time-my.number) % 2)
      return AUMUL_MY_WM;
    else
      return AUMUL_OTHER_WM;
  }
}
