//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////// 
//   $Id: snapshot.h,v 1.34 2001/08/01 01:18:08 mrb Exp $
//   $Manager: cm$
//
//   snapshot.h
//
////////////////////////////////////////////////////////////////////////

#ifndef _SNAPSHOT_H_
#define _SNAPSHOT_H_

#include "sensor.h"         // sensor data types
#include "coordinates.h"    // angle class and vector class for 2-dimensional space
#include "constants.h"      // constants used by soccerserver
#include "minmax.h"         // min, max, sqr

// constants needed by worldmodel classes

const int UNKNOWN = -1;          // this is used for unknown player numbers and types
const int DEFAULT_TYPE = 0;      // default player type
const int ETERNAL = 32000;       // age of objects we have never seen (->invalid objects)
                                 // the value should be larger than match length
const int MAXAGE  = 20;          // objects older than this are (by default) not
                                 // considered in the sorting methods


// utility functions

// test if "pos" is in field (with security distance)
bool inField( const Vector2d &pos, double security = 0.0 );
// test if "pos" is in our penalty area (with security distance)
bool inOurPenaltyArea( const Vector2d &pos, double security = 0.0 );
// test if "pos" is in opponent penalty area (with security distance)
bool inOtherPenaltyArea( const Vector2d &pos, double security = 0.0 );


// Reports if there could have been a player<->ball collision when
// simulating a snapshot.

enum CollisionState
{
  NONE,        // There was no collision (according to our data)
  CRITICAL,    // There MUST have been a collision, at least according to our
               // simulated data, but simulation of this collision might be wrong
               // because soccerserver uses random values in this case (very rare!)
  PROCESSED    // There WAS a collision (acc. to our data) AND it could be simulated
};


// State of a WorldObject (interesting for world model updates)

enum WorldObjectState
{
  UNSAFE,   // State might be undefined or unsafe; e.g. Object is a player and is 
            // simulated from visual data, but the number might be wrong
  FUZZY,    // Object is simulated from data that fitted to some older visual data;
            // e.g. object is a player without visible number but fitted to some
            // older visual information
  VISUAL,   // Object is simulated from visual data
  AUDIAL,   // Object is simulated from audial data
};


// collects data about a player substitution (player number and timestep)

struct Substitution
{
  int number;
  int time;
};


//////////////////////////////////////////////////////////////////////
//
// Class WorldObject
//
//////////////////////////////////////////////////////////////////////

class WorldObject {
public:
  WorldObject( const Vector2d& _pos = Vector2d() ) :
    pos( _pos ), age( ETERNAL ), state( UNSAFE ) {}

  Vector2d  pos;   // absolute position on the field
  int       age;   // number of simulation steps that have passed since this object
                   // has been updated for the last time

  WorldObjectState state;   // tells which data type object was simulated from

  // calculate distance to a given object "obj"
  double    distTo( const WorldObject& obj ) const { return dist( pos, obj.pos ); }

  // check if object is in corridor as sketched below:
  //              ___________________________________ __         _
  //             /                                   |  \         |
  //            /                                    |   \        |
  // source -> /_ _ _ _ _ _ _ _ _ _ _ _ _ target -> _|_   |        \ d
  //           \                                     |    |        /
  //            \                                    |   /        |
  //             \___________________________________|__/        _|
  //
  // On the right there is a half circle with diameter d
  // (which is difficult to draw with ASCII characters...).
  // The angle of the cone on the left is 2 * alpha.
  // 
  // CAUTION: it is also checked if object is in a very small circle around source!
  bool      inCorridor( Vector2d source, Vector2d target, double d, Angle alpha = PI_2 ) const;

  bool      inField( double security = 0.0 ) const;
  bool      inOurPenaltyArea( double security = 0.0 ) const;
  bool      inOtherPenaltyArea( double security = 0.0 ) const;
};


//////////////////////////////////////////////////////////////////////
//
// Class WorldMovedObject
//
//////////////////////////////////////////////////////////////////////

class WorldPlayer;

class WorldMovedObject: public WorldObject {
public:
  WorldMovedObject( double _decay = 0.0,
		    double _vmax = 0.0,
		    double _size = 0.0,
		    const Vector2d& _pos = Vector2d( 0.0, 0.0 ),
		    const Vector2d& _v = Vector2d( 0.0, 0.0 ) ) :
    WorldObject( _pos ), v( _v ), decay( _decay ), vmax( _vmax ), size( _size ), coneTime( 0 ) {}

  // accelerate object, but consider maximum velocity
  void      acc( const Vector2d& _a )
  { v += _a; if( v.getLength() > vmax ) v *= vmax / v.getLength(); }

  // simulate one moving step
  void      simulate() { pos += v; v *= decay; }

  // calculate object's absolute velocity if distChange and dirChange relative to 
  // relObject (usually player himself) are given
  void      calcVelocity( const WorldMovedObject& relObject, double distChange, double dirChange );

  // update object position with sensor data (sensor data = relative position to "my")
  void      setPos( const SensorObject& object, const WorldPlayer& my );

  Vector2d  v;          // velocity vector
  double    decay;      // um... that's how we brake (decay factor)
  double    vmax;       // maximum LENGTH of velocity vector
  double    size;       // object radius
  int       coneTime;   // time object spent in a certain view cone
  // (does not make sense without connection to a player)
};


//////////////////////////////////////////////////////////////////////
//
// Class WorldBall
//
//////////////////////////////////////////////////////////////////////

class WorldBall : public WorldMovedObject {
public:
  WorldBall( const Vector2d& _pos = Vector2d( 0.0, 0.0 ),
	     const Vector2d& _v = Vector2d( 0.0, 0.0 ) ) :
    WorldMovedObject( BALL_DECAY, BALL_SPEED_MAX, BALL_SIZE, _pos, _v ) {}

  // calculate ball position with sensor data (sensor data = relative position to a "player")
  void calcPos( const SensorObject& object, const WorldPlayer& my );
};


//////////////////////////////////////////////////////////////////////
//
// Class WorldPlayer
//
//////////////////////////////////////////////////////////////////////

class WorldPlayer : public WorldMovedObject {
public:
  WorldPlayer( const Vector2d& _pos = Vector2d( 0.0, 0.0 ),
	       const Vector2d& _v = Vector2d( 0.0, 0.0 ),
	       const Angle& _bodyDir = 0.0,
	       const Angle& _faceDir = 0.0,
	       int _number = UNKNOWN,
	       goalie_player _isGoalie = GOALIE_UNKNOWN ) :
    WorldMovedObject( PLAYER_DECAY, PLAYER_SPEED_MAX, PLAYER_SIZE, _pos, _v ),
    bodyDir( _bodyDir ), faceDir( _faceDir ), number( _number ), isGoalie( _isGoalie ),
    staminaIncMax( STAMINA_INC_MAX ), inertiaFactor( INERTIA_MOMENT ),
    dashPowerRate( DASH_POWER_RATE ), kickableMargin( KICKABLE_MARGIN ),
    kickRand( KICK_RAND ), extraStamina( 0.0 ),
    effortMin( EFFORT_MIN ), effortMax( EFFORT_MAX ),
    type( UNKNOWN )
    {}

  // calculate angle between player's BODY direction and "obj"
  Angle      bodyAngleTo( const WorldObject& obj ) const
  { return ( obj.pos - pos ).getAngle() - bodyDir; }

  // calculate angle between player's FACE direction and "obj"
  Angle      faceAngleTo( const WorldObject& obj ) const
  { return ( obj.pos - pos ).getAngle() - faceDir; }

  // calculate player's kick range
  double     kickRange() const { return BALL_SIZE + size + kickableMargin; }

  // check if ball is in kickable area
  bool       ballKickable( const WorldBall& ball ) const
  { return distTo( ball ) < kickRange() ? true : false; }

  // check if ball is in catchable area (for a catch command with catchAngle)
  bool       ballCatchable( const WorldBall& ball, const Angle& catchAngle ) const;

  // calculate actual kick power from angle and distance to ball
  double     kickPowerRate( const WorldBall& ball ) const;

  // calculate inertia moment for turn command
  double     inertiaMoment() const { return v.getLength() * inertiaFactor + 1; }

  Angle      bodyDir;          // player's body-and-running direction
  Angle      faceDir;          // player's facing direction ( = bodyDir + headAngle )

  int        number;           // player's number ( -1 = undefined, 0 = coach )
  goalie_player isGoalie;      // is player goalie?
  bool       isCoach() const   // is player not a field player but the coach?
  { return ( number == 0 ); }

  // Parameters needed since heterogenuous players have been introduced
  double     staminaIncMax;
  double     inertiaFactor;   // CAUTION: This is called "inertia moment" by the server!
  double     dashPowerRate;
  double     kickableMargin;
  double     kickRand;
  double     extraStamina;
  double     effortMin;
  double     effortMax;

  int        type;            // player type (-1 = UNKNOWN, 0 = DEFAULT, 1..)

  // update player with sensor data ( sensor data = relative position to "my")
  void       setBySensorData( SensorPlayer player, const WorldPlayer& my );
};


//////////////////////////////////////////////////////////////////////
//
// Class WorldPlayerType
//
//////////////////////////////////////////////////////////////////////

class WorldPlayerType {
public:
  WorldPlayerType() :
    state( UNKNOWN ), playerSpeedMax( PLAYER_SPEED_MAX ),
    playerDecay( PLAYER_DECAY ), playerSize( PLAYER_SIZE ),
    staminaIncMax( STAMINA_INC_MAX ), inertiaFactor( INERTIA_MOMENT ),
    dashPowerRate( DASH_POWER_RATE ), kickableMargin( KICKABLE_MARGIN ),
    kickRand( KICK_RAND ), extraStamina( 0.0 ),
    effortMin( EFFORT_MIN ), effortMax( EFFORT_MAX )
    {}

  int    state;    // -1 == UNKNOWN, 0 == default, player's type otherwise

  double playerSpeedMax;
  double playerDecay;
  double playerSize;
  double staminaIncMax;
  double inertiaFactor;
  double dashPowerRate;
  double kickableMargin;
  double kickRand;
  double extraStamina;
  double effortMin;
  double effortMax;
};

class ostream & operator<<( class ostream &os, const WorldPlayerType &pt);

//////////////////////////////////////////////////////////////////////
//
// Class WorldSense
//
//////////////////////////////////////////////////////////////////////

class WorldSense {
public:
  WorldSense() 
    : age( ETERNAL ), stamina( 0.0 ), effort( 1.0 ), v( 0.0, 0.0 ), headAngle( 0.0 ),
    kickCount( 0 ), dashCount( 0 ), turnCount( 0 ), sayCount( 0 ), turnNeckCount( 0 ),
    catchCount( 0 ), moveCount( 0 ), changeViewCount( 0 ),
    viewWidth( VIEW_WIDTH_NORMAL ), viewQuality( VIEW_QUALITY_HIGH ), recovery( 1.0 ) {}
  
  int          age;            // time that has passed since this information has been generated

  // all values soccerserver provides

  double       stamina;        // player's stamina
  double       effort;         // player's effort (which shows how "tired" player is)

  Vector2d     v;              // player's velocity
  Angle        headAngle;      // angle of player's head relative to the body

  int          kickCount;      // numbers of commands server has received
  int          dashCount;
  int          turnCount;
  int          sayCount;
  int          turnNeckCount;
  int          catchCount;     // the last three are new in soccerserver 7.0
  int          moveCount;
  int          changeViewCount;

  view_width   viewWidth;      // angle of view cone
  view_quality viewQuality;    // quality of visual information

  static Angle getViewAngle( const view_width viewWidth );
  Angle        getViewAngle() const;   // calculate view angle (= HALF of view cone!) from viewWidth

  double       recovery;   // recovery value for calculation of player's stamina. This value
                           // is not reported by soccerserver, but it makes sense to store
                           // it together with effort and stamina
};

//////////////////////////////////////////////////////////////////////
//
// some stuff for audio multiplexing 
// - functions can be found in Snapshot!
//
//////////////////////////////////////////////////////////////////////

enum AumulChannel 
{
  AUMUL_NONE,             // gap channel - no screaming allowed at all
  AUMUL_EMRG,             // emergency channel
  AUMUL_MY_WM,            // worldmodel broadcast & my turn to scream
  AUMUL_OTHER_WM          // worldmodel broadcast & not my turn
};


//////////////////////////////////////////////////////////////////////
//
// Class Snapshot
//
//////////////////////////////////////////////////////////////////////

class Snapshot {
public:
  Snapshot();

  WorldPlayer      my;               // myself

  // get player with number "n" of our team
  const WorldPlayer      &ourTeam( int n ) const
  {
    if( n < 1 || n > 11 )
      cerr << "ERROR: ourTeam() call with parameter " << n
	   << ", my.number = " << my.number << endl;
    ensure( n >= 1 && n <= 11 );
    return ourTeamArray[n-1];
  }

  // get player with number "n" of other team
  const WorldPlayer      &otherTeam( int n ) const
  {
    if( n < 1 || n > 11 )
      cerr << "ERROR: otherTeam() call with parameter " << n
	   << ", my.number = " << my.number << endl;
    ensure( n >= 1 && n <= 11 );
    return otherTeamArray[n-1];
  }

  // set player with number "n" of our team to "player"
  // CAUTION: do not use this to change snapshots in the worldmodel!
  void setOurTeam( int n, const WorldPlayer& player )
  { ensure( n >= 1 && n <= 11 ); ourTeamArray[n-1] = player; }

  // set player with number "n" of other team to "player"
  // CAUTION: do not use this to change snapshots in the worldmodel!
  void setOtherTeam( int n, const WorldPlayer& player )
  { ensure( n >= 1 && n <= 11 ); otherTeamArray[n-1] = player; }

  int              ourGoalieNo;      // number of our goalie (-1 = unknown)
  int              otherGoalieNo;    // number of opponents' goalie

  WorldBall        ball;
  
  CollisionState   collState;        // reports player<->ball collisions (after simulation)

  WorldObject      ourGoal;
  WorldObject      otherGoal;
  
  WorldSense       senseBody;        // information given by SenseBody (e.g. stamina)

  int              time;             // simulation time that this snapshot belongs to 
  int              visualTime;       // time of latest visual information for this snapshot
  int              lastCatchTime;    // time when last catch command has been sent

  playing_side     ourSide;
  play_mode        playMode;

  int              ourScore;         // number of goals we scored
  int              otherScore;       // number of goals scored by opponent team

  WorldPlayerType* playerType;       // the player types generated by the server
  Substitution*    ourSubs;          // the substitutions we made
  int              numOurSubs;       // ...and the number of them
  Substitution*    otherSubs;        // the substitutions the opponent made
  int              numOtherSubs;     // ...and the number of them

  // the following 3 methods are used to set information invalid, i.e.
  // set its age to ETERNAL

  void    setInvalid();                 // visual and sense information
  void    setSeeInvalid();              // only visual information
  void    setSenseInvalid();            // only sense information

  // Get the number of that player of our resp. other team whose rank is "index" if the
  // players are sorted by distance to "target". Default for "index" is 1 (closest player to
  // "target"). Consider only players younger than "maxAge". Considering players we have never
  // seen before (which for example could mean that their pos is still marked as zero!) can be
  // avoided by setting maxAge < ETERNAL.
  // If no suitable player is found, UNKNOWN is returned. (!)
  // CAUTION: player himself might be returned by ourTeamClosestTo()!

  int     ourTeamClosestTo( const Vector2d& target, int maxAge = MAXAGE, int index = 1 ) const;
  int     otherTeamClosestTo( const Vector2d& target, int maxAge = MAXAGE, int index = 1 ) const;

  // Get the rank of player no. "number" if our resp. other team is sorted by distance to "target".
  // Only players with age <= maxAge are considered. If player with no. "number" is older than maxAge,
  // 0 is returned.

  int     ourTeamDistIndex( int number, const Vector2d& target, int maxAge = MAXAGE ) const;
  int     otherTeamDistIndex( int number, const Vector2d& target, int maxAge = MAXAGE ) const;
  
  // count players in our resp. other team that are in a certain corridor between "source" and
  // "target" (for details of the corridor see WorldObject::inCorridor())
  // Consider only players with age <= maxAge.
  // CAUTION: player himself will be in the corridor and thus counted by ourTeamInCorridor()!

  int     ourTeamInCorridor( const Vector2d& source, const Vector2d& target, double d,
			     const Angle& alpha = PI_2, int maxAge = MAXAGE ) const;
  int     otherTeamInCorridor( const Vector2d& source, const Vector2d& target, double d,
			       const Angle& alpha = PI_2, int maxAge = MAXAGE ) const;
  
  // calculate minimum x-position that is judged to be offside
  // Only opponent players with age <= maxAge are considered
  double  ourOffsideLine( int maxAge ) const;

  // calculate the opponent's offside line (of interest for our defenders)
  double  otherOffsideLine( int maxAge ) const;

  // calculate difference between two snapshots by comparing player and ball positions
  // ( 0 = no difference, otherwise sum of distances between corresponding objects )
  // may be given a playernumber as an argument. In this case sum will be biased
  // towards objects which are close to that player (assume Snapshot s to be 'correct')
  // balldist will do the same biasing towards object close to the ball. 
  // See implementation for amount of biasing.
  static double dist(const Snapshot &s, const Snapshot &t);
  static double dist(const Snapshot &s, const Snapshot &t, int playernr);
  static double balldist(const Snapshot &s, const Snapshot &t);

  // write/read Snapshot as binary to/from stream
  bool    binaryWriteToStream( ostream &out ) const;
  bool    binaryReadFromStream( istream &in );
  // write/read Snapshot non-binary to/from stream
  // OSI: independend from OS
  bool    binaryWriteToStreamOSI( ostream &out ) const;
  bool    binaryReadFromStreamOSI( istream &in );

  // audio multiplexing control functions:
  // get next occurrence of a given channel
  int     nextTimeOfChannel( const AumulChannel ch ) const;
  // get channel for a given timestep
  AumulChannel getChannelTime( const int time ) const;
  // get channel for "now"
  AumulChannel getChannelNow() const
  { return getChannelTime(time); }

private:
  WorldPlayer         ourTeamArray[11];      // players of my team sorted by number
  WorldPlayer         otherTeamArray[11];    // players of other team sorted by number
  // CAUTION: access to player with number i is via our/otherTeamArray[i-1]

  static double dist( const Snapshot &s, const Snapshot &t, const Vector2d &pos );
  
  bool    adjustSense( SensorData &data );   // fill in new sense data
  // fill in new visual data; oldBallRelpos is needed for ball velocity
  bool    adjustSee( SensorData &data, const Vector2d& oldBallRelpos = Vector2d( 0.0, 0.0 ) );
  bool    adjustExact( SensorData &data );   // fill in new exact data

  // Calculate my absolute position and face direction with new sensor data.
  // Face direction is calculated with a line, if only one line is seen
  // (this means we are inside the field), otherwise with flags.
  // Position is calculated with flags.
  // calcMyPos() is true if position could be calculated.
  bool    calcMyPos( SensorData& data );

  bool    adjustPlayers( SensorData &data ); // adjust other players to my (new) position
  // same with ball; ball relpos from one step ago might be needed for
  // calculation of ball velocity with position difference
  bool    adjustBall( SensorData &data, const Vector2d& oldBallRelpos = Vector2d( 0.0, 0.0 ) );
  bool    adjustGoals( SensorData &data );   // same with goals

  friend class Worldmodel;   // Worldmodel calls adjusting methods
};


class ostream& operator<< (class ostream &os, Snapshot& s);

class AudioMsg& operator<< (class AudioMsg& aum, const Snapshot& s);
class AudioMsg& operator>> (class AudioMsg& aum, Snapshot& s);

#endif
