//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////// 
//   $Id: worldmodel.C,v 1.57 2001/08/02 01:02:05 mrb Exp $	
//   $Manager: cm$
//
//   worldmodel.C
//
////////////////////////////////////////////////////////////////////////


#include "unistd.h"
#include "math.h"
#include "mytrace.h"
#include "worldmodel.h"
#include "Ensure.h"


// output generated when checking for collisions
//#define COLLISIONTRACE(output) cout << output << endl

// output generated when merging audial information into worldmodel
//#define AUDIALINPUTTRACE(output) cout << output

// output generated when spreading worldmodel to other players
//#define AUDIALOUTPUTTRACE(output) cout << output


#ifndef COLLISIONTRACE
#define COLLISIONTRACE(output)
#else
#warning COLLISIONTRACE debug trace COLLISIONTRACE is activated!
#endif

#ifndef AUDIALINPUTTRACE
#define AUDIALINPUTTRACE(output)
#else
#warning AUDIALINPUTTRACE debug trace AUDIALINPUTTRACE is activated!
#endif

#ifndef AUDIALOUTPUTTRACE
#define AUDIALOUTPUTTRACE(output)
#else
#warning AUDIALOUTPUTTRACE debug trace AUDIALOUTPUTTRACE is activated!
#endif


//////////////////////////////////////////////////////////////////////
//
// Class Worldmodel
//
//////////////////////////////////////////////////////////////////////


Worldmodel::Worldmodel( int _numSnapshots, Comm* _comm )
{
  numSnapshots = _numSnapshots;
  comm = _comm;

  visualTime = -1;                    // no sensor information before kick off
  visAudTime = -1;
  senseTime = -1;
  simVisAudTime = -1;
  simSenseTime = -1;
  newVisAudData = false;
  newSenseData = false;
  maxTime = -1;
  lastCatchTime = -CATCH_BAN_CYCLE;   // goalie is allowed to catch at once (i.e. in timestep 0)

  ensure( numSnapshots <= 30 );       // just to avoid out of memory (30 IS a lot!)
  
  snapshots = new Snapshot[numSnapshots];
  ensure( snapshots );
  
  actions = new WorldAction[numSnapshots];
  ensure( actions );

  for( int i = 0; i < numSnapshots; i++ )
    {
      snapshots[i].time = i - numSnapshots;   // set time into the past
      snapshots[0].time = 0;                  // ...but not for time step 0
      snapshots[i].visualTime = i - numSnapshots;
      snapshots[0].visualTime = 0;
      snapshots[i].lastCatchTime = -CATCH_BAN_CYCLE;
      snapshots[i].playMode = comm->getPlayMode();
      snapshots[i].ourSide = comm->getMySide();
      snapshots[i].my.number = comm->getMyNumber();
      snapshots[i].my.type = DEFAULT_TYPE;
      snapshots[i].my.isGoalie = comm->amIGoalie() ? GOALIE_TRUE : GOALIE_FALSE;
      if( comm->amIGoalie() )
	{
	  snapshots[i].ourGoalieNo = comm->getMyNumber();
	  for( int j = 0; j < 11; j++ )
	  if( j != comm->getMyNumber() - 1 )
	    snapshots[i].ourTeamArray[j].isGoalie = GOALIE_FALSE;
	}
      snapshots[i].ourScore = comm->getOurScore();
      snapshots[i].otherScore = comm->getOtherScore();
      
      // player types (including default type (ZERO))
      snapshots[i].playerType = new WorldPlayerType[playerParams.player_types];
      snapshots[i].playerType[0].state = DEFAULT_TYPE;

      // space for allowed substitutions plus subsitutions before kick off
      snapshots[i].ourSubs = new Substitution[playerParams.subs_max + 11];
      snapshots[i].otherSubs = new Substitution[playerParams.subs_max + 11];
      snapshots[i].numOurSubs = 0;
      snapshots[i].numOtherSubs = 0;

      for( int j = 0; j < 11; j++)
	{
	  snapshots[i].ourTeamArray[j].type = DEFAULT_TYPE;
	  snapshots[i].ourTeamArray[j].number = j + 1;
	  snapshots[i].otherTeamArray[j].number = j + 1;
	}
    }
}


Worldmodel::Worldmodel( const Worldmodel& wm )
{
  numSnapshots = wm.numSnapshots;
  comm = wm.comm;

  visualTime = wm.visualTime;
  visAudTime = wm.visAudTime;
  senseTime = wm.senseTime;
  simVisAudTime = wm.simVisAudTime;
  simSenseTime = wm.simSenseTime;
  newVisAudData = wm.newVisAudData;
  newSenseData = wm.newSenseData;
  maxTime = wm.maxTime;
  lastCatchTime = wm.lastCatchTime;

  snapshots = new Snapshot[numSnapshots];
  ensure( snapshots );

  actions = new WorldAction[numSnapshots];
  ensure( actions );
  
  for( int i = 0; i < numSnapshots; i++ )
    {
      snapshots[i] = wm.snapshots[i];
      actions[i] = wm.actions[i];

      // player types (including default type (ZERO))
      snapshots[i].playerType = new WorldPlayerType[playerParams.player_types];
      for( int j = 0; j < playerParams.player_types; j++ )
	snapshots[i].playerType[j] = wm.snapshots[i].playerType[j];

      // space for allowed substitutions plus subsitutions before kick off
      snapshots[i].ourSubs = new Substitution[playerParams.subs_max + 11];
      snapshots[i].otherSubs = new Substitution[playerParams.subs_max + 11];
      snapshots[i].numOurSubs = wm.snapshots[i].numOurSubs;
      snapshots[i].numOtherSubs = wm.snapshots[i].numOtherSubs;
      for( int j = 0; j < snapshots[i].numOurSubs; j++ )
	snapshots[i].ourSubs[j % ( playerParams.subs_max + 11 )] =
	  wm.snapshots[i].ourSubs[j % ( playerParams.subs_max + 11 )];
      for( int j = 0; j < snapshots[i].numOtherSubs; j++ )
	snapshots[i].otherSubs[j % ( playerParams.subs_max + 11 )] =
	  wm.snapshots[i].otherSubs[j % ( playerParams.subs_max + 11 )];
    }
}


Worldmodel& 
Worldmodel::operator = ( const Worldmodel& wm )
{
  numSnapshots = wm.numSnapshots;
  comm = wm.comm;

  visualTime = wm.visualTime;
  visAudTime = wm.visAudTime;
  senseTime = wm.senseTime;
  simVisAudTime = wm.simVisAudTime;
  simSenseTime = wm.simSenseTime;
  newVisAudData = wm.newVisAudData;
  newSenseData = wm.newSenseData;
  maxTime = wm.maxTime;

  lastCatchTime = wm.lastCatchTime;

  for( int i = 0; i < numSnapshots; i++ )
    {
      delete [] snapshots[i].playerType;
      delete [] snapshots[i].ourSubs;
      delete [] snapshots[i].otherSubs;
    }

  delete [] snapshots;
  delete [] actions;

  snapshots = new Snapshot[numSnapshots];
  ensure( snapshots );

  actions = new WorldAction[numSnapshots];
  ensure( actions );

  for( int i = 0; i < numSnapshots; i++ )
    {
      snapshots[i] = wm.snapshots[i];
      actions[i] = wm.actions[i];

      // player types (including default type (ZERO))
      snapshots[i].playerType = new WorldPlayerType[playerParams.player_types];
      for( int j = 0; j < playerParams.player_types; j++ )
	snapshots[i].playerType[j] = wm.snapshots[i].playerType[j];

      // space for allowed substitutions plus subsitutions before kick off
      snapshots[i].ourSubs = new Substitution[playerParams.subs_max + 11];
      snapshots[i].otherSubs = new Substitution[playerParams.subs_max + 11];
      snapshots[i].numOurSubs = wm.snapshots[i].numOurSubs;
      snapshots[i].numOtherSubs = wm.snapshots[i].numOtherSubs;
      for( int j = 0; j < snapshots[i].numOurSubs; j++ )
	snapshots[i].ourSubs[j % ( playerParams.subs_max + 11 )] =
	  wm.snapshots[i].ourSubs[j % ( playerParams.subs_max + 11 )];
      for( int j = 0; j < snapshots[i].numOtherSubs; j++ )
	snapshots[i].otherSubs[j % ( playerParams.subs_max + 11 )] =
	  wm.snapshots[i].otherSubs[j % ( playerParams.subs_max + 11 )];
    }
  return *this;
}


Worldmodel::~Worldmodel() 
{
  for( int i = 0; i < numSnapshots; i++ )
    {
      delete [] snapshots[i].playerType;
      delete [] snapshots[i].ourSubs;
      delete [] snapshots[i].otherSubs;
    }

  delete [] snapshots;
  delete [] actions;
}


bool
Worldmodel::getAction()
{
  CycleAction action;

  comm->getLastAction( action );

  int actionIndex = getTimeIndex( action.time );

  // check if comm module gave us a new command (if not, return false)

  if( actions[actionIndex].time == action.time && actions[actionIndex].type == action.command )
    {
      switch( action.command )
	{
	case NO_COMMAND:    // param1 = param2 = 0
	  return false;

	case TURN_COMMAND:  // param1 = moment_rad, param2 = 0
	  if( actions[actionIndex].moment == action.param1 )
	    return false;
	  break;
      
	case DASH_COMMAND:  // param1 = power, param2 = 0
	  if( actions[actionIndex].power == action.param1 )
	    return false;
	  break;
      
	case KICK_COMMAND:  // param1 = power, param2 = dir_rad
	  if( actions[actionIndex].power == action.param1 &&
	      actions[actionIndex].moment == action.param2 )
	    return false;
	  break;
      
	case CATCH_COMMAND: // for goalie only; param1 = dir_rad, param2 = 0
	  if( actions[actionIndex].moment == action.param1 )
	    return false;
	  break;
      
	case MOVE_COMMAND:  // only for coach or in certain playmodes; param1=x, param2=y
	  if( actions[actionIndex].pos.getX() == action.param1 &&
	      actions[actionIndex].pos.getY() == action.param2 )
	    return false;
	  break;

	default:
	  break;
	}
    }

  actions[actionIndex].time = action.time;
  actions[actionIndex].type = action.command;

  actions[actionIndex].moment = 0.0;    // cancel old data
  actions[actionIndex].power  = 0.0;
  actions[actionIndex].pos.setXY( 0.0, 0.0 );

  switch( action.command )
    {
    case NO_COMMAND:    // param1 = param2 = 0
      break;
      
    case TURN_COMMAND:  // param1 = moment_rad, param2 = 0
      actions[actionIndex].moment = action.param1;
      break;
      
    case DASH_COMMAND:  // param1 = power, param2 = 0
      actions[actionIndex].power = action.param1;
      break;
      
    case KICK_COMMAND:  // param1 = power, param2 = dir_rad
      actions[actionIndex].power  = action.param1;
      actions[actionIndex].moment = action.param2;
      break;
      
    case CATCH_COMMAND: // for goalie only; param1 = dir_rad, param2 = 0
      actions[actionIndex].moment = action.param1;
      break;
      
    case MOVE_COMMAND:  // only for coach or in certain playmodes; param1=x, param2=y
      actions[actionIndex].pos.setXY( action.param1, action.param2 );
      break;
      
    default:
      break;
    }

  return true;
}


void
Worldmodel::simVisualData( int time )
{
  // simulate visual information time -> time+1

  double   turnMoment;
  double   actualPower;   // needed for simulation of a dash command

  WorldAction& action = actions[getTimeIndex( time )];
  
  Snapshot& oldSnap = snapshots[getTimeIndex( time )];
  Snapshot& newSnap = snapshots[getTimeIndex( time+1 )];

  // step 1 : copy old information

  newSnap.visualTime = oldSnap.visualTime;
  newSnap.lastCatchTime = oldSnap.lastCatchTime;
  newSnap.my = oldSnap.my;                    // player himself
  newSnap.ball = oldSnap.ball;                // ball
  newSnap.otherGoal = oldSnap.otherGoal;      // goals
  newSnap.ourGoal = oldSnap.ourGoal;
  newSnap.ourScore = oldSnap.ourScore;
  newSnap.otherScore = oldSnap.otherScore;
  newSnap.ourGoalieNo = oldSnap.ourGoalieNo;
  newSnap.otherGoalieNo = oldSnap.otherGoalieNo;

  for( int i = 0; i < 11; i++ )               // players of both teams
    {
      newSnap.ourTeamArray[i] = oldSnap.ourTeamArray[i];
      newSnap.otherTeamArray[i] = oldSnap.otherTeamArray[i];
    }

  // step 2 : Birthday party for all objects

  if( newSnap.my.age < ETERNAL ) newSnap.my.age++;
  if( newSnap.ball.age < ETERNAL ) newSnap.ball.age++;
  if( newSnap.otherGoal.age < ETERNAL ) newSnap.otherGoal.age++;
  if( newSnap.ourGoal.age < ETERNAL ) newSnap.ourGoal.age++;
  for( int i = 0; i < 11; i++ )
    {
      if( newSnap.ourTeamArray[i].age < ETERNAL ) newSnap.ourTeamArray[i].age++;
      if( newSnap.otherTeamArray[i].age < ETERNAL ) newSnap.otherTeamArray[i].age++;
    }
  
  // step 3 : if player has desired an action, simulate it now
  
  if( action.time == time )   // is action valid?
    {
      switch( action.type )
	{
	case DASH_COMMAND: 
	  // simulation of a dash command requires knowledge about the stamina value.
	  // That's why in getSnapshot() sense information is simulated first

	  actualPower = action.power;

	  if( actualPower < 0 )   // dashing backward costs double stamina
	    {
	      actualPower *= -2.0;
	      // "will-power" is new in soccerserver 7.0
	      if( newSnap.senseBody.stamina + newSnap.my.extraStamina < actualPower )
		actualPower = newSnap.senseBody.stamina + newSnap.my.extraStamina;
	      actualPower /= -2.0;
	    }
	  else
	    if( newSnap.senseBody.stamina + newSnap.my.extraStamina < actualPower )
	      actualPower = newSnap.senseBody.stamina + newSnap.my.extraStamina;

	  actualPower *= ( newSnap.senseBody.effort * newSnap.my.dashPowerRate );

	  newSnap.my.acc( radAng2Vector2d( actualPower, newSnap.my.bodyDir ) );

	  // move ball if goalie dashes after a catch command
	  // caution: this does not check if goalie really sent a catch command!
	  if( newSnap.my.isGoalie == GOALIE_TRUE && ( oldSnap.playMode == PLAY_MODE_FREE_KICK_US ) )
	    if( oldSnap.my.ballKickable( oldSnap.ball ) )
	      newSnap.ball.pos += newSnap.my.v;
	  break;
	  
	case TURN_COMMAND:
	  turnMoment = action.moment;
	  if( turnMoment > MAXMOMENT / 180.0 * M_PI ) turnMoment = MAXMOMENT / 180.0 * M_PI;
	  if( turnMoment < MINMOMENT / 180.0 * M_PI ) turnMoment = MINMOMENT / 180.0 * M_PI;
	  newSnap.my.bodyDir += turnMoment / newSnap.my.inertiaMoment();
	  newSnap.my.faceDir += turnMoment / newSnap.my.inertiaMoment();
	  break;
	  
	case MOVE_COMMAND:
	  // check if playMode allows a move command
	  switch( oldSnap.playMode )
	    {
	    case PLAY_MODE_FREE_KICK_US:
	      if( newSnap.my.isGoalie != GOALIE_TRUE )
		break;
	      // caution: this does not check if goalie sent a catch command!
	      if( oldSnap.my.ballKickable( oldSnap.ball ) )
		newSnap.ball.pos = action.pos + oldSnap.ball.pos - oldSnap.my.pos;
	    case PLAY_MODE_BEFORE_KICK_OFF:
	    case PLAY_MODE_GOAL_US:
	    case PLAY_MODE_GOAL_THEM:
	      newSnap.my.pos = action.pos;
	      newSnap.my.v.setXY( 0.0, 0.0 );
	      break;
	    default:
	      break;
	    }
      
	case KICK_COMMAND:
	  // is the ball within kick radius?
	  if( newSnap.my.ballKickable( newSnap.ball ) )
	    {
	      newSnap.ball.acc( radAng2Vector2d( action.power * newSnap.my.kickPowerRate( newSnap.ball ),
						 newSnap.my.bodyDir + Angle( action.moment ) ) );	       
	    }
	  break;
	  
	case CATCH_COMMAND:
	  // check if player is goalie, if player is allowed to catch again and if ball is in catchable area
	  if( newSnap.my.isGoalie == GOALIE_TRUE ) 
	    if( action.time >= newSnap.lastCatchTime + CATCH_BAN_CYCLE )
	      {
		newSnap.lastCatchTime = action.time;
		if( action.time > lastCatchTime ) 
		  lastCatchTime = action.time;
		if( newSnap.my.ballCatchable( newSnap.ball, action.moment ) )
		  {
		    newSnap.ball.v.setXY( 0.0, 0.0 );
		    newSnap.ball.pos =
		      oldSnap.my.pos + oldSnap.my.v +
		      radAng2Vector2d( oldSnap.my.size + BALL_SIZE,
				       ( newSnap.ball.pos - newSnap.my.pos ).getAngle() );
		  }
	      }
	  break;
	  
	default:
	  break;
	}
    }
  
  // step 4 : simulate movement of objects

  newSnap.my.simulate();             // simulate my movement  
  newSnap.ball.simulate();           // simulate ball movement
  for( int i = 0; i < 11; i++ )      // simulate other players' movement
    {
      newSnap.ourTeamArray[i].simulate();
      newSnap.otherTeamArray[i].simulate();
    }

  // step 5 : check for collision with ball (we do NOT check player<->player collisions)

  newSnap.collState = NONE;

  // soccerserver 7.09 algorithm

  if( newSnap.my.distTo( newSnap.ball ) < newSnap.my.size + BALL_SIZE )
    {
      Vector2d middle = ( newSnap.my.pos + newSnap.ball.pos ) * 0.5;
      Vector2d myOffset;
      Vector2d ballOffset;

      if( newSnap.my.pos == newSnap.ball.pos )
	{
	  // Here soccerserver works with a random angle

	  // new player position
	  myOffset.setRadAng( ( newSnap.my.size + BALL_SIZE ) * 0.5, 0.0 );

	  // new ball position
	  ballOffset.setRadAng( ( newSnap.my.size + BALL_SIZE ) * 0.5, PI );

	  newSnap.collState = CRITICAL;
	  COLLISIONTRACE( "player(" << newSnap.my.number << ")<->ball collision, CRITICAL" );
	}
      else
	{
	  // new player position
	  myOffset.setRadAng( ( newSnap.my.size + BALL_SIZE ) * 0.5,
			    ( newSnap.my.pos - middle ).getAngle() );

	  // new ball position
	  ballOffset.setRadAng( ( newSnap.my.size + BALL_SIZE ) * 0.5,
			    ( newSnap.ball.pos - middle ).getAngle() );

	  newSnap.collState = PROCESSED;
	  COLLISIONTRACE( "player(" << newSnap.my.number << ")<->ball collision, PROCESSED" );
	}

      // This is soccerserver implementation. It does not make sense since the ball
      // also can be kicked around in play mode GOAL_KICK_US/THEM!
      if( oldSnap.playMode != PLAY_MODE_PLAY_ON )
	{
	  newSnap.my.pos = middle + myOffset - ballOffset;

	  // this is not implemented in soccerserver 7.10 but it should be...
	  while( newSnap.my.distTo( newSnap.ball ) < newSnap.my.size + BALL_SIZE )
	    {
	      myOffset *= 1.0001;
	      ballOffset *= 1.0001;
	      newSnap.my.pos = middle + myOffset - ballOffset;
	      COLLISIONTRACE( "adjusting distance..." );
	    }
	}
      else
	{
	  newSnap.my.pos = middle + myOffset;
	  newSnap.ball.pos = middle + ballOffset;

	  // this IS implemented in soccerserver 7.10...
	  while( newSnap.my.distTo( newSnap.ball ) < newSnap.my.size + BALL_SIZE )
	    {
	      myOffset *= 1.0001;
	      newSnap.my.pos = middle + myOffset;
	      
	      ballOffset *= 1.0001;
	      newSnap.ball.pos = middle + ballOffset;
	      COLLISIONTRACE( "adjusting distance..." );
	    }
	}

      // set new player and ball velocity
      newSnap.my.v *= -0.1;
      newSnap.ball.v *= -0.1;
    }

  // step 6 : if we received a sense body, use velocity vector and head angle
  // (maybe the velocity vector could also be helpful to detect and handle collisions!)

  if( newSnap.senseBody.age == 0 && newSnap.time == time + 1 )
    {
      newSnap.my.v = newSnap.senseBody.v;
      // calculate faceDir with headAngle
      newSnap.my.faceDir = newSnap.my.bodyDir + newSnap.senseBody.headAngle;
      // Don't forget: v is relative...
      newSnap.my.v.turn( newSnap.my.faceDir );
    }

  // update coneTime for ball and players

  // check if ball is in my view cone or in my "feel radius"
  if( ( fabs( newSnap.my.faceAngleTo( newSnap.ball ) ) < newSnap.senseBody.getViewAngle() ) ||
      newSnap.my.distTo( newSnap.ball ) <= VISIBLE_DISTANCE )
    newSnap.ball.coneTime++;

  // check if players are in my view cone or in my "feel radius"
  for( int i = 0; i < 11; i++ )
    {
      if( i + 1 != newSnap.my.number )
	if( ( fabs( newSnap.my.faceAngleTo( newSnap.ourTeamArray[i] ) ) < newSnap.senseBody.getViewAngle() &&
	      newSnap.my.distTo( newSnap.ourTeamArray[i] ) < TEAM_FAR_LENGTH ) ||
	    newSnap.my.distTo( newSnap.ourTeamArray[i] ) <= VISIBLE_DISTANCE )
	  newSnap.ourTeamArray[i].coneTime++;

      if( ( fabs( newSnap.my.faceAngleTo( newSnap.otherTeamArray[i] ) ) < newSnap.senseBody.getViewAngle() &&
	    newSnap.my.distTo( newSnap.otherTeamArray[i] ) < TEAM_FAR_LENGTH ) ||
	  newSnap.my.distTo( newSnap.otherTeamArray[i] ) <= VISIBLE_DISTANCE )
	newSnap.otherTeamArray[i].coneTime++;
    }

  // copy player into team array (can be done faster)
  // don't try this for the coach!
  if( newSnap.my.number > 0 )
    newSnap.ourTeamArray[newSnap.my.number-1] = newSnap.my;

  // finally set time of new snapshot
  newSnap.time = time + 1;
}


void
Worldmodel::simSenseData( int time )
{
  // simulate sense information time -> time+1
  
  Snapshot& oldSnap = snapshots[getTimeIndex( time )];
  Snapshot& newSnap = snapshots[getTimeIndex( time + 1 )];
  
  WorldAction& action = actions[getTimeIndex( time )];
  
  // step 1 : copy old information
  
  newSnap.senseBody = oldSnap.senseBody;
  newSnap.time = time + 1;
  
  // step 2 : Birthday party
  
  newSnap.senseBody.age++;
  
  // step 3 : decrease stamina if agent desired a dash command
  
  if( action.time == time )   // is action valid?
    if( action.type == DASH_COMMAND )
      {
	if( action.power < 0 )   // dashing backward costs double stamina
	  newSnap.senseBody.stamina += 2.0 * action.power;
	else
	  newSnap.senseBody.stamina -= action.power;
	if( newSnap.senseBody.stamina < 0.0 )
	  newSnap.senseBody.stamina = 0.0;
      }

  // step 4 : increase stamina (considering effort and recovery parameters)

  if( newSnap.senseBody.stamina <= RECOVER_DEC_THR * STAMINA_MAX )
    if( newSnap.senseBody.recovery > RECOVER_MIN )
      newSnap.senseBody.recovery -= RECOVER_DEC ;
  
  if( newSnap.senseBody.stamina <= EFFORT_DEC_THR * STAMINA_MAX )
    if( newSnap.senseBody.effort > newSnap.my.effortMin )
      {
	newSnap.senseBody.effort -= EFFORT_DEC ;
	if( newSnap.senseBody.effort < newSnap.my.effortMin )
	  newSnap.senseBody.effort = newSnap.my.effortMin;
      }

  if( newSnap.senseBody.stamina >= EFFORT_INC_THR * STAMINA_MAX )
    if( newSnap.senseBody.effort < newSnap.my.effortMax )
      {
	newSnap.senseBody.effort += EFFORT_INC ;
	if( newSnap.senseBody.effort > newSnap.my.effortMax )
	  newSnap.senseBody.effort = newSnap.my.effortMax;
      }

  newSnap.senseBody.stamina += newSnap.senseBody.recovery * newSnap.my.staminaIncMax;
  if( newSnap.senseBody.stamina > STAMINA_MAX )
    newSnap.senseBody.stamina = STAMINA_MAX;
}


Snapshot *
Worldmodel::getSnapshot( int time )
{
  int    t;

  // get latest Action sent by agent (we might need that for our simulation)
  // if this action is new simulation is needed!
  if( getAction() )
    newVisAudData = newSenseData = true;

  int    minTime = min( visAudTime, senseTime );

  // desired time too far back into the past
  if( time < maxTime - numSnapshots + 1 ) return NULL;
  
  // no sensor data received at all
  if( ( visAudTime < 0 ) && ( senseTime < 0 ) ) return NULL;
  
  // desired time too far away from minTime. The largest time allowed depends
  // on the size of the "snapshots" array (i.e. numSnapshots)
  if( time - minTime > numSnapshots ) return NULL;
  
  // simulation of sense information (from senseTime to time)
  if( newVisAudData || newSenseData || time > simSenseTime )
    {
      if( senseTime >= 0 )
	for( t = senseTime; t < time; t++ )
	  simSenseData( t );
      else
	snapshots[getTimeIndex( time )].setSenseInvalid();   // no sense information available
      simSenseTime = max( time, senseTime );
      newSenseData = false;
    }

  // simulation of visual information (from visAudTime to time)
  if( newVisAudData || newSenseData || time > simVisAudTime )
    {
      if( visAudTime >= 0 )
	for( t = visAudTime; t < time; t++ )
	  simVisualData( t );
      else
	snapshots[getTimeIndex( time )].setSeeInvalid();   // no visual information available
      simVisAudTime = max( time, visAudTime );
      newVisAudData = false;
    }

  if( time > maxTime ) maxTime = time;

  return &snapshots[getTimeIndex( time )];
}


bool
Worldmodel::adjust()
{
  SensorData        data;              // new data received from soccerserver
  bool              newData = false;   // value of adjust()
  int               i = 0;             // counter
  int               type;              // submitted new player type
  WorldPlayerType   newType;           // for a player change

  // as long as there is data in the buffer - get it

  while ( comm->getNextData( data ) )
    {
      Snapshot *snap = &snapshots[0];
      if( data.time != UNKNOWN )
	snap = &snapshots[getTimeIndex( data.time )];

      switch( data.type ) 
	{

	case SENSOR_TYPE_PLAYER_TYPE :

	  type = data.playerType.player_type;
	  if( type >= 0 && type < playerParams.player_types )
	    {
	      for( i = 0; i < numSnapshots; i++ )
		{
		  snapshots[i].playerType[type].state          = type;

		  snapshots[i].playerType[type].playerSpeedMax = data.playerType.player_speed_max;
		  snapshots[i].playerType[type].playerDecay    = data.playerType.player_decay;
		  snapshots[i].playerType[type].playerSize     = data.playerType.player_size;
		  snapshots[i].playerType[type].staminaIncMax  = data.playerType.stamina_inc_max;
		  snapshots[i].playerType[type].inertiaFactor  = data.playerType.inertia_moment;
		  snapshots[i].playerType[type].dashPowerRate  = data.playerType.dash_power_rate;
		  snapshots[i].playerType[type].kickableMargin = data.playerType.kickable_margin;
		  snapshots[i].playerType[type].kickRand       = data.playerType.kick_rand;
		  snapshots[i].playerType[type].extraStamina   = data.playerType.extra_stamina;
		  snapshots[i].playerType[type].effortMin      = data.playerType.effort_min;
		  snapshots[i].playerType[type].effortMax      = data.playerType.effort_max;
		}
	      newData = true;
	    }
	  break;

	case SENSOR_TYPE_PLAYER_CHANGED :

	  if( data.heteroType == UNKNOWN )
	    newType = snapshots[0].playerType[DEFAULT_TYPE];
	  else
	    newType = snapshots[0].playerType[data.heteroType];

	  if( data.ourTeamChanged )
	    {
	      int numSubs = snapshots[0].numOurSubs;

	      if( newType.state == UNKNOWN )
		cerr << "Nr. " << snapshots[0].my.number
		     << " : Player " << data.number << " was changed to type " 
		     << data.heteroType << ", but I don't know that type!" << endl;

	      for( i = 0; i < numSnapshots; i++ )
		{
		  snapshots[i].ourTeamArray[data.number - 1].type           = data.heteroType;
		  snapshots[i].ourTeamArray[data.number - 1].vmax           = newType.playerSpeedMax;
		  snapshots[i].ourTeamArray[data.number - 1].decay          = newType.playerDecay;
		  snapshots[i].ourTeamArray[data.number - 1].size           = newType.playerSize;
		  snapshots[i].ourTeamArray[data.number - 1].staminaIncMax  = newType.staminaIncMax;
		  snapshots[i].ourTeamArray[data.number - 1].inertiaFactor  = newType.inertiaFactor;
		  snapshots[i].ourTeamArray[data.number - 1].dashPowerRate  = newType.dashPowerRate;
		  snapshots[i].ourTeamArray[data.number - 1].kickableMargin = newType.kickableMargin;
		  snapshots[i].ourTeamArray[data.number - 1].kickRand       = newType.kickRand;
		  snapshots[i].ourTeamArray[data.number - 1].extraStamina   = newType.extraStamina;
		  snapshots[i].ourTeamArray[data.number - 1].effortMin      = newType.effortMin;
		  snapshots[i].ourTeamArray[data.number - 1].effortMax      = newType.effortMax;

		  if( snapshots[i].my.number == data.number )
		    {
		      snapshots[i].my.type           = data.heteroType;
		      snapshots[i].my.vmax           = newType.playerSpeedMax;
		      snapshots[i].my.decay          = newType.playerDecay;
		      snapshots[i].my.size           = newType.playerSize;
		      snapshots[i].my.staminaIncMax  = newType.staminaIncMax;
		      snapshots[i].my.inertiaFactor  = newType.inertiaFactor;
		      snapshots[i].my.dashPowerRate  = newType.dashPowerRate;
		      snapshots[i].my.kickableMargin = newType.kickableMargin;
		      snapshots[i].my.kickRand       = newType.kickRand;
		      snapshots[i].my.extraStamina   = newType.extraStamina;
		      snapshots[i].my.effortMin      = newType.effortMin;
		      snapshots[i].my.effortMax      = newType.effortMax;
		    }

		  snapshots[i].numOurSubs++;
		  snapshots[i].ourSubs[numSubs % ( playerParams.subs_max + 11 )].number = data.number;
		  snapshots[i].ourSubs[numSubs % ( playerParams.subs_max + 11 )].time   = data.time;
		}
	    }
	  else
	    {
	      int numSubs = snapshots[0].numOtherSubs;

	      for( i = 0; i < numSnapshots; i++ )
		{
		  snapshots[i].otherTeamArray[data.number - 1].type           = data.heteroType;
		  snapshots[i].otherTeamArray[data.number - 1].vmax           = newType.playerSpeedMax;
		  snapshots[i].otherTeamArray[data.number - 1].decay          = newType.playerDecay;
		  snapshots[i].otherTeamArray[data.number - 1].size           = newType.playerSize;
		  snapshots[i].otherTeamArray[data.number - 1].staminaIncMax  = newType.staminaIncMax;
		  snapshots[i].otherTeamArray[data.number - 1].inertiaFactor  = newType.inertiaFactor;
		  snapshots[i].otherTeamArray[data.number - 1].dashPowerRate  = newType.dashPowerRate;
		  snapshots[i].otherTeamArray[data.number - 1].kickableMargin = newType.kickableMargin;
		  snapshots[i].otherTeamArray[data.number - 1].kickRand       = newType.kickRand;
		  snapshots[i].otherTeamArray[data.number - 1].extraStamina   = newType.extraStamina;
		  snapshots[i].otherTeamArray[data.number - 1].effortMin      = newType.effortMin;
		  snapshots[i].otherTeamArray[data.number - 1].effortMax      = newType.effortMax;

		  snapshots[i].numOtherSubs++;
		  snapshots[i].otherSubs[numSubs % ( playerParams.subs_max + 11 )].number = data.number;
		  snapshots[i].otherSubs[numSubs % ( playerParams.subs_max + 11 )].time   = data.time;
		}
	    }

	  newVisAudData = newSenseData = true;
	  newData = true;
	  break;

	case SENSOR_TYPE_SEE:      
	  
	  // only care about newer data
	  if( data.time >= visAudTime )
	    {
	      // simulate world until data.time, then update with newer information
	      getSnapshot( data.time );
	      
	      snap->playMode = comm->getPlayMode();
	      snap->ourSide = comm->getMySide();
	      snap->ourScore = comm->getOurScore();
	      snap->otherScore = comm->getOtherScore();

	      if( comm->amIGoalie() && snap->my.isGoalie != GOALIE_TRUE )
		{
		  snap->ourGoalieNo = comm->getMyNumber();
		  for( int j = 0; j < 11; j++ )
		    if( j != comm->getMyNumber() - 1 )
		      snap->ourTeamArray[j].isGoalie = GOALIE_FALSE;
		}
	      snap->my.isGoalie = comm->amIGoalie() ? GOALIE_TRUE : GOALIE_FALSE;

	      if( snap->ball.age == 1 )
		{
		  // report ball relpos from one step ago for the calculation
		  // of ball velocity with position difference
		  snap->adjustSee( data,
				   snapshots[getTimeIndex( data.time - 1 )].ball.pos -
				   snapshots[getTimeIndex( data.time - 1 )].my.pos );
		}
	      else
		snap->adjustSee( data );
	      
	      // try to keep up with headAngle changes
	      // this works because we always process senseBody messages first
	      snap->my.bodyDir = snap->my.faceDir - snap->senseBody.headAngle;
	      
	      // copy player into team array (can be done faster)
	      // don't try this for the coach!
	      if( snap->my.number > 0 )
		snap->ourTeamArray[snap->my.number-1] = snap->my;
  
	      snap->collState = NONE;   // just forget simulation - we have got visual data!

	      snap->visualTime = data.time;
	      visualTime = data.time;
	      visAudTime = data.time;

	      newVisAudData = true;
	      newData = true;
	    }
	  break;
	  
	case SENSOR_TYPE_SENSE_BODY:
	  
	  // only care about new data
	  if ( data.time >= senseTime )
	    {
	      // simulate senseBody-informaton until data.time, then update with newer information
	      // simulation is needed for recovery parameter only (and to fill in missing snapshots)
	      getSnapshot( data.time );
	      
	      snap->playMode = comm->getPlayMode();
	      snap->ourSide = comm->getMySide();
	      snap->ourScore = comm->getOurScore();
	      snap->otherScore = comm->getOtherScore();
	      
	      if( comm->amIGoalie() && snap->my.isGoalie != GOALIE_TRUE )
		{
		  snap->ourGoalieNo = comm->getMyNumber();
		  for( int j = 0; j < 11; j++ )
		    if( j != comm->getMyNumber() - 1 )
		      snap->ourTeamArray[j].isGoalie = GOALIE_FALSE;
		}
	      snap->my.isGoalie = comm->amIGoalie() ? GOALIE_TRUE : GOALIE_FALSE;
	      
	      // update ball (only does something in standard situations in this case!)
	      snap->adjustBall( data );

	      snap->adjustSense( data );

	      snap->my.faceDir = snap->my.bodyDir + Angle( data.head_angle );
	      
	      // update player velocity (caution: velocity is relative!)
	      snap->my.v = radAng2Vector2d( data.speed, Angle( data.speed_angle ) + snap->my.faceDir );
	      
	      // copy player into team array (can be done faster)
	      // don't try this for the coach!
	      if( snap->my.number > 0 )
		snap->ourTeamArray[snap->my.number-1] = snap->my;
	      
	      senseTime = data.time;

	      newSenseData = true;
	      newData = true;
	    }
	  break;
	  
	case SENSOR_TYPE_EXACT:   // for the coach
	  
	  // only care about newer data
	  if( data.time >= visAudTime )
	    {
	      //simulate world until data.time, then overwrite with exact information
	      getSnapshot( data.time );
	      
	      snap->playMode = comm->getPlayMode();
	      snap->ourSide = comm->getMySide();
	      snap->ourScore = comm->getOurScore();
	      snap->otherScore = comm->getOtherScore();
	      
	      // update ball (only does something in standard situations in this case!)
	      snap->adjustBall( data );

	      snap->adjustExact( data );
	      
	      snap->visualTime = data.time;
	      visualTime = data.time;
	      visAudTime = data.time;
	      
	      // As the coach never receives sense_body, we set them invalid.
	      // But we have to set senseTime up, so snapshots will be 
	      // returned nevertheless.
	      senseTime = data.time;
	      snap->setSenseInvalid();

	      newVisAudData = true;
	      newSenseData = true;
	      newData = true;
	    }
	  break;
	  
	default:
	  break;
	}
    }

  return newData;
}


bool
Worldmodel::feedbackAudio( AudioMsg &audioMsg )
{
  bool ok;
  audioMsg >> ok;
  
  if( ok )
    {
      Snapshot audioSnap;
      audioMsg >> audioSnap;

      AUDIALINPUTTRACE( "Audial information from no. " << audioMsg.getSource() << " :" << endl );

      ok |= updateWithSnapshot( audioSnap, audioMsg.getSource() );
    }

  return ok;
}


bool
Worldmodel::updateWithSnapshot( Snapshot audioSnap, int sender )
{
  int   timeDiff   = visAudTime - audioSnap.time;
  bool  changed    = false;   // true if any data was updated
  bool  feltPlayer = false;   // did I "feel" players around me?

  if( timeDiff > 0 )
    {
      // simulate audio information until visAudTime
      if( audioSnap.ball.age < ETERNAL )
	{
	  for( int i = 0; i < timeDiff; i++ )
	    audioSnap.ball.simulate();
	  audioSnap.ball.age += timeDiff;
	}
      for( int i = 0; i < 11; i++ )
	{
	  if( audioSnap.ourTeamArray[i].age < ETERNAL )
	    audioSnap.ourTeamArray[i].age += timeDiff;
	  if( audioSnap.otherTeamArray[i].age < ETERNAL )
	    audioSnap.otherTeamArray[i].age += timeDiff;
	}
    }

  // get the snapshot that is compared with audioSnap
  Snapshot& worldSnap = snapshots[getTimeIndex( max( visAudTime, audioSnap.time ) )];

  // check for felt players
  for( int i = 0; i < 11; i++ )
    {
      if( i != worldSnap.my.number - 1 )
	if( dist( worldSnap.ourTeamArray[i].pos, worldSnap.my.pos ) <= VISIBLE_DISTANCE )
	  {
	    feltPlayer = true;
	    break;
	  }
      if( dist( worldSnap.otherTeamArray[i].pos, worldSnap.my.pos ) <= VISIBLE_DISTANCE )
	{
	  feltPlayer = true;
	  break;
	}		
    }

  AUDIALINPUTTRACE( "Using friends : " );
  
  // update players of my team
  for( int i = 0; i < 11; i++ )
    {
      // only update with valid information
      if( audioSnap.ourTeamArray[i].age >= ETERNAL )
	continue;

      // never update player himself
      if( i == worldSnap.my.number - 1 )
	continue;

      // use of audial information depends on its quality and age
      bool updatePlayer = false;
      bool checkAge = false;
      WorldObjectState state = UNSAFE;

      switch( worldSnap.ourTeamArray[i].state )
	{
	case UNSAFE :
	  switch( audioSnap.ourTeamArray[i].state )
	    {
	    case UNSAFE : updatePlayer = true;  checkAge = true;  state = UNSAFE; break;
	    case FUZZY  : updatePlayer = true;  checkAge = false; state = FUZZY;  break;
	    case VISUAL : updatePlayer = true;  checkAge = false; state = AUDIAL; break;
	    case AUDIAL : updatePlayer = true;  checkAge = false; state = AUDIAL; break;
	    }
	  break;

	case FUZZY :
	  switch( audioSnap.ourTeamArray[i].state )
	    {
	    case UNSAFE : updatePlayer = false; checkAge = false; state = UNSAFE; break;
	    case FUZZY  : updatePlayer = true;  checkAge = true;  state = FUZZY;  break;
	    case VISUAL : updatePlayer = true;  checkAge = false; state = AUDIAL; break;
	    case AUDIAL : updatePlayer = true;  checkAge = false; state = AUDIAL; break;
	    }
	  break;

	case VISUAL :
	  switch( audioSnap.ourTeamArray[i].state )
	    {
	    case UNSAFE : updatePlayer = false; checkAge = false; state = UNSAFE; break;
	    case FUZZY  : updatePlayer = true;  checkAge = true;  state = FUZZY;  break;
	    case VISUAL : updatePlayer = true;  checkAge = true;  state = AUDIAL; break;
	    case AUDIAL : updatePlayer = true;  checkAge = true;  state = AUDIAL; break;
	    }
	  break;

	case AUDIAL :
	  switch( audioSnap.ourTeamArray[i].state )
	    {
	    case UNSAFE : updatePlayer = false; checkAge = false; state = UNSAFE; break;
	    case FUZZY  : updatePlayer = true;  checkAge = true;  state = FUZZY;  break;
	    case VISUAL : updatePlayer = true;  checkAge = true;  state = AUDIAL; break;
	    case AUDIAL : updatePlayer = true;  checkAge = true;  state = AUDIAL; break;
	    }
	  break;
	}

      // if player is very close I should have felt him (player might be my shadow!)
      if( updatePlayer )
	if( dist( audioSnap.ourTeamArray[i].pos, worldSnap.my.pos ) <= VISIBLE_DISTANCE )
	  if( !feltPlayer )
	    updatePlayer = false;

      // if I felt that player I have to be careful with updates
      if( updatePlayer )
	if( dist( worldSnap.ourTeamArray[i].pos, worldSnap.my.pos ) <= VISIBLE_DISTANCE )
	  checkAge = true;

      // update player if age is not to be checked or he is young enough
      if( updatePlayer )
	if( ( checkAge && audioSnap.ourTeamArray[i].age < worldSnap.ourTeamArray[i].age ) ||
	    ( !checkAge && audioSnap.ourTeamArray[i].age - 5 < worldSnap.ourTeamArray[i].age ) )
	  {
	    worldSnap.ourTeamArray[i].pos = audioSnap.ourTeamArray[i].pos;
	    worldSnap.ourTeamArray[i].age = audioSnap.ourTeamArray[i].age;
	    worldSnap.ourTeamArray[i].state = state;
	    worldSnap.ourTeamArray[i].coneTime = 0;

	    changed = true;
	    AUDIALINPUTTRACE( i + 1 << " " );
	  }
    }
  
  AUDIALINPUTTRACE( endl << "Using opponents : " );

  // update players of opponent team
  for( int i = 0; i < 11; i++ )
    {
      // in case of a coach message: update player types (in all snapshots!)
      if( sender == 0 )
	for( int j = 0; j < numSnapshots; j++ )
	  snapshots[j].otherTeamArray[i].type = audioSnap.otherTeamArray[i].type;

      // only update with valid information
      if( audioSnap.otherTeamArray[i].age >= ETERNAL )
	continue;

      // use of audial information depends on its quality and age
      bool updatePlayer = false;
      bool checkAge = false;
      WorldObjectState state = UNSAFE;

      switch( worldSnap.otherTeamArray[i].state )
	{
	case UNSAFE :
	  switch( audioSnap.otherTeamArray[i].state )
	    {
	    case UNSAFE : updatePlayer = true;  checkAge = true;  state = UNSAFE; break;
	    case FUZZY  : updatePlayer = true;  checkAge = false; state = FUZZY;  break;
	    case VISUAL : updatePlayer = true;  checkAge = false; state = AUDIAL; break;
	    case AUDIAL : updatePlayer = true;  checkAge = false; state = AUDIAL; break;
	    }
	  break;

	case FUZZY :
	  switch( audioSnap.otherTeamArray[i].state )
	    {
	    case UNSAFE : updatePlayer = false; checkAge = false; state = UNSAFE; break;
	    case FUZZY  : updatePlayer = true;  checkAge = true;  state = FUZZY;  break;
	    case VISUAL : updatePlayer = true;  checkAge = false; state = AUDIAL; break;
	    case AUDIAL : updatePlayer = true;  checkAge = false; state = AUDIAL; break;
	    }
	  break;

	case VISUAL :
	  switch( audioSnap.otherTeamArray[i].state )
	    {
	    case UNSAFE : updatePlayer = false; checkAge = false; state = UNSAFE; break;
	    case FUZZY  : updatePlayer = true;  checkAge = true;  state = FUZZY;  break;
	    case VISUAL : updatePlayer = true;  checkAge = true;  state = AUDIAL; break;
	    case AUDIAL : updatePlayer = true;  checkAge = true;  state = AUDIAL; break;
	    }
	  break;

	case AUDIAL :
	  switch( audioSnap.otherTeamArray[i].state )
	    {
	    case UNSAFE : updatePlayer = false; checkAge = false; state = UNSAFE; break;
	    case FUZZY  : updatePlayer = true;  checkAge = true;  state = FUZZY;  break;
	    case VISUAL : updatePlayer = true;  checkAge = true;  state = AUDIAL; break;
	    case AUDIAL : updatePlayer = true;  checkAge = true;  state = AUDIAL; break;
	    }
	  break;
	}

      // if player is very close I should have felt him (player might be my shadow!)
      if( updatePlayer )
	if( dist( audioSnap.otherTeamArray[i].pos, worldSnap.my.pos ) <= VISIBLE_DISTANCE )
	  if( !feltPlayer )
	    updatePlayer = false;

      // if I felt that player I have to be careful with updates
      if( updatePlayer )
	if( dist( worldSnap.otherTeamArray[i].pos, worldSnap.my.pos ) <= VISIBLE_DISTANCE )
	  checkAge = true;

      // update player if age is not to be checked or he is young enough
      if( updatePlayer )
	if( ( checkAge && audioSnap.otherTeamArray[i].age < worldSnap.otherTeamArray[i].age ) ||
	    ( !checkAge && audioSnap.otherTeamArray[i].age - 5 < worldSnap.otherTeamArray[i].age ) )
	  {
	    worldSnap.otherTeamArray[i].pos = audioSnap.otherTeamArray[i].pos;
	    worldSnap.otherTeamArray[i].age = audioSnap.otherTeamArray[i].age;
	    worldSnap.otherTeamArray[i].state = state;
	    worldSnap.otherTeamArray[i].coneTime = 0;

	    changed = true;
	    AUDIALINPUTTRACE( i + 1 << " " );
	  }
    }

  AUDIALINPUTTRACE( endl );

  // update ball if audial information is younger than my visual information
  // CAUTION: ball.v may be bad if sending player is far away from the ball
  if( audioSnap.ball.age < worldSnap.ball.age )
    {
      // be careful: do not overwrite good information with audio information
      // from players that are far from the ball
      if( audioSnap.ball.age < worldSnap.ball.age - 3 ||
	  sender == 0 ||   // coach!
	  sqrnorm( audioSnap.ourTeam( sender ).pos - worldSnap.ball.pos ) <
	  sqrnorm( worldSnap.my.pos - worldSnap.ball.pos ) )
	{
	  worldSnap.ball.pos = audioSnap.ball.pos;
	  worldSnap.ball.age = audioSnap.ball.age;
	  worldSnap.ball.coneTime = 0;

	  // if the audio ball is FUZZY the velocity information might be bad
	  // (ball was only felt by sending player)
	  if( audioSnap.ball.state != FUZZY )
	    {
	      worldSnap.ball.state = AUDIAL;   // we can trust in the audio information
	      worldSnap.ball.v = audioSnap.ball.v;
	    }

	  changed = true;
	  AUDIALINPUTTRACE( "Using Ball; pos = " << worldSnap.ball.pos << " , v = " << worldSnap.ball.v << endl );
	}
    }

  // update goalie numbers (if sending player knows them but we do not)
  if( audioSnap.ourGoalieNo != UNKNOWN && worldSnap.ourGoalieNo == UNKNOWN )
    {
      worldSnap.ourGoalieNo = audioSnap.ourGoalieNo;
      for( int i = 0; i < 11; i++ )
	if( i == worldSnap.ourGoalieNo - 1 )
	  worldSnap.ourTeamArray[i].isGoalie = GOALIE_TRUE;
	else
	  worldSnap.ourTeamArray[i].isGoalie = GOALIE_FALSE;
    }
  if( audioSnap.otherGoalieNo != UNKNOWN && worldSnap.otherGoalieNo == UNKNOWN )
    {
      worldSnap.otherGoalieNo = audioSnap.otherGoalieNo;
      for( int i = 0; i < 11; i++ )
	if( i == worldSnap.otherGoalieNo - 1 )
	  worldSnap.otherTeamArray[i].isGoalie = GOALIE_TRUE;
	else
	  worldSnap.otherTeamArray[i].isGoalie = GOALIE_FALSE;
    }

  if( changed && audioSnap.time > visAudTime )
    visAudTime = audioSnap.time;

  if( changed )
    {
      newVisAudData = true;
      newSenseData = true;
    }

  // maxTime = audioSnap.time;

  return changed;
}


bool
Worldmodel::getSayMsg( AudioMsg &audioMsg )
{
  Snapshot audioSnap;
  Snapshot& worldSnap = snapshots[getTimeIndex( maxTime )];
  bool ok = true;

  AUDIALOUTPUTTRACE( "Sending worldmodel : " );

  // send ball
  audioSnap.ball = worldSnap.ball;

  AUDIALOUTPUTTRACE( "ball with pos = " << audioSnap.ball.pos << " , v = " << audioSnap.ball.v << endl );
  
  // send teams
  for( int i = 0; i < 11; i++ )
    {
      audioSnap.ourTeamArray[i] = worldSnap.ourTeamArray[i];
      audioSnap.otherTeamArray[i] = worldSnap.otherTeamArray[i];
    }

  audioSnap.ourGoalieNo = worldSnap.ourGoalieNo;
  audioSnap.otherGoalieNo = worldSnap.otherGoalieNo;

  audioMsg.markWrite();
  audioMsg << true;
  audioMsg << audioSnap;
      
  // check if error occurred
  if( !audioMsg.getState() )
    {
      audioMsg.rollbackWrite().clearState();
      audioMsg << false;
      ok = false;
      AUDIALOUTPUTTRACE( "Didn't send anything!" << endl );
    }
  return ok;
}
