//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
//
// MRB player module ballhandling
//
// $Id: ballhandlingmodule.h,v 1.67 2001/08/06 02:56:13 mrb Exp $
// $manager: fel_und_eri$
//
// handles all situations in which a more or less responsible player
// or goalie has to go for the ball or to keep the ball to himself
//
//////////////////////////////////////////////////////////////////////


#ifndef __BALL_HANDLING_MODULE_H__
#define __BALL_HANDLING_MODULE_H__


#include "grade.h"

#include "worldmodel.h"
#include "tacticalmodule.h"
#include "audiomgr.h"
#include "AdvSkills.h"

/*
  
  short description:

  When called for feedback(), snap and tsnap are stored locally and some
  auxiliary variables (my, ball, etc) are set. A timestamp of snap and tsnap 
  is saved to be compared against when called later for eval() and act()

  In evaluate() we first test for some simple special cases, and then call the
  more complex evaluatePlayer() and evaluateGoalie() routines. They do the rest
  of the evaluation, and make use of the further singled out 
  dribble evaluation. At the end of the eval() method, the variables action and
  g (for grade) have to be set to the desired values to be returnd to 
  master control. Also take care to set any local temporary variables (e.g. 
  for values already computed for the later execution of the command). All
  this is implemented in ballhandlingEvaluate.C.

  In act() we split up the different cases in a switch{}. Auxiliary routines
  then do the rest of the action execution in most cases. All this is 
  implemented in ballhandlingAct.C. Some simulation routines and specific 
  turnHead routines (taking care of the players current action) are 
  included in this file as well.

  Things that need some more serious thinking are marked with [!] (or more
  !!s for urgency).

  Known bugs:

  major ToDo:

  - see the [!..!]s

  minor ToDo:
  
*/


// dribble directions

const Vector2d LEFT                  = radAng2Vector2d( 1.0, -M_PI/2.0 );
const Vector2d LEFT_LEFT_FORWARD     = radAng2Vector2d( 1.0, -3.0*M_PI/8.0 );
const Vector2d LEFT_FORWARD          = radAng2Vector2d( 1.0, -M_PI/4.0 );
const Vector2d FORWARD_LEFT_FORWARD  = radAng2Vector2d( 1.0, -M_PI/8.0 );
const Vector2d FORWARD               = radAng2Vector2d( 1.0, 0.0 );
const Vector2d FORWARD_RIGHT_FORWARD = radAng2Vector2d( 1.0, M_PI/8.0 );
const Vector2d RIGHT_FORWARD         = radAng2Vector2d( 1.0, M_PI/4.0 );
const Vector2d RIGHT_RIGHT_FORWARD   = radAng2Vector2d( 1.0, 3.0*M_PI/8.0 );
const Vector2d RIGHT                 = radAng2Vector2d( 1.0, M_PI/2.0 );
const Vector2d RIGHT_RIGHT_BACK      = radAng2Vector2d( 1.0, 5.0*M_PI/8.0 );
const Vector2d RIGHT_BACK            = radAng2Vector2d( 1.0, 3.0*M_PI/4.0 );
const Vector2d BACK_RIGHT_BACK       = radAng2Vector2d( 1.0, 7.0*M_PI/8.0 );
const Vector2d BACK                  = radAng2Vector2d( 1.0, M_PI );
const Vector2d BACK_LEFT_BACK        = radAng2Vector2d( 1.0, -7.0*M_PI/8.0 );
const Vector2d LEFT_BACK             = radAng2Vector2d( 1.0, -3.0*M_PI/4.0 );
const Vector2d LEFT_LEFT_BACK        = radAng2Vector2d( 1.0, -5.0*M_PI/8.0 );


class BallHandlingModuleInfo; // forward decl. needed


/////////////////////////////////////////////////////////////////////
//
// Class Ballhandling Module
//
/////////////////////////////////////////////////////////////////////


class BallHandlingModule {
  
 public:
  
  BallHandlingModule(bool _goalie);
  
  ~BallHandlingModule();



  /// feedback of communicated information
  bool feedbackWorld( const Snapshot* snap, 
		      TacticalSnapshot* tsnap );
  
  bool feedbackAudio( AudioMsg &audioMsg );
  
  /// evaluation of given situation into a Grade
  Grade evaluate( AdvSkills* compat_skills, const Snapshot *compat_snap, 
		  TacticalSnapshot* compat_tsnap);

  /// Decide whether a turn head should be done.
  Grade turnHeadEvaluate(AdvSkills* compat_skills,
			 AdvSkills::TurnHeadCommand &thc);
    
  // perform action according to last evaluation
  bool act( AdvSkills* skills, const Snapshot* compat_snap, 
	    TacticalSnapshot* compat_tsnap);
  
  /// pipe information to be communicated into the communication stream
  bool getSayMsg(AudioMsg &audioMsg);

 private:
  
  // action decision
  ///////////////////////////

  enum planned_action {
    
    WAIT = 0,            // turn perpetual to ball direction
    FIND_BALL,           // turn to get ball in sight
    
    GOALIE_INTERCEPT,    // go for a rolling ball (uses also player intercept)
    GOALIE_CATCH,        // catch ball
    GOALIE_DRIBBLE,      // dribble to penalty area
    GOALIE_PLACE,        // keep ball away from opponent

    PLAYER_INTERCEPT,    // go for a rolling ball
    PLAYER_INTERCEPT_ALT,// go for alternate icpt position
    PLAYER_FAST_DRIBBLE, // kick ball ahead a few steps
    PLAYER_DRIBBLE,      // keep ball close, but move
    PLAYER_PLACE,        // just keep ball away from opponent
  };

  // the decision made in evaluate(), to be executed in act()
  planned_action action;
  // grade of this action
  Grade grade;
  // if action actually was executed
  bool  acted;

  ///////////////////////////////////////////////////////////////////////////
  // local data
  const Snapshot   *snap;
  TacticalSnapshot *tsnap;

  int         time;

  play_mode   playMode;

  WorldPlayer my;
  double      stamina;          // total
  double      availableStamina; // total - recovery threshold
  WorldBall   ball;
  double      distToBall;
  Angle       faceAngleToBall;
  Angle       bodyAngleToBall;
  WorldObject ourGoal;

  bool                    goalie;   

  bool                    sendSomething; // send flag
  BallHandlingModuleInfo* sendInfo;      // send info

  int  lastKickTime;
  bool opponentKicked;
  bool findBallTurnedBody;   // find ball was called AND sen a turn command

  ///////////////////////////////////////////////////////////////////////////

  // eval methods
  Grade evaluateGoalie();

  Grade evaluatePlayer();

  Grade evaluateDribble(); 

  Grade goalieEvaluateDribble(); 

  void updateTA(TacticalActionState ta);

  // tacticalAction blocking
  int lastReportTime;

  // intercept
  AdvSkills::TurtleCommand turtleCmd;

  // findBall/turnHead
  bool  lookingForBall;
  Angle turnAngle; // angle by which to turn for findBall
  Grade turnHeadGrade;
  view_width reqWidth;

  // checks if ball is too old
  Grade evaluateFindBall();
  // turns to look for the ball
  void   findBall();
  
  // intercept
  /////////////////////////////////////
  
  // intercepts the freely rolling ball at the computed intercept point
  // (used for player AND goalie)
  void   playerIntercept();

  // dribble
  /////////////////////////////////////
  double fd_dashPower;
  int    fd_numSteps;
  double fd_dist;

  Angle dribbleDir; // for fast dribble and dribble
  Angle holdDir;    // for fast dribble and dribble
  Angle placeDir;   // for place
  Angle kickDir;    // for kickaway
  double dribbleStamina;

  struct DribbleAreaData {
    int      numDirs;
    struct {
      Vector2d dir;
      Grade    grade;
      bool     dribble;
    } dirs[9];

    Angle    prefHeadDir;
  };

  static const DribbleAreaData areas[];

  // test current ball position for the area code
  int  dribbleArea();
  // check if that point is valid for (fast)dribble
  bool checkDribbleDir(Vector2d target, double secDist = 5.0); 
  // kick ball ahead slightly
  void fastdribble();
  // and keep ball under control at all times
  void dribble();
  // move ball around ourselves away from opponent
  void place();

  // goalie catch
  /////////////////////////////////////
  // catch the ball immediately
  void goalieCatch();

  AdvSkills::CatchCommand catchParams;
};

//////////////////////////////////////////////////////////////////////
//
// Class to keep information for modules of other agents.
//
//////////////////////////////////////////////////////////////////////
class BallHandlingModuleInfo {

  // internal communication between ballhandlingModules of different players
  // of our team is not used yet.
  
 public:

  BallHandlingModuleInfo();
  ~BallHandlingModuleInfo();
  
};

/** put information to AudioMsg-Stream */
class AudioMsg& operator<< (AudioMsg &aumsg, const BallHandlingModuleInfo &info);

/** fill with information from AudioMsg-Stream */
class AudioMsg& operator>> (AudioMsg &aumsg, BallHandlingModuleInfo &info);

#endif












