//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#include "tacticalmodule.h"
#include "tacticalmoduleinfo.C"   // implementation of class TacticalModuleInfo

#include <string>
#include <strstream.h>

// output for the coach's player type detection
//#define USETYPEDETECTTRACE

// output after determining responsible players and most dangerous opponents
//#define RESPONSIBLETRACE(output) cout << snap->time << output << endl

// output for simulation of opponent stamina
//#define STAMSIMTRACE(output) cout << snap->time << output << endl


#ifndef USETYPEDETECTTRACE
#define TYPEDETECTTRACE(output)
#else
#define TYPEDETECTTRACE(output) { \
 ostrstream ost; \
 ost << "TM typedetect:" << output << (char)0; \
 Comm::sendMCmd(MLog(ost.str())); } 
#warning TYPEDETECTTRACE debug trace TYPEDETECTTRACE is activated!
#endif

#ifndef RESPONSIBLETRACE
#define RESPONSIBLETRACE(output)
#else
#warning RESPONSIBLETRACE debug trace RESPONSIBLETRACE is activated!
#endif

#ifndef STAMSIMTRACE
#define STAMSIMTRACE(output)
#else
#warning STAMSIMTRACE debug trace STAMSIMTRACE is activated!
#endif


//////////////////////////////////////////////////////////////////////
//
// Class TacticalSystem
//
//////////////////////////////////////////////////////////////////////

TacticalSystem::TacticalSystem( int _numDef, int _numDefMid, int _numOffMid, int _numOff )
{
  ensure( _numDef >= 0 && _numDefMid >= 0 && _numOffMid >= 0 && _numOff >= 0 );
  ensure( _numDef + _numDefMid + _numOffMid + _numOff == 10 );

  numDefenders   = _numDef;
  numDefMidfield = _numDefMid;
  numOffMidfield = _numOffMid;
  numMidfield    = _numDefMid + _numOffMid;
  numOffenders   = _numOff;
}

//////////////////////////////////////////////////////////////////////
//
// Class TacticalAction
//
//////////////////////////////////////////////////////////////////////

ostream &operator << (ostream &os, const TacticalAction &a)
{
  char *state;
  switch (a.state) {
  case TA_CONTROL:
    state="CONTROL";
    break;
  case TA_PASS:
    state="PASS";
    break;
  case TA_GOALSHOT:
    state="GOALSHOT";
    break;
  case TA_REACT:
    state="REACT";
    break;
  case TA_UNKNOWN:
    state="UNKNOWN";
    break;
  default:
    state="UNDEFINED";
  }
  
  return os << "(Action st=" << state << " start= " 
	    << a.startingTime << " perf= " << a.performer
	    << " recv= " << a.receiver << ")";
}

//////////////////////////////////////////////////////////////////////
//
// Class TacticalSnapshot
//
//////////////////////////////////////////////////////////////////////

TacticalSnapshot::TacticalSnapshot()
{
  time = 0;

  ourBallControl = BC_UNKNOWN;

  myTacticalNo = UNKNOWN;
  ourGoalieFound = false;
  otherGoalieFound = false;

  ourSystem.numDefenders     = Config::getValueAsInt( "ourDefenders" );
  ourSystem.numDefMidfield   = Config::getValueAsInt( "ourDefMidfield" );
  ourSystem.numOffMidfield   = Config::getValueAsInt( "ourOffMidfield" );
  ourSystem.numMidfield      = ourSystem.numDefMidfield + ourSystem.numOffMidfield;
  ourSystem.numOffenders     = Config::getValueAsInt( "ourOffenders" );

  otherSystem.numDefenders   = Config::getValueAsInt( "otherDefenders" );
  otherSystem.numDefMidfield = Config::getValueAsInt( "ourDefMidfield" );
  otherSystem.numOffMidfield = Config::getValueAsInt( "ourOffMidfield" );
  otherSystem.numMidfield    = otherSystem.numDefMidfield + otherSystem.numOffMidfield;
  otherSystem.numOffenders   = Config::getValueAsInt( "otherOffenders" );

  ourEdgeDefender      = Config::getValueAsInt( "tac_ourEdgeDefender" );
  ourEdgeDefMidfield   = Config::getValueAsInt( "tac_ourEdgeDefMidfield" );
  ourEdgeOffMidfield   = Config::getValueAsInt( "tac_ourEdgeOffMidfield" );
  ourEdgeOffender      = Config::getValueAsInt( "tac_ourEdgeOffender" );
  ourEdgeGoalie        = Config::getValueAsInt( "tac_ourEdgeGoalie" );

  otherEdgeDefender    = Config::getValueAsInt( "tac_otherEdgeDefender" );
  otherEdgeDefMidfield = Config::getValueAsInt( "tac_otherEdgeDefMidfield" );
  otherEdgeOffMidfield = Config::getValueAsInt( "tac_otherEdgeOffMidfield" );
  otherEdgeOffender    = Config::getValueAsInt( "tac_otherEdgeOffender" );
  otherEdgeGoalie      = Config::getValueAsInt( "tac_otherEdgeGoalie" );

  firstResponsiblePlayer  = UNKNOWN;
  secondResponsiblePlayer = UNKNOWN;
  thirdResponsiblePlayer  = UNKNOWN;

  firstDangerousOpponent  = UNKNOWN;
  secondDangerousOpponent = UNKNOWN;
  thirdDangerousOpponent  = UNKNOWN;

  for( int i = 0; i < 11; i++ )
    {
      ourTeamRBTime[i]   = -1;
      otherTeamRBTime[i] = -1;

      ourTeamRBPos[i].setXY( 0, 0 );
      otherTeamRBPos[i].setXY( 0,0 );

      ourTeamTNo[i]   = UNKNOWN;
      otherTeamTNo[i] = UNKNOWN;

      ourTeamStam[i]   = STAMINA_MAX;
      otherTeamStam[i] = STAMINA_MAX;

      otherTeamSimStamPossible[i] = false;

      // only do this for the coach (ToDo)
      playerTypeData[i] = new PlayerTypeData[playerParams.player_types];
      for( int j = 0; j < playerParams.player_types; j++ )
	{
	  playerTypeData[i][j].state = PTS_UNKNOWN;
	  playerTypeData[i][j].score = 0;
	}
      playerType[i] = UNKNOWN;
    }

  oldTime = -2;   // HACK
  oldNumSubs = 0;
}


TacticalSnapshot::~TacticalSnapshot()
{
  for( int i = 0; i < 11; i++ )
    delete [] playerTypeData[i];
}


bool
TacticalSnapshot::reportOurTacticalAction( TacticalActionState state, int perPlayer, int recPlayer )
{
  bool newAction = false;

  switch( state )
    {
    case TA_CONTROL :
      if( perPlayer >= 1 && perPlayer <= 11 )
	{
	  newAction = true;
	  reportedTAction.performer = perPlayer;
	  reportedTAction.receiver = UNKNOWN;
	}
      break;

    case TA_PASS :
      if( perPlayer >= 1 && perPlayer <= 11 )
	// REMARK: have to handle emergency kicks sometime...
	if( recPlayer >= 1 && recPlayer <= 11 )
	{
	  newAction = true;
	  reportedTAction.performer = perPlayer;
	  reportedTAction.receiver = recPlayer;
	}
      break;
     
    case TA_GOALSHOT :
      if( perPlayer >= 1 && perPlayer <= 11 )
	{
	  newAction = true;
	  reportedTAction.performer = perPlayer;
	  reportedTAction.receiver = UNKNOWN;
	}
      break;

    case TA_REACT :
      newAction = true;
      reportedTAction.performer = UNKNOWN;
      reportedTAction.receiver = UNKNOWN;
      break;

    case TA_UNKNOWN :
    default :
      break;
    }

  if( newAction )
    {
      reportedTAction.state = state;
      reportedTAction.quality = TQ_REPORTED;
      reportedTAction.startingTime = time;
      return true;
    }

  return false;
}


void
TacticalSnapshot::updateTacticalNumbers( Snapshot *snap )
{
  // try to determine goalie numbers

  if( !ourGoalieFound )
    if( snap->ourGoalieNo != UNKNOWN )
      {
	ourGoalieFound = true;

	// set tactical numbers of our team
	for( int i = 1; i <= 11; i++ )
	  {
	    if( i < snap->ourGoalieNo )
	      {
		ourTeamTNo[i-1] = i + 1;
	      }
	    else if( i > snap->ourGoalieNo )
	      {
		ourTeamTNo[i-1] = i;
	      }
	    else
	      {
		ourTeamTNo[i-1] = 1;
	      }
	  }
      }

  if( !otherGoalieFound )
    if( snap->otherGoalieNo != UNKNOWN )
      {
	otherGoalieFound = true;

	// set tactical numbers of other team
	for( int i = 1; i <= 11; i++ )
	  {
	    if( i < snap->otherGoalieNo )
	      {
		otherTeamTNo[i-1] = i + 1;
	      }
	    else if( i > snap->otherGoalieNo )
	      {
		otherTeamTNo[i-1] = i;
	      }
	    else
	      {
		otherTeamTNo[i-1] = 1;
	      }
	  }
      }

  // set my tactical number if it is still UNKNOWN
  if( myTacticalNo == UNKNOWN && ourGoalieFound && snap->my.number != UNKNOWN )
    {
      if( snap->my.number < snap->ourGoalieNo )
	{
	  myTacticalNo = snap->my.number + 1;
	}
      else if( snap->my.number > snap->ourGoalieNo )
	{
	  myTacticalNo = snap->my.number;
	}
      else
	{
	  myTacticalNo = 1;
	}
    }
}


bool
TacticalSnapshot::ourTeamDefender( int tn ) const
{
  ensure( tn >= 1 && tn <= 11 );
  if( tn < 1 || tn > 11 )
    return false;
  if( tn >= 2 && tn <= ourSystem.numDefenders + 1 )
    return true;
  return false;
}


bool
TacticalSnapshot::ourTeamDefMidfield( int tn ) const
{
  ensure( tn >= 1 && tn <= 11 );
  if( tn < 1 || tn > 11 )
    return false;
  if( tn > ourSystem.numDefenders + 1 &&
      tn <= ourSystem.numDefMidfield + ourSystem.numDefenders + 1 )
    return true;
  return false;
}


bool
TacticalSnapshot::ourTeamOffMidfield( int tn ) const
{
  ensure( tn >= 1 && tn <= 11 );
  if( tn < 1 || tn > 11 )
    return false;
  if( tn > ourSystem.numDefMidfield + ourSystem.numDefenders + 1 &&
      tn <= ourSystem.numMidfield + ourSystem.numDefenders + 1 )
    return true;
  return false;
}


bool
TacticalSnapshot::ourTeamMidfield( int tn ) const
{
  ensure( tn >= 1 && tn <= 11 );
  if( tn < 1 || tn > 11 )
    return false;
  if( tn > ourSystem.numDefenders + 1 &&
      tn <= ourSystem.numMidfield + ourSystem.numDefenders + 1 )
    return true;
  return false;
}


bool
TacticalSnapshot::ourTeamOffender( int tn ) const
{
  ensure( tn >= 1 && tn <= 11 );
  if( tn < 1 || tn > 11 )
    return false;
  if( tn > ourSystem.numMidfield + ourSystem.numDefenders + 1 )
    return true;
  return false;
}


bool
TacticalSnapshot::otherTeamDefender( int tn ) const
{
  ensure( tn >= 1 && tn <= 11 );
  if( tn < 1 || tn > 11 )
    return false;
  if( tn >= 2 && tn <= otherSystem.numDefenders + 1 )
    return true;
  return false;
}


bool
TacticalSnapshot::otherTeamDefMidfield( int tn ) const
{
  ensure( tn >= 1 && tn <= 11 );
  if( tn < 1 || tn > 11 )
    return false;
  if( tn > otherSystem.numDefenders + 1 &&
      tn <= otherSystem.numDefMidfield + otherSystem.numDefenders + 1 )
    return true;
  return false;
}


bool
TacticalSnapshot::otherTeamOffMidfield( int tn ) const
{
  ensure( tn >= 1 && tn <= 11 );
  if( tn < 1 || tn > 11 )
    return false;
  if( tn > otherSystem.numDefMidfield + otherSystem.numDefenders + 1 &&
      tn <= otherSystem.numMidfield + otherSystem.numDefenders + 1 )
    return true;
  return false;
}


bool
TacticalSnapshot::otherTeamMidfield( int tn ) const
{
  ensure( tn >= 1 && tn <= 11 );
  if( tn < 1 || tn > 11 )
    return false;
  if( tn > otherSystem.numDefenders + 1 &&
      tn <= otherSystem.numMidfield + otherSystem.numDefenders + 1 )
    return true;
  return false;
}


bool
TacticalSnapshot::otherTeamOffender( int tn ) const
{
  ensure( tn >= 1 && tn <= 11 );
  if( tn < 1 || tn > 11 )
    return false;
  if( tn > otherSystem.numMidfield + otherSystem.numDefenders + 1 )
    return true;
  return false;
}


void
TacticalSnapshot::updateStamina( Snapshot *snap, const TacticalModuleInfo& receivedInfo, int currentTime )
{
  for( int i = 0; i < 11; i++ )
    {
      // kick-off, halftime, extended time -> stamina refresh
      if( snap->playMode == PLAY_MODE_HALF_TIME ||
	  snap->playMode == PLAY_MODE_TIME_EXTENDED ||
	  snap->playMode == PLAY_MODE_BEFORE_KICK_OFF )
	{
	  ourTeamStam[i].value = STAMINA_MAX;
	  ourTeamStam[i].age = 0;
	  otherTeamStam[i].value = STAMINA_MAX;
	  otherTeamStam[i].age = 0;
	  otherTeamSimStamPossible[i] = false;
	}
      else
	{
	  const WorldPlayer& player = snap->otherTeam( i + 1 );
	  // try to simulate opponent stamina
	  if( player.age <= 1 )
	    {
	      if( otherTeamSimStamPossible[i] )
		{
		  // player movement since last stamina simulation
		  double d = dist( otherTeamLastStamPos[i], player.pos );
		  
		  int n = otherTeamStam[i].age - player.age;

		  d *= 0.9;   // error of about 10 % because of noise and sight blur
		  
		  double tmp = ( 1 - power( player.decay, n ) ) / ( 1 - player.decay );
		  
		  // maximum effect achieved by "sliding"
		  d = max( 0.0, d - player.vmax * player.decay * tmp );
		  
		  // subtract minimum stamina needed for that position change
		  otherTeamStam[i].value = otherTeamLastStamValue[i] +
		    n * player.staminaIncMax - d / tmp / player.dashPowerRate;
		  otherTeamStam[i].value = min( STAMINA_MAX,
						max( EFFORT_DEC_THR * STAMINA_MAX,
						     otherTeamStam[i].value ) );

		  STAMSIMTRACE( " : Updated stamina of opp. " << i + 1 << " to "
				<< otherTeamStam[i].value << ", had not done that for "
				<< n << " steps." );
		}
	      otherTeamStam[i].age        = player.age;
	      otherTeamSimStamPossible[i] = true;
	      otherTeamLastStamValue[i]   = min( STAMINA_MAX, otherTeamStam[i].value +
						 player.age * player.staminaIncMax );
	      otherTeamLastStamPos[i]     = player.pos;
	    }

	  // get my own stamina from snapshot
	  if( i == snap->my.number - 1 )
	    {
	      ourTeamStam[i].value = snap->senseBody.stamina;
	      ourTeamStam[i].age = 0;
	    }
	  // use audial information for the rest of my team
	  else if( receivedInfo.valid )
	    {
	      int stamAge = receivedInfo.ourTeamStam[i].age;

	      if( receivedInfo.time < currentTime )
		stamAge += currentTime - receivedInfo.time;
	      if( stamAge < ourTeamStam[i].age )
		{
		  ourTeamStam[i].age = stamAge;
		  ourTeamStam[i].value = receivedInfo.ourTeamStam[i].value;
		}
	    }
	}
    }
}


void
TacticalSnapshot::determineResponsibleFriends( Snapshot *snap )
{
  WorldBall simBall = snap->ball;      // simulated ball
  int       simTime = 0;               // number of simulated steps

  double    useStamina[11];            // guessed stamina used for simulation
  double    myTeamDist[11];            // radius of players' movements
  double    myTeamV[11];               // players' velocity
  bool      sorted[11];                // has player "reached the ball"?
  int       numSimulatedPlayers = 0;   // number of players that have "reached the ball"
  bool      playerLeft = true;         // is there a player left to be sorted?

  // if ball is close to a player we can not rely on its velocity
  // so we set it to zero
  for( int i = 1; i <= 11; i++ )
    {
      if( snap->ourTeam( i ).distTo( simBall ) < 1.5 * snap->ourTeam( i ).kickRange() ||
	  snap->otherTeam( i ).distTo( simBall ) < 1.5 * snap->otherTeam( i ).kickRange() )
	{
	  simBall.v.setXY( 0.0, 0.0 );
	  break;
	}
    }

  for( int i = 0; i < 11; i++ )
    {
      // use simulated stamina for team members if it is not too old
      // player's own stamina is well known
      if( i == snap->my.number - 1 )
	useStamina[i] = ourTeamStam[i].value - EFFORT_DEC_THR * STAMINA_MAX;
      else
	useStamina[i] = max( 5 * MAXPOWER,
			     ourTeamStam[i].value - EFFORT_DEC_THR * STAMINA_MAX -
			     ( MAXPOWER - snap->ourTeam( i + 1 ).staminaIncMax ) * ourTeamStam[i].age );


      sorted[i] = false;
      myTeamDist[i] = 0.0;
      myTeamV[i] = 0.0;
    }

  // only find the (up to) three "most responsible" players
  while( playerLeft && numSimulatedPlayers < 3 && simTime < 50 )
    {
      playerLeft = false;
      for( int i = 0; i < 11; i++ )
	{
	  const WorldPlayer& player = snap->ourTeam( i + 1 );

	  if( player.age <= MAXAGE && !sorted[i] && numSimulatedPlayers < 3 )
	    {
	      playerLeft = true;
	      // has player already reached the ball?
	      if( dist( player.pos, simBall.pos ) < myTeamDist[i] + player.kickRange() )
		{
		  sorted[i] = true;
		  ourTeamRBTime[i] = simTime;
		  ourTeamRBPos[i] = simBall.pos;
		  switch( numSimulatedPlayers )
		    {
		    case 0  : firstResponsiblePlayer = i + 1; break;
		    case 1  : secondResponsiblePlayer = i + 1; break;
		    case 2  : thirdResponsiblePlayer = i + 1; break;
		    default : break;
		    }
		  numSimulatedPlayers++;
		}
	      else   // if not: simulate player's movement
		{
		  myTeamV[i] = min( player.vmax,
				    ( player.staminaIncMax +
				      max( 0.0, min( MAXPOWER - player.staminaIncMax, useStamina[i] ) ) )
				    * player.dashPowerRate + myTeamV[i] * player.decay );
		  
		  useStamina[i] -= max( 0.0, min( MAXPOWER - player.staminaIncMax, useStamina[i] ) );
		  useStamina[i] = max( 0.0, useStamina[i] );
		  myTeamDist[i] += myTeamV[i];
		}
	    }
	}

      simBall.simulate();
      simTime++;
    }

  // check responsible players for their age (maybe throw away old simulated information!)
  while( firstResponsiblePlayer != UNKNOWN && snap->ourTeam( firstResponsiblePlayer ).age > MAXAGE )
    {
      firstResponsiblePlayer = secondResponsiblePlayer;
      secondResponsiblePlayer = thirdResponsiblePlayer;
      thirdResponsiblePlayer = UNKNOWN;
    }
  while( secondResponsiblePlayer != UNKNOWN && snap->ourTeam( secondResponsiblePlayer ).age > MAXAGE )
    {
      secondResponsiblePlayer = thirdResponsiblePlayer;
      thirdResponsiblePlayer = UNKNOWN;      
    }
  if( thirdResponsiblePlayer != UNKNOWN && snap->ourTeam( thirdResponsiblePlayer ).age > MAXAGE )
    thirdResponsiblePlayer = UNKNOWN;

  // check for double players (mix of simulated and now calculated information)
  if( firstResponsiblePlayer == thirdResponsiblePlayer )
    thirdResponsiblePlayer = UNKNOWN;
  if( firstResponsiblePlayer == secondResponsiblePlayer )
    {
      secondResponsiblePlayer = thirdResponsiblePlayer;
      thirdResponsiblePlayer = UNKNOWN;
    }
  if( secondResponsiblePlayer == thirdResponsiblePlayer )
    thirdResponsiblePlayer = UNKNOWN;

  RESPONSIBLETRACE( " : Closest friends: " << firstResponsiblePlayer
		    << " (" << ourTeamReachBallTime( firstResponsiblePlayer ) << ") "
		    << secondResponsiblePlayer
		    << " (" << ourTeamReachBallTime( secondResponsiblePlayer ) << ") "
		    << thirdResponsiblePlayer
		    << " (" << ourTeamReachBallTime( thirdResponsiblePlayer ) << ")" );
}


void
TacticalSnapshot::determineDangerousOpponents( Snapshot *snap )
{
  WorldBall simBall = snap->ball;      // simulated ball
  int       simTime = 0;               // number of simulated steps

  double    useStamina[11];            // players' simulated stamina
  double    otherTeamDist[11];         // radius of players' movements
  double    otherTeamV[11];            // players' velocity  
  bool      sorted[11];                // has player "reached the ball"?
  int       numSimulatedPlayers = 0;   // number of players that have "reached the ball"
  int       numValidPlayers = 0;       // number of players that are young enough
  bool      playerLeft = true;         // is there a player left to be sorted?

  // if ball is close to a player we can not rely on its velocity
  // so we set it to zero
  for( int i = 1; i <= 11; i++ )
    {
      if( snap->ourTeam( i ).distTo( simBall ) < 1.5 * snap->ourTeam( i ).kickRange() ||
	  snap->otherTeam( i ).distTo( simBall ) < 1.5 * snap->otherTeam( i ).kickRange() )
	{
	  simBall.v.setXY( 0.0, 0.0 );
	  break;
	}
    }

  for( int i = 0; i < 11; i++ )
    {
      sorted[i] = false;
      useStamina[i] = STAMINA_MAX - EFFORT_DEC_THR * STAMINA_MAX;   // (almost) worst case scenario
      otherTeamDist[i] = 0.0;
      otherTeamV[i] = 0.0;
    }

  for( int i = 1; i <= 11; i++ )
    if( snap->otherTeam( i ).age <= MAXAGE )
      numValidPlayers++;

  // only find the (up to) three most dangerous opponents
  while( playerLeft && numSimulatedPlayers < 3 && simTime < 50 )
    {
      playerLeft = false;
      for( int i = 0; i < 11; i++ )
	{
	  const WorldPlayer& player = snap->otherTeam( i + 1 );

	  if( player.age < MAXAGE && !sorted[i] && numSimulatedPlayers < 3 )
	    {
	      playerLeft = true;
	      // has player already reached the ball?
	      if( dist( player.pos, simBall.pos ) < otherTeamDist[i] + player.kickRange() )
		{
		  sorted[i] = true;
		  otherTeamRBTime[i] = simTime;
		  otherTeamRBPos[i] = simBall.pos;
		  switch( numSimulatedPlayers )
		    {
		    case 0  : firstDangerousOpponent = i + 1; break;
		    case 1  : secondDangerousOpponent = i + 1; break;
		    case 2  : thirdDangerousOpponent = i + 1; break;
		    default : break;
		    }
		  numSimulatedPlayers++;
		}
	      else   // if not: simulate player's movement
		{
		  otherTeamV[i] = min( player.vmax,
				       ( player.staminaIncMax +
					 max( 0.0, min( MAXPOWER - player.staminaIncMax, useStamina[i] ) ) )
				       * player.dashPowerRate + otherTeamV[i] * player.decay );
		  
		  useStamina[i] -= max( 0.0, min( MAXPOWER - player.staminaIncMax, useStamina[i] ) );
		  useStamina[i] = max( 0.0, useStamina[i] );
		  otherTeamDist[i] += otherTeamV[i];
		}
	    }
	}

      simBall.simulate();
      simTime++;
    }

  // check responsible players for their age (maybe throw away old simulated information!)
  while( firstDangerousOpponent != UNKNOWN && snap->otherTeam( firstDangerousOpponent ).age > MAXAGE )
    {
      firstDangerousOpponent = secondDangerousOpponent;
      secondDangerousOpponent = thirdDangerousOpponent;
      thirdDangerousOpponent = UNKNOWN;
    }
  while( secondDangerousOpponent != UNKNOWN && snap->otherTeam( secondDangerousOpponent ).age > MAXAGE )
    {
      secondDangerousOpponent = thirdDangerousOpponent;
      thirdDangerousOpponent = UNKNOWN;      
    }
  if( thirdDangerousOpponent != UNKNOWN && snap->otherTeam( thirdDangerousOpponent ).age > MAXAGE )
    thirdDangerousOpponent = UNKNOWN;

  // check for double players (mix of simulated and now calculated information)
  if( firstDangerousOpponent == thirdDangerousOpponent )
    thirdDangerousOpponent = UNKNOWN;
  if( firstDangerousOpponent == secondDangerousOpponent )
    {
      secondDangerousOpponent = thirdDangerousOpponent;
      thirdDangerousOpponent = UNKNOWN;
    }
  if( secondDangerousOpponent == thirdDangerousOpponent )
    thirdDangerousOpponent = UNKNOWN;

  RESPONSIBLETRACE( " : Closest opponents: " << firstDangerousOpponent
		    << " (" << otherTeamReachBallTime( firstDangerousOpponent ) << ") "
		    << secondDangerousOpponent
		    << " (" << otherTeamReachBallTime( secondDangerousOpponent ) << ") "
		    << thirdDangerousOpponent
		    << " (" << otherTeamReachBallTime( thirdDangerousOpponent ) << ")" );
}


void
TacticalSnapshot::determineBallControl( Snapshot *snap,
					BallControlState ourBallControlAudio )
{
  BallControlState ourBallControlNew = BC_UNKNOWN;     // new value for ourBallControl

  int firstFriendReachTime = -1;     // time when firstResponsiblePlayer will reach the ball
  int firstOpponentReachTime = -1;   // time when firstDangerousOpponent will reach the ball

  switch( snap->playMode )
    {
    case PLAY_MODE_UNKNOWN :
    case PLAY_MODE_BEFORE_KICK_OFF :
    case PLAY_MODE_TIME_OVER :
    case PLAY_MODE_HALF_TIME :
    case PLAY_MODE_TIME_EXTENDED :
      
      // there is no game running
      ourBallControlNew = BC_UNKNOWN;
      break;

    case PLAY_MODE_KICK_OFF_US :
    case PLAY_MODE_KICK_IN_US :
    case PLAY_MODE_FREE_KICK_US :
    case PLAY_MODE_CORNER_KICK_US :
    case PLAY_MODE_GOAL_KICK_US :
    case PLAY_MODE_GOAL_THEM :
    case PLAY_MODE_OFFSIDE_THEM :

      // in a standard situation for us ball control is obvious
      ourBallControlNew = BC_SAFE;
      break;

    case PLAY_MODE_KICK_OFF_THEM :
    case PLAY_MODE_KICK_IN_THEM :
    case PLAY_MODE_FREE_KICK_THEM :
    case PLAY_MODE_CORNER_KICK_THEM :
    case PLAY_MODE_GOAL_KICK_THEM :
    case PLAY_MODE_GOAL_US :
    case PLAY_MODE_OFFSIDE_US :

      // in a standard situation for the opponent team ball control is obvious
      ourBallControlNew = BC_LOST;
      break;

    case PLAY_MODE_DROP_BALL :
    case PLAY_MODE_PLAY_ON :

      // is there a player of our team? (there should be at least one...)
      if( firstResponsiblePlayer != UNKNOWN )
	{
	  firstFriendReachTime = ourTeamReachBallTime( firstResponsiblePlayer );

	  int ourEdge = ( ourTeamDefender( firstResponsiblePlayer ) ? ourEdgeDefender :
			  ourTeamDefMidfield( firstResponsiblePlayer ) ? ourEdgeDefMidfield :
			  ourTeamOffMidfield( firstResponsiblePlayer ) ? ourEdgeOffMidfield :
			  ourTeamOffender( firstResponsiblePlayer ) ? ourEdgeOffender :
			  ourEdgeGoalie );

	  int otherEdge = ( ourTeamDefender( firstResponsiblePlayer ) ? otherEdgeDefender :
			    ourTeamDefMidfield( firstResponsiblePlayer ) ? otherEdgeDefMidfield :
			    ourTeamOffMidfield( firstResponsiblePlayer ) ? otherEdgeOffMidfield :
			    ourTeamOffender( firstResponsiblePlayer ) ? otherEdgeOffender :
			    ourEdgeGoalie );

	  // is there a player of the opponent team?
	  if( firstDangerousOpponent != UNKNOWN )
	    {
	      firstOpponentReachTime = otherTeamReachBallTime( firstDangerousOpponent );

	      // opponent is faster
	      if( firstOpponentReachTime + otherEdge < firstFriendReachTime )
		{
		  ourBallControlNew = ( ourBallControl == BC_SAFE ? BC_DANGER : BC_LOST );
		}
	      // we are faster
 	      else if( firstFriendReachTime + ourEdge < firstOpponentReachTime )
		{
		  ourBallControlNew = ( ourBallControl == BC_LOST ? BC_DANGER : BC_SAFE );
		}
	      // dangerous situation
	      else
		{
		  ourBallControlNew = BC_DANGER;
		}
	    }
	  else   // no opponent seen
	    {
	      ourBallControlNew = ( ourBallControl == BC_LOST ? BC_DANGER : BC_SAFE);
	    }
	}
      else   // weird case: no player of our team found...
	{
	  // is there at least an opponent?
	  if( firstDangerousOpponent != UNKNOWN )
	    {
	      firstOpponentReachTime = otherTeamReachBallTime( firstDangerousOpponent );
	  
	      ourBallControlNew = ( ourBallControl == BC_SAFE ? BC_DANGER : BC_LOST);
	    }
	  else   // there is NOBODY around... try to use friend's ball control values
	    {
	      ourBallControlNew = ourBallControlAudio;
	    }
	}
      break;

    default :

      ensure( 0 );
      break;
    }

  ourBallControl = ourBallControlNew;
}


bool
TacticalSnapshot::determineOurTacticalAction( Snapshot *snap, const TacticalAction& audioAction, int currentTime )
{
  bool useEmChannel = false;   // use emergency channel if agent reported an action
  bool useHistory = false;     // are tactical history and snap consistent?

  int actionAge = currentTime - ourAction.startingTime;

  TacticalAction ourActionNew( TA_UNKNOWN, UNKNOWN, UNKNOWN, currentTime );

  switch( snap->playMode )
    {
    case PLAY_MODE_UNKNOWN :
    case PLAY_MODE_BEFORE_KICK_OFF :
    case PLAY_MODE_TIME_OVER :
    case PLAY_MODE_HALF_TIME :
    case PLAY_MODE_TIME_EXTENDED :
      
      ourActionNew.quality = TQ_GUESSED;
      break;

    case PLAY_MODE_KICK_OFF_US :
    case PLAY_MODE_KICK_IN_US :
    case PLAY_MODE_FREE_KICK_US :
    case PLAY_MODE_CORNER_KICK_US :
    case PLAY_MODE_GOAL_THEM :
    case PLAY_MODE_OFFSIDE_THEM :
    case PLAY_MODE_GOAL_KICK_US :   // this is not a "standard standard situation"!!!
      // REMARK: in a standard situation for us the state can only be TA_CONTROL or
      // TA_PASS, and the latter only in PLAY_MODE_GOAL_KICK_US

      // ---< 1 >--- take reported information into account
      if( reportedTAction.quality == TQ_REPORTED &&
	  ( reportedTAction.state == TA_CONTROL ||
	    snap->playMode == PLAY_MODE_GOAL_KICK_US && reportedTAction.state == TA_PASS ) &&
	  reportedTAction.performer != UNKNOWN )
	{
	  useEmChannel           = true;
	  ourActionNew.state     = reportedTAction.state;
	  ourActionNew.quality   = TQ_REPORTED;
	  ourActionNew.performer = reportedTAction.performer;
	  ourActionNew.receiver  = reportedTAction.receiver;
	  break;
	}

      // ---< 2 >--- take audial information into account
      if( ( audioAction.state == TA_CONTROL ||
	    ( snap->playMode == PLAY_MODE_GOAL_KICK_US && audioAction.state == TA_PASS ) ) &&
	  audioAction.performer != UNKNOWN )
	{
	  // if sending player is sure about what he sent... believe him
	  if( audioAction.quality == TQ_REPORTED )
	    {
	      ourActionNew.state     = audioAction.state;
	      ourActionNew.quality   = TQ_AUDIAL_REPORTED;
	      ourActionNew.performer = audioAction.performer;
	      ourActionNew.receiver  = audioAction.receiver;
	      break;
	    }
	  // if he is not so sure compare it with our results
	  if( audioAction.quality == TQ_AUDIAL_REPORTED )
	    {
	      if( audioAction.performer == firstResponsiblePlayer ||
		  audioAction.performer == secondResponsiblePlayer ||
		  audioAction.performer == thirdResponsiblePlayer )
		{
		  ourActionNew.state     = audioAction.state;
		  ourActionNew.quality   = TQ_AUDIAL;
		  ourActionNew.performer = audioAction.performer;
		  ourActionNew.receiver  = audioAction.receiver;
		  break;
		}
	    }
	}

      // ---< 3 >--- try to determine performer by history

      // first case : goal kick situation and pass action
      if( snap->playMode == PLAY_MODE_GOAL_KICK_US &&
	  ourAction.state == TA_PASS &&
	  ( ourAction.quality == TQ_REPORTED || ourAction.quality == TQ_AUDIAL_REPORTED ) &&
	  ourAction.performer != UNKNOWN )
	{
	  // keep that until some player changes it (to control)
	  ourActionNew = ourAction;
	  break;
	}
	  
      // second case : control action
      if( ourAction.state     == TA_CONTROL &&
	  ourAction.quality   != TQ_UNKNOWN &&
	  ourAction.performer != UNKNOWN )
	{
	  // if old performer fits into new worldmodel or is a reported one keep him
	  if( ourAction.performer == firstResponsiblePlayer ||
	      ourAction.performer == secondResponsiblePlayer ||
	      ourAction.performer == thirdResponsiblePlayer ||
	      ourAction.quality   == TQ_REPORTED ||
	      ourAction.quality   == TQ_AUDIAL_REPORTED )
	    {
	      ourActionNew = ourAction;
	      break;
	    }
	}

      // ---< 4 >--- if all else fails, take firstResponsiblePlayer as performer of CONTROL
      ourActionNew.state     = TA_CONTROL;
      ourActionNew.quality   = TQ_GUESSED;
      ourActionNew.performer = firstResponsiblePlayer;
      ourActionNew.receiver  = UNKNOWN;
      break;

    case PLAY_MODE_KICK_OFF_THEM :
    case PLAY_MODE_KICK_IN_THEM :
    case PLAY_MODE_FREE_KICK_THEM :
    case PLAY_MODE_CORNER_KICK_THEM :
    case PLAY_MODE_GOAL_KICK_THEM :
    case PLAY_MODE_GOAL_US :
    case PLAY_MODE_OFFSIDE_US :

      // in a standard situation for the opponent team our tactical action is obvious
      ourActionNew.state   = TA_REACT;
      ourActionNew.quality = TQ_GUESSED;
      break;

    case PLAY_MODE_DROP_BALL :
    case PLAY_MODE_PLAY_ON :

      // CAUTION: very simple implementation!

      // step 1 : first take reported information into account
      if( reportedTAction.quality == TQ_REPORTED && reportedTAction.state != TA_UNKNOWN )
	{
	  useEmChannel           = true;
	  ourActionNew.quality   = TQ_REPORTED;
	  ourActionNew.state     = reportedTAction.state;
	  ourActionNew.performer = reportedTAction.performer;
	  ourActionNew.receiver  = reportedTAction.receiver;
	  break;
	}

      // step 2 : if there is no reported information take audio information into account
      if( audioAction.state != TA_UNKNOWN )
	if( audioAction.quality == TQ_REPORTED || audioAction.quality == TQ_AUDIAL_REPORTED )
	  // do not overwrite better information
	  if( ( ourAction.quality != TQ_REPORTED && ourAction.quality != TQ_AUDIAL_REPORTED ) ||
	      audioAction.startingTime > ourAction.startingTime )
	    // do not use information that mentions myself as performer - I should know what I do!!!
	    if( audioAction.performer != snap->my.number || ourAction.state == TA_UNKNOWN )
	      {
		ourActionNew.quality   = ( audioAction.quality == TQ_REPORTED ? TQ_AUDIAL_REPORTED : TQ_AUDIAL );
		ourActionNew.state     = audioAction.state;
		ourActionNew.performer = audioAction.performer;
		ourActionNew.receiver  = audioAction.receiver;
		// use our team mate's starting time for that action
		ourActionNew.startingTime = audioAction.startingTime;
		break;
	      }

      // step 3 : if there is no good audio information try to use history
      if( ourAction.quality != TQ_UNKNOWN )
	{
	  switch( ourAction.state )
	    {
	    case TA_CONTROL :
	      // keep old action if ball is not lost
	      // and if performer is still responsible
	      if( ourBallControl != BC_LOST &&
		  ( ourAction.performer == firstResponsiblePlayer ||
		    ourAction.performer == secondResponsiblePlayer ||
		    ourAction.performer == thirdResponsiblePlayer ) )
		// before: || ourAction.quality == TQ_REPORTED ||
		// ourAction.quality == TQ_AUDIAL_REPORTED ) )
		// does not make sense in non-standard situations
		{
		  useHistory = true;
		}
	      break;
	      
	    case TA_PASS :
	      // keep old action if ball is not lost...
	      if( ( ourBallControl != BC_LOST ||
		    firstDangerousOpponent == UNKNOWN ||
		    otherTeamReachBallTime( firstDangerousOpponent ) > 5 ) &&
		  // ...and if no friend has it under control
		  (
		    firstResponsiblePlayer == ourAction.performer ||
		    firstResponsiblePlayer == UNKNOWN ||
		    ourTeamReachBallTime( firstResponsiblePlayer ) > 2 ) )
		{
		  useHistory = true;
		}
	      break;

	    case TA_GOALSHOT :
	      // keep old action if ball is not lost
	      if( ourBallControl != BC_LOST ||
		  firstDangerousOpponent == UNKNOWN ||
		  otherTeamReachBallTime( firstDangerousOpponent ) > 2 )
		{
		  useHistory = true;
		}
	      break;

	    case TA_REACT :
	      // keep old action if ball control is not safe again
	      if( ourBallControl != BC_SAFE )
		{
		  useHistory = true;
		}
	      break;

	    case TA_UNKNOWN :
	    default :
	      break;
	    }
	}
      
      // use history if it is consistent with world and not too old
      if( useHistory &&
	  ( ourAction.state == TA_CONTROL  ? actionAge < 20 :
	    ( ourAction.state == TA_PASS     ? actionAge < 10 :
	      ( ourAction.state == TA_GOALSHOT ? actionAge < 15 :   // 10 ?
		( ourAction.state == TA_REACT    ? actionAge < 20 : false ) ) ) ) )
	{
	  ourActionNew = ourAction;
	  break;
	}

      // history and world are not consistent -> try to determine tactical action
      // by looking at the world

      switch( ourBallControl )
	{
	case BC_SAFE :
	  if( firstResponsiblePlayer == UNKNOWN )
	    ourActionNew.state = TA_REACT;
	  else
	    {
	      ourActionNew.state = TA_CONTROL;
	      ourActionNew.performer = firstResponsiblePlayer;
	    }
	  break;

	case BC_DANGER :
	  // use history in that case
	  if( ourAction.state == TA_REACT || firstResponsiblePlayer == UNKNOWN )
	    ourActionNew.state = TA_REACT;
	  else
	    {
	      ourActionNew.state = TA_CONTROL;
	      ourActionNew.performer = firstResponsiblePlayer;
	    }
	  break;

	case BC_LOST :
	  ourActionNew.state = TA_REACT;
	  break;

	case BC_UNKNOWN :
	default :
	  ourActionNew.state = TA_REACT;
	  break;
	}

      ourActionNew.quality = TQ_GUESSED;
      break;

    default :

      ensure( 0 );
      break;
    }

  // if old and new action for our team are the same, don't change startingTime
  // (only if ourActionNew comes from audial information)
  if( ourAction.state == ourActionNew.state &&
      ourAction.performer == ourActionNew.performer &&
      ourAction.receiver == ourActionNew.receiver )
    {
      ourActionNew.startingTime = min( ourAction.startingTime, ourActionNew.startingTime );

      // Don't use emergency channel too often for the same action
      // if( ( currentTime - ourAction.startingTime ) % 5 != 0 &&
      //       ourAction.quality == TQ_REPORTED )
      //   useEmChannel = false;
    }

  ourAction = ourActionNew;

  // if action is very old, downgrade its quality
  if( ourAction.startingTime < currentTime - 10 )
    {
      if( ourAction.quality == TQ_REPORTED )
	ourAction.quality = TQ_GUESSED;
      if( ourAction.quality == TQ_AUDIAL_REPORTED )
	ourAction.quality = TQ_AUDIAL;
    }

  return useEmChannel;
}


void
TacticalSnapshot::determineOtherPlayerTypes( Snapshot *snap )
{
  int newTime = snap->time;
  int numTypes = playerParams.player_types;
  bool playerSubstituted;

  // only evaluate in PLAY_ON mode and if we had a snapshot one timestep ago
  if( snap->playMode == PLAY_MODE_PLAY_ON && newTime - oldTime == 1 )
    {
      for( int i = 0; i < 11; i++ )
	{
	  playerSubstituted = false;
	  if( snap->numOtherSubs > oldNumSubs )
	    {
	      for( int k = 0; k < snap->numOtherSubs - oldNumSubs; k++ )
		{
		  if( snap->otherSubs[( snap->numOtherSubs - 1 - k ) %
				     ( playerParams.subs_max + 11 )].number == i + 1 )
		    {
		      TYPEDETECTTRACE( "Player " << i + 1 << " was substituted!" );
		      playerSubstituted = true;
		      playerType[i] = UNKNOWN;

		      for( int j = 0; j < numTypes; j++ )
			{
			  playerTypeData[i][j].state = PTS_UNKNOWN;
			  playerTypeData[i][j].score = 0;
			}
		    }
		}
	    }

	  if( playerSubstituted )
	    continue;
	  if( playerType[i] != UNKNOWN )
	    continue;

	  const WorldPlayer& newPlayer = snap->otherTeam( i + 1 );
	  double posDiff = norm( newPlayer.pos - oldPlayer[i].pos );
	  for( int j = 0; j < numTypes; j++ )
	    if( playerTypeData[i][j].state == PTS_UNKNOWN )
	      {
		if( posDiff > snap->playerType[j].playerSpeedMax )
		  {
		    TYPEDETECTTRACE( "excluded type " << j << " for player " << i + 1 );
		    playerTypeData[i][j].state = PTS_IMPOSSIBLE;
		    playerTypeData[i][j].score = 0;
		    continue;
		  }

		if( ( fabs( MAXMOMENT ) /
		      fabs( newPlayer.bodyDir - oldPlayer[i].bodyDir ) - 1 )
		    / norm( oldPlayer[i].v ) < snap->playerType[j].inertiaFactor )
		  {
		    TYPEDETECTTRACE( "excluded type " << j << " for player " << i + 1 );
		    playerTypeData[i][j].state = PTS_IMPOSSIBLE;
		    playerTypeData[i][j].score = 0;
		    continue;
		  }

		Vector2d accel = newPlayer.pos - oldPlayer[i].pos - oldPlayer[i].v;
		if( norm( accel ) / max( MINPOWER, MAXPOWER ) > snap->playerType[j].dashPowerRate )
		  {
		    TYPEDETECTTRACE( "excluded type " << j << " for player " << i + 1 );
		    playerTypeData[i][j].state = PTS_IMPOSSIBLE;
		    playerTypeData[i][j].score = 0;
		    continue;
		  }

		double decay = norm( oldPlayer[i].v + accel ) / norm( newPlayer.v );
		if( decay > snap->playerType[j].playerDecay )
		  {
		    TYPEDETECTTRACE( "excluded type " << j << " for player " << i + 1 );
		    playerTypeData[i][j].state = PTS_IMPOSSIBLE;
		    playerTypeData[i][j].score = 0;
		    continue;
		  }
	      }

	  if( playerType[i] == UNKNOWN )
	    {
	      int lastNotImpossible = -1;
	      int numImpossible = 0;
	      // check if all types but one are excluded
	      for( int j = 0; j < numTypes; j++ )
		{
		  if( playerTypeData[i][j].state == PTS_IMPOSSIBLE )
		    numImpossible++;
		  else
		    lastNotImpossible = j;
		}
	      if( numImpossible == numTypes - 1 )
		{
		  playerTypeData[i][lastNotImpossible].state = PTS_SURE;
		  playerType[i] = lastNotImpossible;
		  TYPEDETECTTRACE( "Player " << i + 1 << " has type " << lastNotImpossible );
		}
	    }
	}
    }

  // finally reset oldPlayer array and oldTime
  for( int i = 0; i < 11; i++ )
    oldPlayer[i] = snap->otherTeam( i + 1 );
  oldTime = newTime;
  oldNumSubs = snap->numOtherSubs;
}

ostream &operator << (ostream &os, const TacticalSnapshot &ts)
{
  return os << "TSnap: " << ts.ourAction
	    << " (1stR: " << ts.firstResponsiblePlayer
	    << ", 2ndR: " << ts.secondResponsiblePlayer
	    << ", 3rdR: " << ts.thirdResponsiblePlayer
	    << ")(1stO: " << ts.firstDangerousOpponent
	    << ", 2ndO: " << ts.secondDangerousOpponent
	    << ", 3rdO: " << ts.thirdDangerousOpponent;
}


//////////////////////////////////////////////////////////////////////
//
// Class TacticalModule
//
//////////////////////////////////////////////////////////////////////

TacticalModule::TacticalModule( int _numSnapshots )
{
  numSnapshots = _numSnapshots;
  ensure( numSnapshots <= 30 );   // just to avoid out of memory (30 IS enough!)

  snapshots = new TacticalSnapshot[numSnapshots];
  ensure( snapshots );
  
  currentTime = 0;

  useEmergencyChannel = false;
}


TacticalModule::TacticalModule( const TacticalModule& tm )
{
  numSnapshots = tm.numSnapshots;

  snapshots = new TacticalSnapshot[numSnapshots];
  ensure( snapshots );
  
  for( int i = 0; i < numSnapshots; i++ )
    snapshots[i] = tm.snapshots[i];

  currentTime = tm.currentTime;
  
  useEmergencyChannel = tm.useEmergencyChannel;
}


TacticalModule& 
TacticalModule::operator = ( const TacticalModule& tm )
{
  numSnapshots = tm.numSnapshots;

  delete [] snapshots;

  snapshots = new TacticalSnapshot[numSnapshots];
  ensure( snapshots );

  for( int i = 0; i < numSnapshots; i++ )
    snapshots[i] = tm.snapshots[i];

  currentTime = tm.currentTime;

  useEmergencyChannel = tm.useEmergencyChannel;

  return *this;
}


TacticalModule::~TacticalModule()
{
  delete [] snapshots;
}


void
TacticalModule::simulate( Snapshot* snap, int time )
{
  TacticalSnapshot &oldSnap = snapshots[getTimeIndex( time )];
  TacticalSnapshot &newSnap = snapshots[getTimeIndex( time + 1 )];

  newSnap.time = oldSnap.time + 1;
  
  // copy old information

  newSnap.ourBallControl = oldSnap.ourBallControl;

  newSnap.ourAction = oldSnap.ourAction;

  newSnap.ourSystem = oldSnap.ourSystem;
  newSnap.otherSystem = oldSnap.otherSystem;

  newSnap.myTacticalNo = oldSnap.myTacticalNo;

  newSnap.firstResponsiblePlayer  = oldSnap.firstResponsiblePlayer;
  newSnap.secondResponsiblePlayer = oldSnap.secondResponsiblePlayer;
  newSnap.thirdResponsiblePlayer  = oldSnap.thirdResponsiblePlayer;

  newSnap.firstDangerousOpponent  = oldSnap.firstDangerousOpponent;
  newSnap.secondDangerousOpponent = oldSnap.secondDangerousOpponent;
  newSnap.thirdDangerousOpponent  = oldSnap.thirdDangerousOpponent;

  newSnap.ourGoalieFound   = oldSnap.ourGoalieFound;
  newSnap.otherGoalieFound = oldSnap.otherGoalieFound;

  for( int i = 0; i < 11; i++ )
    {
      newSnap.ourTeamRBTime[i]   = oldSnap.ourTeamRBTime[i];
      newSnap.otherTeamRBTime[i] = oldSnap.otherTeamRBTime[i];

      newSnap.ourTeamRBPos[i] = oldSnap.ourTeamRBPos[i];
      newSnap.otherTeamRBPos[i] = oldSnap.otherTeamRBPos[i];

      newSnap.ourTeamTNo[i]   = oldSnap.ourTeamTNo[i];
      newSnap.otherTeamTNo[i] = oldSnap.otherTeamTNo[i];

      newSnap.ourTeamStam[i].value = oldSnap.ourTeamStam[i].value;
      newSnap.ourTeamStam[i].age   = oldSnap.ourTeamStam[i].age + 1;

      newSnap.otherTeamStam[i].value = min( STAMINA_MAX, oldSnap.otherTeamStam[i].value +
					    snap->otherTeam( i + 1 ).staminaIncMax );
      newSnap.otherTeamStam[i].age   = oldSnap.ourTeamStam[i].age + 1;

      newSnap.otherTeamLastStamValue[i]   = oldSnap.otherTeamLastStamValue[i];
      newSnap.otherTeamLastStamPos[i]     = oldSnap.otherTeamLastStamPos[i];
      newSnap.otherTeamSimStamPossible[i] = oldSnap.otherTeamSimStamPossible[i];
    }

  // copy player type data if in coach mode
  if( snap->my.isCoach() )
    {
      for( int i = 0; i < 11; i++ )
	{
	  newSnap.oldPlayer[i] = oldSnap.oldPlayer[i];
	  for( int j = 0; j < playerParams.player_types; j++ )
	    {
	      newSnap.playerTypeData[i][j].state = oldSnap.playerTypeData[i][j].state;
	      newSnap.playerTypeData[i][j].score = oldSnap.playerTypeData[i][j].score;
	    }
	  newSnap.playerType[i] = oldSnap.playerType[i];
	}
      newSnap.oldTime = oldSnap.oldTime;
      newSnap.oldNumSubs = oldSnap.oldNumSubs;
    }

  newSnap.reportedTAction = oldSnap.reportedTAction;
}


bool
TacticalModule::feedbackWorld( Snapshot* snap )
{
  if( !snap )
    return false;

  // CAUTION: What if audio information comes first?
  // well: at the moment there is no audio information...
  // ??? This is something I wrote in 2000 and do not understand in 2001...
  if( snap->time < currentTime )
    return false;

  // simulate snapshot
  for( int t = currentTime; t < snap->time; t++ )
    simulate( snap, t );

  currentTime = snap->time;
  TacticalSnapshot &tsnap = snapshots[getTimeIndex( currentTime )];

  // ------- COACH : determine opponent player types -------
  if( snap->my.isCoach() )
    {
      tsnap.determineOtherPlayerTypes( snap );
      return true;
    }

  // ------- NO COACH : -------
  // try to determine goalie numbers and my tactical number
  tsnap.updateTacticalNumbers( snap );

  // try to update stamina of our team members
  // (and of the opponent in those situations when stamina is reset)
  tsnap.updateStamina( snap, receivedInfo, currentTime );

  // find out how long it will take players to reach the ball
  // (and determine the 3 "most responsible ones") 
  tsnap.determineResponsibleFriends( snap );

  // find out how long it will take opponents to reach the ball
  // (and determine the 3 "most dangerous ones") 
  tsnap.determineDangerousOpponents( snap );

  // determine ball control
  tsnap.determineBallControl( snap, receivedInfo.valid ? receivedInfo.ourBallControl : BC_UNKNOWN );

  // tactical action evaluation for our team
  // set emergency channel flag if there is a reported action to broadcast
  if( tsnap.determineOurTacticalAction( snap,
					receivedInfo.valid ? receivedInfo.ourAction :
					TacticalAction( TA_UNKNOWN ), currentTime ) )
    useEmergencyChannel = true;

  // reported action is processed, so set it invalid
  tsnap.reportedTAction.quality = TQ_UNKNOWN;
  tsnap.reportedTAction.state = TA_UNKNOWN;
  tsnap.reportedTAction.performer = UNKNOWN;
  tsnap.reportedTAction.receiver = UNKNOWN;

  // audial information is processed, so set receivedInfo invalid
  receivedInfo.valid = false;

  return true;
}


bool
TacticalModule::feedbackAudio( AudioMsg &audioMsg )
{
  TacticalModuleInfo recInfo( false );
  audioMsg >> recInfo;

  if( recInfo.valid )
    {
      receivedInfo = recInfo;
      return true;
    }

  return false;
}


TacticalSnapshot *
TacticalModule::getSnapshot()
{
  // invalid time
  if( currentTime < 0 ) return NULL;

  // invalid snapshot
  if( snapshots[getTimeIndex( currentTime )].time != currentTime ) return NULL;

  return &snapshots[getTimeIndex( currentTime )];
}


bool
TacticalModule::getSayMsg( AudioMsg &audioMsg )
{
  bool ok = true;

  TacticalSnapshot &dummy = snapshots[getTimeIndex( currentTime )];
  TacticalModuleInfo sendInfo( ok, dummy );

  audioMsg.markWrite();
  audioMsg << sendInfo;
      
  // check if error occurred
  if( !audioMsg.getState() )
    {
      audioMsg.rollbackWrite().clearState();
      audioMsg << false;
      ok = false;
    }

  return ok;
}


class AudioMsg& operator << ( AudioMsg &aumsg, int *playerType )
{
  for( int i = 0; i < 11; i++ )
    aumsg << NodaChar( char( playerType[i] + 1 ) );

  return aumsg;
}
