//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////
//
// Class TacticalModuleInfo
//
//////////////////////////////////////////////////////////////////////

TacticalModuleInfo::TacticalModuleInfo( bool _valid, const TacticalSnapshot &snap,
					int _sender, int _time )
{
  // if any info is to be sent, use valid = true;
  // to save space in the AudioMsg stream, use valid = false.
  
  valid = _valid;

  if( valid )
    {
      sender = _sender;
      time = _time;

      // set all other variables to values from TacticalSnapshot
      ourBallControl = snap.ourBallControl;

      ourAction = snap.ourAction;

      firstResponsiblePlayer = snap.firstResponsiblePlayer;
      secondResponsiblePlayer = snap.secondResponsiblePlayer;
      thirdResponsiblePlayer = snap.thirdResponsiblePlayer;
      
      firstDangerousOpponent = snap.firstDangerousOpponent;
      secondDangerousOpponent = snap.secondDangerousOpponent;
      thirdDangerousOpponent = snap.thirdDangerousOpponent;

      for( int i = 0; i < 11; i++ )
	{
	  ourTeamTNo[i] = snap.ourTeamTacticalNo( i + 1 );
	  otherTeamTNo[i] = snap.otherTeamTacticalNo( i + 1 );
	  ourTeamStam[i] = snap.ourTeamStamina( i + 1 );
	  otherTeamStam[i] = snap.otherTeamStamina( i + 1 );
	}
    }
  else
    {
      // set all variables invalid
      sender = UNKNOWN;
      time = -1;
      
      ourBallControl = BC_UNKNOWN;

      firstResponsiblePlayer = UNKNOWN;
      secondResponsiblePlayer = UNKNOWN;
      thirdResponsiblePlayer = UNKNOWN;

      firstDangerousOpponent = UNKNOWN;
      secondDangerousOpponent = UNKNOWN;
      thirdDangerousOpponent = UNKNOWN;
      
      for( int i = 0; i < 11; i++ )
	{
	  ourTeamTNo[i] = UNKNOWN;
	  otherTeamTNo[i] = UNKNOWN;
	  ourTeamStam[i].value = STAMINA_MAX;
	  ourTeamStam[i].age = ETERNAL;
	  otherTeamStam[i].value = STAMINA_MAX;
	  otherTeamStam[i].age = ETERNAL;
	}
    }
}


class AudioMsg &operator << ( AudioMsg &audioMsg, const TacticalModuleInfo &info )
{
  audioMsg << info.valid;

  // send tactical module information
  // space needed:
  // validity                        :    1 byte  (bool)
  // ball control                    :    1 byte  (1 NodaChar)
  // responsible players             :    3 bytes (3 NodaChars)
  // dangerous opponents             :    3 bytes (same)
  // tactical action for our team    :    7 bytes (1 for state, 1 for quality, 3 for time, 2 for players)
  // tactical numbers for our team   :   11 bytes (11 NodaChars)
  // tactical numbers for other team :   11 bytes (same)
  // stamina for our team            :   44 bytes (3*11 for value, 1*11 for age)
  // --------------------------------------------
  // total                               81 bytes

  if( info.valid )
    {
       // send ballControl variable
       char c = 0;
       switch( info.ourBallControl )
	 {
	 case BC_SAFE    : c = 0; break;
	 case BC_DANGER  : c = 1; break;
	 case BC_LOST    : c = 2; break;
	 case BC_UNKNOWN : c = 3; break;
	 default         : break;
	 }
       audioMsg << NodaChar( c );

       // send responsible and dangerous players

       // CAUTION : NodaChar does not know negative values (like UNKNOWN)
       audioMsg << NodaChar( char( info.firstResponsiblePlayer + 1 ) );
       audioMsg << NodaChar( char( info.secondResponsiblePlayer + 1 ) );
       audioMsg << NodaChar( char( info.thirdResponsiblePlayer + 1 ) );

       audioMsg << NodaChar( char( info.firstDangerousOpponent + 1 ) );
       audioMsg << NodaChar( char( info.secondDangerousOpponent + 1 ) );
       audioMsg << NodaChar( char( info.thirdDangerousOpponent + 1 ) );

       // send tactical actions for our team

       switch( info.ourAction.state )
	 {
	 case TA_CONTROL  : c = 0; break;
	 case TA_PASS     : c = 1; break;
	 case TA_GOALSHOT : c = 2; break;
	 case TA_REACT    : c = 3; break;
	 case TA_UNKNOWN  : c = 4; break;
	 default : break;
	 }
       audioMsg << NodaChar( c );

       switch( info.ourAction.quality )
 	{
 	case TQ_REPORTED        : c = 0; break;
 	case TQ_AUDIAL_REPORTED : c = 1; break;
 	case TQ_AUDIAL          : c = 2; break;
 	case TQ_GUESSED         : c = 3; break;
 	case TQ_UNKNOWN         : c = 4; break;
 	default : break;
 	}
       audioMsg << NodaChar( c );

       audioMsg << Int3B( info.ourAction.startingTime );
       audioMsg << NodaChar( char( info.ourAction.performer + 1 ) );
       audioMsg << NodaChar( char( info.ourAction.receiver + 1 ) );

       // send tactical numbers for both teams

       for( int i = 0; i < 11; i++ )
 	{
	  audioMsg << NodaChar( char( info.ourTeamTNo[i] + 1 ) );
  	  audioMsg << NodaChar( char( info.otherTeamTNo[i] + 1 ) );
 	}

       // send stamina for our team

       for( int i = 0; i < 11; i++ )
	 {
	   audioMsg << Int3B( int( info.ourTeamStam[i].value ) );
	   audioMsg << NodaChar( char( min( 63, info.ourTeamStam[i].age ) ) );
	 }
    }
  
  return audioMsg;
}


class AudioMsg &operator >> ( AudioMsg &audioMsg, TacticalModuleInfo &info )
{
  audioMsg >> info.valid;
  if( info.valid )
    {
       NodaChar c;
       Int3B d;

       info.sender = audioMsg.getSource();
       info.time = audioMsg.getServerTime();

       // get ballControl variable

       audioMsg >> c;
       switch( int( c ) )
	 {
	 case 0  : info.ourBallControl = BC_SAFE; break;
	 case 1  : info.ourBallControl = BC_DANGER; break;
	 case 2  : info.ourBallControl = BC_LOST; break;
	 case 3  : info.ourBallControl = BC_UNKNOWN; break;
	 default : break;
	 }

       // get responsible and dangerous players

       // CAUTION : NodaChar does not know negative values (like UNKNOWN)
       audioMsg >> c;
       info.firstResponsiblePlayer = int( char( c ) ) - 1;
       audioMsg >> c;
       info.secondResponsiblePlayer = int( char( c ) ) - 1;
       audioMsg >> c;
       info.thirdResponsiblePlayer = int( char( c ) ) - 1;

       audioMsg >> c;
       info.firstDangerousOpponent = int( char( c ) ) - 1;
       audioMsg >> c;
       info.secondDangerousOpponent = int( char( c ) ) - 1;
       audioMsg >> c;
       info.thirdDangerousOpponent = int( char( c ) ) - 1;

       // get tactical actions for our team

       audioMsg >> c;
       switch( char( c ) )
 	{
 	case 0  : info.ourAction.state = TA_CONTROL; break;
 	case 1  : info.ourAction.state = TA_PASS; break;
 	case 2  : info.ourAction.state = TA_GOALSHOT; break;
 	case 3  : info.ourAction.state = TA_REACT; break;
 	case 4  : info.ourAction.state = TA_UNKNOWN; break;
 	default : break;
 	}

       audioMsg >> c;
       switch( char( c ) )
 	{
 	case 0 : info.ourAction.quality = TQ_REPORTED; break;
 	case 1 : info.ourAction.quality = TQ_AUDIAL_REPORTED; break;
 	case 2 : info.ourAction.quality = TQ_AUDIAL; break;
 	case 3 : info.ourAction.quality = TQ_GUESSED; break;
 	case 4 : info.ourAction.quality = TQ_UNKNOWN; break;
 	default : break;
 	}
       
       audioMsg >> d;
       info.ourAction.startingTime = int( d );
       audioMsg >> c;
       info.ourAction.performer = int( char( c ) ) - 1;
       audioMsg >> c;
       info.ourAction.receiver = int( char( c ) ) - 1;
       
       // get tactical numbers for both teams

       for( int i = 0; i < 11; i++ )
	 {
	  audioMsg >> c;
 	  info.ourTeamTNo[i] = int( char( c ) ) - 1;
  	  audioMsg >> c;
  	  info.otherTeamTNo[i] = int( char( c ) ) - 1;
 	}

       // get stamina for our team

       for( int i = 0; i < 11; i++ )
	 {
	   audioMsg >> d;
	   info.ourTeamStam[i].value = double( int( d ) );
	   audioMsg >> c;
	   if( int( char( c ) ) == 63 )
	     info.ourTeamStam[i].age = ETERNAL;
	   else
	     info.ourTeamStam[i].age = int( char( c ) );
	 }
    }

  return audioMsg;
}
